//@ts-check

const fs = require('fs');
const path = require('path');

const baseDir = path.join(__dirname, '..');

const backstoryPath = path.join(baseDir, 'backstory-gen', 'backstory.md');

const backstoryReadablePath = path.join(baseDir, 'code', 'data', 'backstory-readable.mjs');

/** @typedef {{ location: string, type: string, others: Array<{ name: string, expected: string }>, content: string }} MdEventEntry  */

/**
 * @param {string} fileContent
 * @returns {Array<MdEventEntry>}
 */
function parseMdFile(fileContent) {
  const lines = fileContent.split('\n');
  /** @type {Array<MdEventEntry>} */
  const result = [];
  /** @type {MdEventEntry | undefined} */
  let currentEntry = undefined;
  for (const line of lines) {
    if (line.startsWith('# ')) continue;
    if (line.startsWith('## ')) {
      const header = line.substring('## '.length).trim();
      const sections = header.split(' ');
      let location = '';
      let type = '';
      /** @type {MdEventEntry['others']} */
      let others = [];
      for (const section of sections) {
        if (!location) {
          location = section;
        } else if (!type) {
          type = section;
        } else {
          const entries = section.split('=');
          others.push({ name: entries[0], expected: entries[1] });
        }
      }
      currentEntry = { location, type, others, content: '' };
      result.push(currentEntry);
      continue;
    }
    if (!currentEntry) continue;
    const trimmedLine = line.trim();
    if (!trimmedLine && !currentEntry.content) continue;
    currentEntry.content += trimmedLine;
  }

  for (const entry of result) {
    entry.content = entry.content.trim();
  }

  return result;
}

/**
 * @param {MdEventEntry} entry
 * @returns {string}
 */
function translateEntryContent(entry) {
  let result = entry.content;
  result = result.replace(/\[character\]/g, '${character.name}');
  result = result.replace(/\[group\]/g, '${getNames(names, content.group)}');
  result = result.replace(/\[teammate\]/g, '${getName(names, content.teammate)}');
  result = result.replace(/\[enemies\]/g, '${getEnemies(content.enemies)}');
  result = result.replace(/\[characterAndGroup\]/g, '${getCharacterAndNames(character, names, content.group)}');
  result = result.replace(/\[rescuer\]/g, '${getName(names, content.rescuer)}');
  result = result.replace(/\[rescuee\]/g, '${getName(names, content.rescuee)}');
  result = result.replace(/\[captured\]/g, '${getName(names, content.captured)}');
  result = result.replace(/\[capturedGroup\]/g, '${getNames(names, content.capturedGroup)}');
  result = result.replace(
    /\[characterAndCapturedGroup\]/g,
    '${getCharacterAndNames(character, names, content.capturedGroup)}',
  );
  result = result.replace(
    /\[characterGirlfriendAndCapturedGroup\]/g,
    '${getCharacterAndNames(character, names, content.capturedGroup.concat(content.girlfriend))}',
  );
  result = result.replace(/\[rescuees\]/g, '${getNames(names, content.rescuees)}');
  result = result.replace(/\[other\]/g, '${getName(names, content.other)}');
  result = result.replace(/\[friend\]/g, '${getName(names, content.friend)}');
  result = result.replace(/\[rival\]/g, '${getName(names, content.rival)}');
  result = result.replace(/\[girlfriend\]/g, '${getName(names, content.girlfriend)}');
  result = result.replace(/\[targetGirlfriend\]/g, '${getName(names, content.targetGirlfriend)}');
  result = result.replace(/\[item_name\]/g, '${content.item_name)}');
  result = result.replace(/\[new_level\]/g, '${content.new_level)}');
  if (result.includes('[') && result.includes(']')) {
    console.warn(
      `Unhandled translation for entry type in ${entry.location}/${entry.type}${entry.others.length === 0 ? '' : ` (Others: ${JSON.stringify(entry.others)})`}`,
    );
    console.warn(`Original text: ${entry.content}`);
    console.warn(`Result text: ${result}`);
  } else if (!result) {
    console.warn(
      `Incomplete translation for entry type in ${entry.location}/${entry.type}${entry.others.length === 0 ? '' : ` (Others: ${JSON.stringify(entry.others)})`}: no content provided`,
    );
  }
  return result;
}

/**
 * @param {Array<MdEventEntry>} entries
 * @returns {string}
 */
function generateCodeForEntries(entries) {
  let result = '';
  const availableTypes = [];
  for (const entry of entries) if (!availableTypes.includes(entry.type)) availableTypes.push(entry.type);
  for (const type of availableTypes) {
    const entriesForType = entries.filter((e) => e.type === type);
    if (entriesForType.length === 1) {
      result += `
  if (content.type === '${type}') {
    story.push(\`Day \${content.cycle}: ${translateEntryContent(entriesForType[0])}\`)
    return;
  }`;
    } else {
      result += `
  if (content.type === '${type}') {
`;
      const others = entriesForType[0].others;
      if (others.length === 1) {
        const criteriaName = others[0].name;
        for (const entry of entriesForType) {
          result += `
    if (content.${criteriaName} === '${entry.others[0].expected}') {
      story.push(\`Day \${content.cycle}: ${translateEntryContent(entry)}\`)
      return;
    }
  `;
        }
        result += `
      assertTypeIsHandled(content.${criteriaName}, 'backstory event extra condition');
    }`;
      } else {
        console.error(`Could not find proper conditions for the several entries using type = ${type}`);
        result += `
      assertTypeIsHandled(content, 'backstory choices not handled properly');
    }`;
      }
    }
  }
  return result;
}

/**
 * @param {Array<MdEventEntry>} entries
 * @returns {string}
 */
function generateCode(entries) {
  return `import { assertTypeIsHandled } from '../utils.mjs';
import { getAllLiveTransmutedAndDustedCharacters } from './utils.mjs';

/**
 * @param {import("../types").GameData} game_data
 * @param {import("../types").StoredCharacterData} character
 * @returns {string[]}
 */
export function generateReadableBackstory(game_data, character) {
  const characters = getAllLiveTransmutedAndDustedCharacters(game_data);
  /** @type {Record<import('../types').CharacterId, string>} */
  const names = {};
  for (const character of characters) {
    names[character.id] = character.name;
  }

  /** @type {string[]} */
  const story = [];
  for (const event of character.story.events) {
    if (event.location === 'origins') {
      for (const content of event.content) {
        readableOriginsBackstoryEvent(story, character, event, content, names);
      }
      continue;
    }
    if (event.location === 'town') {
      for (const content of event.content) {
        readableTownBackstoryEvent(story, character, event, content, names);
      }
      continue;
    }
    if (event.location === 'exploration') {
      for (const content of event.content) {
        readableExplorationBackstoryEvent(story, character, event, content, names);
      }
      continue;
    }
    if (event.location === 'captured') {
      for (const content of event.content) {
        readableCapturedBackstoryEvent(story, character, event, content, names);
      }
      continue;
    }
    if (event.location === 'transmuted') {
      for (const content of event.content) {
        readableTransmutedBackstoryEvent(story, character, event, content, names);
      }
      continue;
    }
    assertTypeIsHandled(event, 'backstory event');
  }
  return story;
}


/**
 * @param {Record<import('../types').CharacterId, string>} names
 * @param {import('../types').EventCharacterInfo} info
 */
function getName(names, info) {
  if (info.does_not_recognize_target) return 'someone';
  if (info.relationship.is_partner) {
    return \`her girlfriend \${names[info.character]}\`;
  } else if (info.relationship.is_friend) {
    return \`her friend \${names[info.character]}\`;
  } else if (info.relationship.is_rival_bully) {
    return \`her rival \${names[info.character]}\`;
  } else if (info.relationship.is_rival_target) {
    return \`her rival \${names[info.character]}\`;
  }
  return names[info.character];
}

/**
 * @param {Record<import('../types').CharacterId, string>} names
 * @param {import('../types').EventCharacterInfo[]} info
 */
function getNames(names, info) {
  let result = '';
  for (let index = 0; index < info.length; ++index) {
    result += getName(names, info[index]);
    if (index === info.length - 1) {
    } else if (index === info.length - 2) {
      result += ' and ';
    } else {
      result += ', ';
    }
  }
  return result;
}

/**
 * @param {import('../types').StoredCharacterData} character
 * @param {Record<import('../types').CharacterId, string>} names
 * @param {import('../types').EventCharacterInfo[]} info
 */
function getCharacterAndNames(character, names, info) {
  let result = character.name;
  for (let index = 0; index < info.length; ++index) {
    if (index === info.length - 1) {
      result += ' and ';
    } else {
      result += ', ';
    }
    result += getName(names, info[index]);
  }
  return result;
}

/**
 * @param {import('../types').EventEnemyInfo[]} enemies
 */
function getEnemies(enemies) {
  return 'a group of mages';
}

/**
 * @param {string[]} story
 * @param {import('../types').StoredCharacterData} character
 * @param {import('../types').CharacterBackstoryEvent} event
 * @param {import('../types').CharacterBackstoryOriginsEvent} content
 * @param {Record<import('../types').CharacterId, string>} names
 */
function readableOriginsBackstoryEvent(story, character, event, content, names) {
${generateCodeForEntries(entries.filter((e) => e.location === 'origins'))}
  assertTypeIsHandled(content, 'backstory event type');
}

/**
 * @param {string[]} story
 * @param {import('../types').StoredCharacterData} character
 * @param {import('../types').CharacterBackstoryEvent} event
 * @param {import('../types').CharacterBackstoryTownEvent} content
 * @param {Record<import('../types').CharacterId, string>} names
 */
function readableTownBackstoryEvent(story, character, event, content, names) {
${generateCodeForEntries(entries.filter((e) => e.location === 'town'))}
  assertTypeIsHandled(content, 'backstory event type');
}

/**
 * @param {string[]} story
 * @param {import('../types').StoredCharacterData} character
 * @param {import('../types').CharacterBackstoryEvent} event
 * @param {import('../types').CharacterBackstoryExplorationEvent} content
 * @param {Record<import('../types').CharacterId, string>} names
 */
function readableExplorationBackstoryEvent(story, character, event, content, names) {
${generateCodeForEntries(entries.filter((e) => e.location === 'exploration'))}
  assertTypeIsHandled(content, 'backstory event type');
}

/**
 * @param {string[]} story
 * @param {import('../types').StoredCharacterData} character
 * @param {import('../types').CharacterBackstoryEvent} event
 * @param {import('../types').CharacterBackstoryCaptureEvent} content
 * @param {Record<import('../types').CharacterId, string>} names
 */
function readableCapturedBackstoryEvent(story, character, event, content, names) {
${generateCodeForEntries(entries.filter((e) => e.location === 'captured'))}
  assertTypeIsHandled(content, 'backstory event type');
}

/**
 * @param {string[]} story
 * @param {import('../types').StoredCharacterData} character
 * @param {import('../types').CharacterBackstoryEvent} event
 * @param {import('../types').CharacterBackstoryTransmutedEvent} content
 * @param {Record<import('../types').CharacterId, string>} names
 */
function readableTransmutedBackstoryEvent(story, character, event, content, names) {
${generateCodeForEntries(entries.filter((e) => e.location === 'transmuted'))}
  assertTypeIsHandled(content, 'backstory event type');
}`;
}

const rawStr = fs.readFileSync(backstoryPath, { encoding: 'utf-8' });
const data = parseMdFile(rawStr);
fs.writeFileSync(backstoryReadablePath, generateCode(data), { encoding: 'utf-8' });
