/** @typedef {string} AudioSampleId */

import { assertTypeIsHandled } from '../utils.mjs';

/**
 * @param {import("../types").AudioAssets} audio_assets The sound snippets available in-game
 */
export function getAudioEngine(audio_assets) {
  const context = new window.AudioContext();
  const music_gain = context.createGain();
  music_gain.gain.value = 1.0;
  const effect_gain = context.createGain();
  effect_gain.gain.value = 1.0;
  /** @type {Record<AudioSampleId, { audio: HTMLMediaElement, gain: GainNode, track: MediaElementAudioSourceNode }> } */
  const registered_tracks = {};
  for (const asset_id in audio_assets.snippets) {
    const asset = audio_assets.snippets[asset_id];
    if (registered_tracks[asset_id]) throw new Error(`Cannot register the audio asset ${asset_id} twice`);
    const track = context.createMediaElementSource(asset.media);
    const gain = context.createGain();
    let connected_node = track.connect(gain);
    if (asset.kind === 'music') connected_node = connected_node.connect(music_gain);
    else if (asset.kind === 'effect') connected_node = connected_node.connect(effect_gain);
    connected_node.connect(context.destination);
    gain.gain.value = 0.0;
    registered_tracks[asset_id] = { audio: asset.media, gain: gain, track: track };
  }

  return {
    /**
     * @param {import("../types").AudioAssetKind} kind
     * @param {number} value
     */
    set_volume: (kind, value) => {
      if (kind === 'effect') {
        effect_gain.gain.setTargetAtTime(value, context.currentTime, 0.01);
        return;
      }
      if (kind === 'music') {
        music_gain.gain.setTargetAtTime(value, context.currentTime, 0.01);
        return;
      }
      assertTypeIsHandled(kind, 'Audio track kind');
    },
    /**
     * @param {AudioSampleId} identifier
     */
    start_loop: (identifier) => {
      const data = registered_tracks[identifier];
      if (!data) throw new Error(`Cannot enable audio for unknown identifier ${identifier}`);
      if (data.audio.paused) {
        try {
          data.audio.loop = true;
          data.audio.currentTime = 0;
          data.audio.play();
        } catch (e) {}
      }
    },
    /**
     * @param {AudioSampleId} identifier
     */
    stop_loop: (identifier) => {
      const data = registered_tracks[identifier];
      if (!data) throw new Error(`Cannot enable audio for unknown identifier ${identifier}`);
      if (!data.audio.paused) {
        try {
          data.audio.pause();
        } catch (e) {}
      }
    },
    /**
     * @param {AudioSampleId} identifier
     */
    loop_disable: (identifier) => {
      const data = registered_tracks[identifier];
      if (!data) throw new Error(`Cannot enable audio for unknown identifier ${identifier}`);
      try {
        data.gain.gain.setTargetAtTime(0.0, context.currentTime, 0.2);
      } catch (e) {}
    },
    /**
     * @param {AudioSampleId} identifier
     */
    loop_enable: (identifier) => {
      const data = registered_tracks[identifier];
      if (!data) throw new Error(`Cannot enable audio for unknown identifier ${identifier}`);
      try {
        data.gain.gain.setTargetAtTime(1.0, context.currentTime, 0.2);
      } catch (e) {}
    },
    /**
     * @param {AudioSampleId} identifier
     */
    play_once: (identifier) => {
      const data = registered_tracks[identifier];
      if (!data) throw new Error(`Cannot enable audio for unknown identifier ${identifier}`);
      try {
        data.gain.gain.value = 1.0;
        data.audio.loop = false;
        data.audio.currentTime = 0;
        data.audio.play();
      } catch (e) {}
    },
    stop_all: () => {
      for (const sample_id in registered_tracks) {
        try {
          registered_tracks[sample_id].audio.pause();
        } catch (e) {}
      }
    },
  };
}
