import { mouseIsInArea } from './checks.mjs';

/**
 * @param {import("../engine.mjs").Game} game
 * @param {boolean} is_enabled
 * @param {{ x: number, y: number, total_width: number, total_height: number, padding: number, padding_between: number, elevator_height: number, border_size: number }} layout
 * @param {{ border: string, background: string, elevator: string, hovered_elevator: string }} color
 * @param {import('../types').ScrollController} controller
 * @param {Array<{ width: number, display: (x: number, y: number) => void }>} content
 */
export function showScrollableHorizontalList(game, is_enabled, layout, color, controller, content) {
  const content_start_x = layout.x;
  const content_start_y = layout.y;
  const content_width = layout.total_width;
  const content_height = layout.total_height;
  const border_size = layout.border_size;
  const elevator_height = layout.elevator_height;
  const view_padding_x = layout.padding;
  const view_padding_y = layout.padding;
  const view_start_x = content_start_x + border_size;
  const view_start_y = content_start_y + border_size;
  const elevator_start_y = view_start_y + content_height - 2 * border_size - elevator_height;
  const view_width = content_width - 2 * border_size;
  const view_height = content_height - elevator_height - 3 * border_size;
  const elevator_area_start_x = view_start_x;
  const elevator_area_start_y = view_start_y + content_height - 2 * border_size - elevator_height;
  const elevator_area_width = view_width;
  const elevator_area_height = elevator_height;

  const total_content_width = content.reduce(
    (p, c) => p + c.width + layout.padding_between,
    2 * view_padding_x - layout.padding_between,
  );

  // Border and text contents
  game.context.strokeStyle = color.border;
  game.context.strokeRect(content_start_x, content_start_y, content_width, content_height);
  game.context.beginPath();
  game.context.moveTo(elevator_area_start_x, elevator_area_start_y);
  game.context.lineTo(elevator_area_start_x + elevator_area_width, elevator_area_start_y);
  game.context.stroke();
  game.context.strokeStyle = 'none';
  game.context.fillStyle = color.background;
  game.context.fillRect(view_start_x, view_start_y, view_width, view_height);

  // Elevator sidebar and contents
  const elevator_width = view_width * Math.min(1.0, view_width / total_content_width);
  const elevator_start_x = view_start_x + (view_width - elevator_width) * controller.scroll_offset;

  const is_elevator_hovered = mouseIsInArea(game, {
    x: elevator_start_x,
    y: elevator_start_y,
    width: elevator_width,
    height: elevator_height,
  });

  game.context.fillStyle = color.background;
  game.context.fillRect(elevator_area_start_x, elevator_area_start_y, elevator_area_width, elevator_area_height);
  game.context.fillStyle = is_elevator_hovered ? color.hovered_elevator : color.elevator;
  game.context.fillRect(elevator_start_x, elevator_start_y, elevator_width, elevator_height);
  if (is_enabled && game.mouse.primary_button && is_elevator_hovered) {
    controller.is_elevator_selected = true;
  }

  if (!game.mouse.primary_button) {
    controller.is_elevator_selected = false;
  }

  if (controller.is_elevator_selected) {
    const equivalent_center = game.mouse.x - view_start_x - elevator_width / 2;
    const equivalent_width = view_width - elevator_width;
    controller.scroll_offset = equivalent_center / equivalent_width;
  }

  if (mouseIsInArea(game, { x: content_start_x, y: content_start_y, width: content_width, height: content_height })) {
    if (is_enabled && game.mouse.scroll_down) controller.scroll_offset -= 100 / total_content_width;
    if (is_enabled && game.mouse.scroll_up) controller.scroll_offset += 100 / total_content_width;
  }
  if (controller.scroll_offset > 1.0) controller.scroll_offset = 1.0;
  if (controller.scroll_offset < 0.0) controller.scroll_offset = 0.0;

  game.context.save();
  game.context.beginPath();
  game.context.rect(view_start_x, view_start_y, view_width, view_height);
  game.context.clip();
  const offset = (total_content_width - view_width) * controller.scroll_offset;
  let start_x = view_start_x + view_padding_x;
  if (offset > 0) {
    start_x -= offset;
  }

  for (const entry of content) {
    if (start_x + entry.width > 0 && start_x < view_start_x + view_width)
      entry.display(start_x, view_start_y + view_padding_y);
    start_x += entry.width + layout.padding_between;
  }
  game.context.restore();
}
