import { mouseIsInArea } from './checks.mjs';

/**
 * @param {import("../engine.mjs").Game} game
 * @param {boolean} is_enabled
 * @param {{ x: number, y: number, total_width: number, total_height: number, padding: number, padding_between: number, item_width: number, item_height: number, elevator_width: number, border_size: number }} layout
 * @param {{ border: string, background: string, elevator: string, hovered_elevator: string }} color
 * @param {import('../types').ScrollController} controller
 * @param {Array<{ display: (x: number, y: number) => void }>} content
 */
export function showScrollableVerticalGrid(game, is_enabled, layout, color, controller, content) {
  const content_start_x = layout.x;
  const content_start_y = layout.y;
  const content_width = layout.total_width;
  const content_height = layout.total_height;
  const border_size = layout.border_size;
  const elevator_width = layout.elevator_width;
  const view_padding_x = layout.padding;
  const view_padding_y = layout.padding;
  const view_start_x = content_start_x + border_size;
  const view_start_y = content_start_y + border_size;
  const elevator_start_x = view_start_x + content_width - 2 * border_size - elevator_width;
  const view_width = content_width - elevator_width - 3 * border_size;
  const view_height = content_height - 2 * border_size;
  const elevator_area_start_x = view_start_x + content_width - 2 * border_size - elevator_width;
  const elevator_area_start_y = view_start_y;
  const elevator_area_width = elevator_width;
  const elevator_area_height = view_height;

  const items_per_line = Math.floor(
    (view_width - view_padding_x * 2 + layout.padding_between) / (layout.item_width + layout.padding_between),
  );
  const total_lines = Math.ceil(content.length / items_per_line);
  const total_content_height =
    2 * view_padding_y + total_lines * (layout.item_height + layout.padding_between) - layout.padding_between;

  // Border and text contents
  game.context.strokeStyle = color.border;
  game.context.strokeRect(content_start_x, content_start_y, content_width, content_height);
  game.context.beginPath();
  game.context.moveTo(elevator_area_start_x, elevator_area_start_y);
  game.context.lineTo(elevator_area_start_x, elevator_area_start_y + elevator_area_height);
  game.context.stroke();
  game.context.strokeStyle = 'none';
  game.context.fillStyle = color.background;
  game.context.fillRect(view_start_x, view_start_y, view_width, view_height);

  // Elevator sidebar and contents
  const elevator_height = view_height * Math.min(1.0, view_height / total_content_height);
  const elevator_start_y = view_start_y + (view_height - elevator_height) * controller.scroll_offset;

  const is_elevator_hovered = mouseIsInArea(game, {
    x: elevator_start_x,
    y: elevator_start_y,
    width: elevator_width,
    height: elevator_height,
  });

  game.context.fillStyle = color.background;
  game.context.fillRect(elevator_area_start_x, elevator_area_start_y, elevator_area_width, elevator_area_height);
  game.context.fillStyle = is_elevator_hovered ? color.hovered_elevator : color.elevator;
  game.context.fillRect(elevator_start_x, elevator_start_y, elevator_width, elevator_height);
  if (is_enabled && game.mouse.primary_button && is_elevator_hovered) {
    controller.is_elevator_selected = true;
  }

  if (!game.mouse.primary_button) {
    controller.is_elevator_selected = false;
  }

  if (controller.is_elevator_selected) {
    const equivalent_center = game.mouse.y - view_start_y - elevator_height / 2;
    const equivalent_height = view_height - elevator_height;
    controller.scroll_offset = equivalent_center / equivalent_height;
  }

  if (mouseIsInArea(game, { x: content_start_x, y: content_start_y, width: content_width, height: content_height })) {
    if (is_enabled && game.mouse.scroll_down) controller.scroll_offset -= 100 / total_content_height;
    if (is_enabled && game.mouse.scroll_up) controller.scroll_offset += 100 / total_content_height;
  }
  if (controller.scroll_offset > 1.0) controller.scroll_offset = 1.0;
  if (controller.scroll_offset < 0.0) controller.scroll_offset = 0.0;

  game.context.save();
  game.context.beginPath();
  game.context.rect(view_start_x, view_start_y, view_width, view_height);
  game.context.clip();
  const offset = (total_content_height - view_height) * controller.scroll_offset;
  let start_x = view_start_x + view_padding_x;
  let start_y = view_start_y + view_padding_y;
  if (offset > 0) {
    start_y -= offset;
  }

  for (const entry of content) {
    if (start_y + layout.item_height > view_start_y && start_y < view_start_y + view_height) {
      entry.display(start_x, start_y);
    }
    start_x += layout.item_width + layout.padding_between;
    if (start_x + layout.item_width > view_start_x + view_width - view_padding_x) {
      start_x = view_start_x + view_padding_x;
      start_y += layout.item_height + layout.padding_between;
    }
  }
  game.context.restore();
}
