import { mouseIsInArea } from './checks.mjs';

/**
 * @param {import("../engine.mjs").Game} game
 * @param {boolean} is_enabled
 * @param {{ x: number, y: number, total_width: number, total_height: number, padding: number, elevator_width: number, border_size: number }} layout
 * @param {{ border: string, background: string, text: string, elevator: string, hovered_elevator: string }} color
 * @param {import('../types').ScrollController} controller
 * @param {Array<{ text: string, font_size: number }>} raw_text_content
 */
export function showScrollableTextBox(game, is_enabled, layout, color, controller, raw_text_content) {
  /**
   * @param {Array<{ text: string, font_size: number }>} raw_text
   * @param {number} view_width
   * @returns {Array<{ text: string, font_size: number }>}
   */
  function splitTextAtLength(raw_text, view_width) {
    const result = [];
    for (const entry of raw_text) {
      game.context.font = `${entry.font_size}px sans-serif`;
      let split_entries = entry.text.split(' ');
      while (split_entries.length > 0) {
        let end = split_entries.length;
        while (true) {
          const reconstructed = split_entries.slice(0, end).join(' ');
          const size = game.context.measureText(reconstructed);
          if (size.width < view_width) {
            result.push({ text: reconstructed, font_size: entry.font_size });
            break;
          }
          const percent = view_width / size.width;
          end = Math.min(Math.ceil(split_entries.length * percent) + 5, end - 1);
        }
        split_entries.splice(0, end);
      }
    }
    return result;
  }

  const content_start_x = layout.x;
  const content_start_y = layout.y;
  const content_width = layout.total_width;
  const content_height = layout.total_height;
  const border_size = layout.border_size;
  const elevator_width = layout.elevator_width;
  const view_padding_x = layout.padding;
  const view_padding_y = layout.padding;
  const view_start_x = content_start_x + border_size;
  const view_start_y = content_start_y + border_size;
  const elevator_start_x = view_start_x + content_width - 2 * border_size - elevator_width;
  const view_width = content_width - elevator_width - 3 * border_size;
  const view_height = content_height - 2 * border_size;
  const elevator_area_start_x = view_start_x + content_width - 2 * border_size - elevator_width;
  const elevator_area_start_y = view_start_y;
  const elevator_area_width = elevator_width;
  const elevator_area_height = view_height;

  const backstory = splitTextAtLength(raw_text_content, view_width - 2 * view_padding_x);
  const total_content_height = 2 * view_padding_y + backstory.length * 17;

  // Border and text contents
  game.context.fillStyle = color.border;
  game.context.fillRect(content_start_x, content_start_y, content_width, content_height);
  game.context.fillStyle = color.background;
  game.context.fillRect(view_start_x, view_start_y, view_width, view_height);

  // Elevator sidebar and contents
  const elevator_height = view_height * Math.min(1.0, view_height / total_content_height);
  const elevator_start_y = view_start_y + (view_height - elevator_height) * controller.scroll_offset;

  const is_elevator_hovered = mouseIsInArea(game, {
    x: elevator_start_x,
    y: elevator_start_y,
    width: elevator_width,
    height: elevator_height,
  });

  game.context.fillStyle = color.background;
  game.context.fillRect(elevator_area_start_x, elevator_area_start_y, elevator_area_width, elevator_area_height);
  game.context.fillStyle = is_elevator_hovered ? color.hovered_elevator : color.elevator;
  game.context.fillRect(elevator_start_x, elevator_start_y, elevator_width, elevator_height);
  if (is_enabled && game.mouse.primary_button && is_elevator_hovered) {
    controller.is_elevator_selected = true;
  }

  if (!game.mouse.primary_button) {
    controller.is_elevator_selected = false;
  }

  if (controller.is_elevator_selected) {
    const equivalent_center = game.mouse.y - view_start_y - elevator_height / 2;
    const equivalent_height = view_height - elevator_height;
    controller.scroll_offset = equivalent_center / equivalent_height;
  }

  if (mouseIsInArea(game, { x: content_start_x, y: content_start_y, width: content_width, height: content_height })) {
    if (is_enabled && game.mouse.scroll_down) controller.scroll_offset -= 100 / total_content_height;
    if (is_enabled && game.mouse.scroll_up) controller.scroll_offset += 100 / total_content_height;
  }
  if (controller.scroll_offset > 1.0) controller.scroll_offset = 1.0;
  if (controller.scroll_offset < 0.0) controller.scroll_offset = 0.0;

  game.context.save();
  game.context.beginPath();
  game.context.rect(view_start_x, view_start_y, view_width, view_height);
  game.context.clip();
  const offset = (total_content_height - view_height) * controller.scroll_offset;
  let start_y = view_start_y + view_padding_y;
  if (offset > 0) {
    start_y -= offset;
  }

  game.context.fillStyle = color.text;
  for (const entry of backstory) {
    start_y += entry.font_size + 2;
    game.context.font = `${entry.font_size}px sans-serif`;
    if (start_y < 0) continue;
    game.context.fillText(entry.text, view_start_x + view_padding_x, start_y - view_padding_y);
  }
  game.context.restore();
}
