import { assertTypeIsHandled } from '../utils.mjs';
import { getRandomInArray } from '../random.mjs';
import { getStatsModifierDescription } from './stats-modifiers.mjs';
import { getRequiredExperienceForPlayerLevel } from './player-character.mjs';

/**
 * @typedef {"cloth" | "leather" | "plated"} BodyEquipmentType
 * @typedef {"diadem" | "necklace" | "amulet"} AccessoryEquipmentType
 * @typedef {"staff" | "sword"} WeaponEquipmentType
 */

/**
 * @type {{
 *     body: { [key in BodyEquipmentType]: { name: string, image: string } },
 *     accessory: { [key in AccessoryEquipmentType]: { name: string, image: string } },
 *     weapon: { [key in WeaponEquipmentType]: { name: string, image: string } },
 * }} */
const raw_data = {
  body: {
    cloth: { name: 'Cloth Dress', image: 'item_cloth_dress' },
    leather: { name: 'Leather Armor', image: 'item_leather_armor' },
    plated: { name: 'Plated Armor', image: 'item_plated_armor' },
  },
  accessory: {
    diadem: { name: 'Serted Diadem', image: 'item_diadem' },
    necklace: { name: 'Sapphire Necklace', image: 'item_gem_necklace' },
    amulet: { name: 'Sapphire Amulet', image: 'item_gem_wrist_amulet' },
  },
  weapon: {
    staff: { name: "Healer's Staff", image: 'item_staff' },
    sword: { name: "Adventurer's Sword", image: 'item_sword' },
  },
};

/**
 * @param {BodyEquipmentType} type
 * @param {import('../types').StatsModifierCharacteristics} stats
 * @param {number} level
 * @returns {import("../types").StoredEquipmentData}
 */
export function generateStoredBodyEquipment(type, stats, level) {
  const entry = raw_data.body[type];
  return { type: 'body', name: entry.name, image: entry.image, level: level, stats: stats };
}

/**
 * @param {AccessoryEquipmentType} type
 * @param {import('../types').StatsModifierCharacteristics} stats
 * @param {number} level
 * @returns {import("../types").StoredEquipmentData}
 */
export function generateStoredAccessoryEquipment(type, stats, level) {
  const entry = raw_data.accessory[type];
  return { type: 'accessory', name: entry.name, image: entry.image, level: level, stats: stats };
}

/**
 * @param {WeaponEquipmentType} type
 * @param {import('../types').StatsModifierCharacteristics} stats
 * @param {number} level
 * @returns {import("../types").StoredEquipmentData}
 */
export function generateStoredWeaponEquipment(type, stats, level) {
  const entry = raw_data.weapon[type];
  return { type: 'weapon', name: entry.name, image: entry.image, level: level, stats: stats };
}

/**
 * @param {BodyEquipmentType | AccessoryEquipmentType | WeaponEquipmentType} type
 * @param {number} level
 * @returns {import("../types").StatsModifierCharacteristics}
 */
export function generateEquipmentStatisticsForType(type, level) {
  if (type === 'cloth') return generateEquipmentStatisticsForModifier('mp_regen_modifier', level);
  if (type === 'leather') return generateEquipmentStatisticsForModifier('cooldown_modifier', level);
  if (type === 'plated') return generateEquipmentStatisticsForModifier('hp_modifier', level);
  if (type === 'diadem') return generateEquipmentStatisticsForModifier('cooldown_modifier', level);
  if (type === 'necklace') return generateEquipmentStatisticsForModifier('mp_modifier', level);
  if (type === 'amulet') return generateEquipmentStatisticsForModifier('mp_regen_modifier', level);
  if (type === 'staff') return generateEquipmentStatisticsForModifier('heal_modifier', level);
  if (type === 'sword') return generateEquipmentStatisticsForModifier('damage_modifier', level);
  assertTypeIsHandled(type, 'equipment type');
}

/**
 * @param {import('../types').StatsModifierCharacteristics["type"]} type
 * @param {number} level
 * @returns {import("../types").StatsModifierCharacteristics}
 */
export function generateEquipmentStatisticsForModifier(type, level) {
  if (type === 'mp_regen_modifier') return { type: 'mp_regen_modifier', additional_mp_regen: level * 0.1 };
  if (type === 'mp_cost_modifier') return { type: 'mp_cost_modifier', mp_cost_divider: level * 0.05 };
  if (type === 'cooldown_modifier') return { type: 'cooldown_modifier', cooldown_divider: level * 0.05 };
  if (type === 'hp_modifier') return { type: 'hp_modifier', additional_hp: level * 10 };
  if (type === 'mp_modifier') return { type: 'mp_modifier', additional_mp: level * 10 };
  if (type === 'heal_modifier') return { type: 'heal_modifier', additional_heal: level * 5 };
  if (type === 'damage_modifier') return { type: 'damage_modifier', additional_damage: level * 5 };
  if (type === 'counterspell_self_modifier')
    return { type: 'counterspell_self_modifier', increase_percent: 0.1 * level };
  if (type === 'counterspell_team_modifier')
    return { type: 'counterspell_team_modifier', increase_percent: 0.1 * level };
  if (type === 'evasion_modifier') return { type: 'evasion_modifier', increase_percent: 0.1 * level };
  if (type === 'redirect_attack_to_character_modifier')
    return { type: 'redirect_attack_to_character_modifier', increase_percent: 0.1 * level };
  if (type === 'redirected_attack_damage_modifier')
    return { type: 'redirected_attack_damage_modifier', damage_multiplier: 1 - 0.1 * level };
  assertTypeIsHandled(type, 'equipment statistics type');
}

/**
 * @param {number} level
 * @returns {import("../types").StoredEquipmentData}
 */
export function generateRandomStoredEquipment(level) {
  /** @type {Array<import("../types").StoredEquipmentData["type"]>} */
  const kinds = ['body', 'accessory', 'weapon'];
  const kind = getRandomInArray(kinds);
  if (kind === 'body') {
    /** @type {BodyEquipmentType[]} */
    const equipment_types = ['cloth', 'leather', 'plated'];
    const type = getRandomInArray(equipment_types);
    return generateStoredBodyEquipment(type, generateEquipmentStatisticsForType(type, level), level);
  }
  if (kind === 'accessory') {
    /** @type {AccessoryEquipmentType[]} */
    const equipment_types = ['amulet', 'diadem', 'necklace'];
    const type = getRandomInArray(equipment_types);
    return generateStoredAccessoryEquipment(type, generateEquipmentStatisticsForType(type, level), level);
  }
  if (kind === 'weapon') {
    /** @type {WeaponEquipmentType[]} */
    const equipment_types = ['sword', 'staff'];
    const type = getRandomInArray(equipment_types);
    return generateStoredWeaponEquipment(type, generateEquipmentStatisticsForType(type, level), level);
  }
  assertTypeIsHandled(kind, 'equipment kind');
}

/**
 * @param {import("../types").StoredCharacterData} character
 * @param {number} item_to_character_exp_ratio If greater than 1, it's inefficent. If less than 1, it's efficient.
 * @returns {import("../types").StoredEquipmentData}
 */
export function getEquipmentFromCharacter(character, item_to_character_exp_ratio) {
  let total_available_exp = character.partial_exp;
  while (character.level > 1) {
    total_available_exp += getRequiredExperienceForPlayerLevel(character.level);
    character.level -= 1;
  }

  let total_levels = 1;
  while (total_available_exp > 0) {
    let next_exp_requirement = getRequiredExperienceForPlayerLevel(total_levels) * item_to_character_exp_ratio;
    total_available_exp -= next_exp_requirement;
    if (total_available_exp < 0) break;
    total_levels += 1;
  }

  const device = generateRandomStoredEquipment(total_levels);
  device.character = { is_discovered_by_player: true, data: character };
  return device;
}

/**
 * @param {import("../types").StoredEquipmentData} equipment
 * @param {import("../types").StoredCharacterData} character
 * @param {number} item_to_character_exp_ratio If greater than 1, it's inefficent. If less than 1, it's efficient.
 */
export function drainLevelIntoEquipment(equipment, character, item_to_character_exp_ratio) {
  let remaining_exp_drain = getRequiredExperienceForPlayerLevel(equipment.level + 1) * item_to_character_exp_ratio;
  while (remaining_exp_drain > 0) {
    let available = Math.min(character.partial_exp, remaining_exp_drain);
    character.partial_exp -= available;
    remaining_exp_drain -= available;
    if (remaining_exp_drain > 0) {
      if (character.level > 1) {
        character.partial_exp = getRequiredExperienceForPlayerLevel(character.level);
        character.level -= 1;
      } else {
        break;
      }
    }
  }
  equipment.level += 1;
  equipment.stats = generateEquipmentStatisticsForModifier(equipment.stats.type, equipment.level);
}

/** @param {import("../types").StoredEquipmentData} equipment */
export function getFaceImageFromTransmutedCharacter(equipment) {
  return 'character_face_transmuted';
}

/**
 * @param {import("../types").StoredEquipmentData} equipment
 * @returns {string}
 */
export function getEquipmentDescription(equipment) {
  let text = `lv. ${equipment.level} ${equipment.name}`;
  text += `\n${getStatsModifierDescription(equipment.stats)}`;
  if (equipment.character && equipment.character.is_discovered_by_player) {
    text += `\nTransmuted from ${equipment.character.data.name}, lv. ${equipment.character.data.level} ${equipment.character.data.role}`;
  }
  return text;
}

/**
 * @param {import("../types").GameData} game_data The current game data
 * @param {import('../types').TownGameData} town_data The current game
 */
export function recomputeEquipmentTargetCycle(game_data, town_data) {
  if (!town_data.upgrade_equipment.active) return;
  if (!town_data.upgrade_equipment.target) {
    town_data.upgrade_equipment.retarget_end_cycle = 0;
    return;
  }
  if (!town_data.upgrade_equipment.handler) {
    town_data.upgrade_equipment.retarget_end_cycle = 0;
    return;
  }
  let level_diff = town_data.upgrade_equipment.handler.level - town_data.upgrade_equipment.target.level - 1;
  if (town_data.use_dust_for_upgrade) {
    if (level_diff < 10) level_diff = Math.min(10, level_diff + town_data.available_dust);
  }
  if (level_diff < 0) {
    town_data.upgrade_equipment.retarget_end_cycle = 0;
    return;
  }
  if (level_diff >= 10) {
    town_data.upgrade_equipment.retarget_end_cycle = game_data.game_state.current_cycle + 1;
    return;
  }
  const cycle_count = Math.ceil((10 - level_diff) * 1.5);
  town_data.upgrade_equipment.retarget_end_cycle = game_data.game_state.current_cycle + cycle_count;
}
