import { assertTypeIsHandled } from '../utils.mjs';
import { getInGameSkillsFromList, getSkillDescriptionForInGameCharacter } from './player-skills.mjs';
import {
  binaryChoiceResult,
  generateUUID,
  getRandomInArray,
  getRandomInteger,
  getRandomInWeightedArray,
} from '../random.mjs';
import { getInGameStatisticsFromModifiers } from './stats-modifiers.mjs';
import { getEventCharacterInfo } from '../events/add-backstory.mjs';
import { getRelationsFor } from './backstory.mjs';

/** @type {{ [key in import("../types").CharacterRole]: import("../types").SkillId[] }} */
export const skills_data = {
  adventurer: ['attack_first', 'magic_attack_group', 'heal_self'],
  mage: ['magic_attack_first', 'heal_group', 'dispel_group'],
  rogue: ['attack_first', 'attack_group', 'heal_self'],
  warrior: ['attack_first_heavy', 'magic_attack_group', 'heal_self'],
  mirror: [],
  statue: [],
};

/**
 * @param {string[]} [exception_list] List of names to not generate (will be ignored if there's no names not in the list)
 * @returns {string}
 */
export function generateHumanName(exception_list) {
  const base_choices = [
    'Jessica',
    'Ashley',
    'Emily',
    'Samantha',
    'Sarah',
    'Taylor',
    'Hannah',
    'Brittany',
    'Amanda',
    'Elizabeth',
    'Kayla',
    'Rachel',
    'Megan',
    'Lauren',
    'Stephanie',
    'Courtney',
    'Jennifer',
    'Nicole',
    'Victoria',
    'Brianna',
    'Amber',
    'Morgan',
    'Danielle',
    'Jasmine',
    'Alexandra',
    'Alyssa',
    'Rebecca',
    'Madison',
    'Katherine',
    'Anna',
    'Haley',
    'Kelsey',
    'Allison',
    'Melissa',
    'Abigail',
    'Kimberly',
    'Olivia',
    'Mary',
    'Michelle',
    'Kaitlyn',
    'Sydney',
    'Maria',
    'Christina',
    'Tiffany',
    'Chelsea',
    'Sara',
    'Erin',
    'Ophelia',
    'Jane',
    'Yasmin',
    'Felicia',
  ];
  const choices = exception_list ? base_choices.filter((name) => !exception_list.includes(name)) : base_choices;
  return getRandomInArray(choices);
}

/**
 * @param {import('../types').StoredCharacterData} character
 */
export function addRandomOriginStory(character) {
  const choices = /** @type {const} */ ([
    { data: 'vagrant', weight: 20 },
    { data: 'vagrant_magic', weight: 4 },
    { data: 'commoner', weight: 100 },
    { data: 'commoner_money', weight: 6 },
    { data: 'commoner_magic', weight: 3 },
    { data: 'noble_money', weight: 20 },
    { data: 'noble_magic', weight: 1 },
  ]);
  const selectedType = getRandomInWeightedArray(choices);
  character.story.events.push({
    location: 'origins',
    content: [{ type: selectedType, cycle: 0 }],
  });
}

/**
 * @param {import('../types').StoredCharacterData} character1
 * @param {import('../types').StoredCharacterData} character2
 */
export function addGirlfriendsOriginBackstory(character1, character2) {
  const c1Origins = character1.story.events.find((l) => l.location === 'origins');
  const c2Origins = character2.story.events.find((l) => l.location === 'origins');
  if (!c1Origins) return;
  if (!c2Origins) return;
  const c1Relations = getRelationsFor(character1, character2.id);
  const c2Relations = getRelationsFor(character2, character1.id);
  c1Relations.status.is_partner = true;
  c2Relations.status.is_partner = true;
  const friendships = getRandomInteger(12, 36);
  c1Relations.interactions.friendship_interactions = friendships;
  c2Relations.interactions.friendship_interactions = friendships;
  const romances = getRandomInteger(2, 14);
  c1Relations.interactions.romance_interactions = romances;
  c2Relations.interactions.romance_interactions = romances;

  const encounter = getRandomInArray(
    /** @type {const} */ (['girlfriend_encounter_childhood', 'girlfriend_encounter_adulthood']),
  );
  c1Origins.content.push({ type: encounter, cycle: 0, girlfriend: getEventCharacterInfo(character1, character2.id) });
  c2Origins.content.push({ type: encounter, cycle: 0, girlfriend: getEventCharacterInfo(character2, character1.id) });
}

/**
 * @param {import('../types').StoredCharacterData} character1
 * @param {import('../types').StoredCharacterData} character2
 */
export function addFriendsOriginBackstory(character1, character2) {
  const c1Origins = character1.story.events.find((l) => l.location === 'origins');
  const c2Origins = character2.story.events.find((l) => l.location === 'origins');
  if (!c1Origins) return;
  if (!c2Origins) return;
  const c1Relations = getRelationsFor(character1, character2.id);
  const c2Relations = getRelationsFor(character2, character1.id);
  c1Relations.status.is_friend = true;
  c2Relations.status.is_friend = true;
  const friendships = getRandomInteger(5, 9);
  c1Relations.interactions.friendship_interactions = friendships;
  c2Relations.interactions.friendship_interactions = friendships;

  const encounter = getRandomInArray(
    /** @type {const} */ (['friend_encounter_childhood', 'friend_encounter_adulthood']),
  );
  c1Origins.content.push({ type: encounter, cycle: 0, friend: getEventCharacterInfo(character1, character2.id) });
  c2Origins.content.push({ type: encounter, cycle: 0, friend: getEventCharacterInfo(character2, character1.id) });
}

/**
 * @param {import('../types').StoredCharacterData} character1
 * @param {import('../types').StoredCharacterData} character2
 */
export function addRivalsOriginBackstory(character1, character2) {
  const c1Origins = character1.story.events.find((l) => l.location === 'origins');
  const c2Origins = character2.story.events.find((l) => l.location === 'origins');
  if (!c1Origins) return;
  if (!c2Origins) return;
  const c1Relations = getRelationsFor(character1, character2.id);
  const c2Relations = getRelationsFor(character2, character1.id);
  c1Relations.status.is_rival_bully = true;
  c2Relations.status.is_rival_target = true;
  const friendships = getRandomInteger(1, 3);
  c1Relations.interactions.friendship_interactions = friendships;
  c2Relations.interactions.friendship_interactions = friendships;
  const events = getRandomInteger(10, 20);
  c1Relations.interactions.rivalry_bullies_interactions = events;
  c2Relations.interactions.rivalry_is_bullied_interactions = events;
  const counters = getRandomInteger(0, 4);
  c1Relations.interactions.rivalry_is_bullied_interactions = counters;
  c2Relations.interactions.rivalry_bullies_interactions = counters;

  const encounter = getRandomInArray(/** @type {const} */ (['childhood', 'adulthood']));
  if (encounter === 'childhood') {
    c1Origins.content.push({
      type: 'rival_encounter_childhood_bullies',
      cycle: 0,
      rival: getEventCharacterInfo(character1, character2.id),
    });
    c2Origins.content.push({
      type: 'rival_encounter_childhood_is_bullied',
      cycle: 0,
      rival: getEventCharacterInfo(character2, character1.id),
    });
  } else if (encounter === 'adulthood') {
    c1Origins.content.push({
      type: 'rival_encounter_adulthood_bullies',
      cycle: 0,
      rival: getEventCharacterInfo(character1, character2.id),
    });
    c2Origins.content.push({
      type: 'rival_encounter_adulthood_is_bullied',
      cycle: 0,
      rival: getEventCharacterInfo(character2, character1.id),
    });
  }
}

/**
 * @param {import('../types').StoredCharacterData} character
 */
export function addTownArrivalAloneOriginStory(character) {
  const origins = character.story.events.find((l) => l.location === 'origins');
  if (!origins) return;
  origins.content.push({ type: 'alone_town_event', cycle: 0 });
  character.story.events.push({
    location: 'town',
    cycle_start: 0,
    content: [{ type: 'arrival_base_alone', cycle: 0 }],
  });
}

/**
 * @param {import('../types').StoredCharacterData} character
 * @param {import('../types').StoredCharacterData[]} team
 */
export function addTownArrivalTeamOriginStory(character, team) {
  const origins = character.story.events.find((l) => l.location === 'origins');
  if (!origins) return;
  const group = team.map((t) => getEventCharacterInfo(character, t.id));
  origins.content.push({ type: 'group_town_event', cycle: 0, group: group });
  character.story.events.push({
    location: 'town',
    cycle_start: 0,
    content: [{ type: 'arrival_base_group', cycle: 0, group: group }],
  });
}

/**
 * @param {import('../types').StoredCharacterData} character
 * @param {number} cycle
 */
export function addCaptureAloneOriginStory(character, cycle) {
  const origins = character.story.events.find((l) => l.location === 'origins');
  if (!origins) return;
  origins.content.push({ type: 'alone_capture_event', cycle: cycle, enemies: [{ kind: 'mage' }] });
  character.story.events.push({
    location: 'captured',
    cycle_start: cycle,
    content: [{ type: 'alone_capture_start', cycle: cycle }],
  });
}

/**
 * @param {import('../types').StoredCharacterData} character
 * @param {import('../types').StoredCharacterData[]} team
 * @param {number} cycle
 */
export function addCaptureTeamOriginStory(character, team, cycle) {
  const origins = character.story.events.find((l) => l.location === 'origins');
  if (!origins) return;
  const group = team.map((t) => getEventCharacterInfo(character, t.id));
  origins.content.push({ type: 'group_capture_event', cycle: cycle, enemies: [{ kind: 'mage' }], group: group });
  character.story.events.push({
    location: 'captured',
    cycle_start: cycle,
    content: [{ type: 'group_capture_start', cycle: cycle, capturedGroup: group }],
  });
}

/**
 * @returns {import('../types').StoredCharacterData[]}
 */
export function generateDefaultAvailableTeam() {
  const names = [generateHumanName([])];
  names.push(generateHumanName(names));
  names.push(generateHumanName(names));
  names.push(generateHumanName(names));
  /** @type {import('../types').StoredCharacterData[]} */
  const characters = [
    generateStoredCharacter('adventurer', 1, { relations: [], events: [] }, names[0]),
    generateStoredCharacter('adventurer', 1, { relations: [], events: [] }, names[1]),
    generateStoredCharacter('adventurer', 1, { relations: [], events: [] }, names[2]),
    generateStoredCharacter('adventurer', 1, { relations: [], events: [] }, names[3]),
  ];
  addRandomOriginStory(characters[0]);
  addRandomOriginStory(characters[1]);
  addRandomOriginStory(characters[2]);
  addRandomOriginStory(characters[3]);
  const groupSize = getRandomInArray([1, 2, 3, 4]);
  if (groupSize === 4) {
    if (binaryChoiceResult(0.1)) {
      addGirlfriendsOriginBackstory(characters[0], characters[1]);
      addFriendsOriginBackstory(characters[0], characters[2]);
      addFriendsOriginBackstory(characters[0], characters[3]);
      addFriendsOriginBackstory(characters[1], characters[2]);
      addFriendsOriginBackstory(characters[1], characters[3]);
      addFriendsOriginBackstory(characters[2], characters[3]);
    } else if (binaryChoiceResult(0.1)) {
      addRivalsOriginBackstory(characters[0], characters[1]);
      addFriendsOriginBackstory(characters[0], characters[2]);
      addFriendsOriginBackstory(characters[0], characters[3]);
      addFriendsOriginBackstory(characters[1], characters[2]);
      addFriendsOriginBackstory(characters[1], characters[3]);
      addFriendsOriginBackstory(characters[2], characters[3]);
    } else if (binaryChoiceResult(0.05)) {
      addRivalsOriginBackstory(characters[0], characters[1]);
      addFriendsOriginBackstory(characters[0], characters[2]);
      addFriendsOriginBackstory(characters[0], characters[3]);
      addFriendsOriginBackstory(characters[1], characters[2]);
      addFriendsOriginBackstory(characters[1], characters[3]);
      addGirlfriendsOriginBackstory(characters[2], characters[3]);
    } else if (binaryChoiceResult(0.2)) {
      addFriendsOriginBackstory(characters[0], characters[1]);
      addFriendsOriginBackstory(characters[2], characters[3]);
    } else {
      addFriendsOriginBackstory(characters[0], characters[1]);
      addFriendsOriginBackstory(characters[0], characters[2]);
      addFriendsOriginBackstory(characters[0], characters[3]);
      addFriendsOriginBackstory(characters[1], characters[2]);
      addFriendsOriginBackstory(characters[1], characters[3]);
      addFriendsOriginBackstory(characters[2], characters[3]);
    }
    addTownArrivalTeamOriginStory(characters[0], [characters[1], characters[2], characters[3]]);
    addTownArrivalTeamOriginStory(characters[1], [characters[0], characters[2], characters[3]]);
    addTownArrivalTeamOriginStory(characters[2], [characters[0], characters[1], characters[3]]);
    addTownArrivalTeamOriginStory(characters[3], [characters[0], characters[1], characters[2]]);
  } else if (groupSize === 3) {
    if (binaryChoiceResult(0.1)) {
      addGirlfriendsOriginBackstory(characters[0], characters[1]);
      addFriendsOriginBackstory(characters[0], characters[2]);
      addFriendsOriginBackstory(characters[1], characters[2]);
    } else if (binaryChoiceResult(0.1)) {
      addRivalsOriginBackstory(characters[0], characters[1]);
      addFriendsOriginBackstory(characters[0], characters[2]);
      addFriendsOriginBackstory(characters[1], characters[2]);
    } else if (binaryChoiceResult(0.05)) {
      addRivalsOriginBackstory(characters[0], characters[1]);
      addGirlfriendsOriginBackstory(characters[0], characters[2]);
      addFriendsOriginBackstory(characters[1], characters[2]);
    } else {
      addFriendsOriginBackstory(characters[0], characters[1]);
      addFriendsOriginBackstory(characters[0], characters[2]);
      addFriendsOriginBackstory(characters[1], characters[2]);
    }
    addTownArrivalTeamOriginStory(characters[0], [characters[1], characters[2]]);
    addTownArrivalTeamOriginStory(characters[1], [characters[0], characters[2]]);
    addTownArrivalTeamOriginStory(characters[2], [characters[0], characters[1]]);
    addTownArrivalAloneOriginStory(characters[3]);
  } else if (groupSize === 2) {
    if (binaryChoiceResult(0.1)) {
      addGirlfriendsOriginBackstory(characters[0], characters[1]);
    } else if (binaryChoiceResult(0.1)) {
      addRivalsOriginBackstory(characters[0], characters[1]);
    } else {
      addFriendsOriginBackstory(characters[0], characters[1]);
    }
    addTownArrivalTeamOriginStory(characters[0], [characters[1]]);
    addTownArrivalTeamOriginStory(characters[1], [characters[0]]);
    addTownArrivalAloneOriginStory(characters[2]);
    addTownArrivalAloneOriginStory(characters[3]);
  } else {
    addTownArrivalAloneOriginStory(characters[0]);
    addTownArrivalAloneOriginStory(characters[1]);
    addTownArrivalAloneOriginStory(characters[2]);
    addTownArrivalAloneOriginStory(characters[3]);
  }
  return characters;
}

/**
 * @param {import("../types").BaseCharacterRole} role
 * @param {number} level
 * @param {import('../types').CharacterBackstory} story
 * @param {string} [name]
 * @returns {import("../types").StoredCharacterData}
 */
export function generateStoredCharacter(role, level, story, name) {
  return {
    id: generateUUID(),
    name: name ?? generateHumanName(),
    role: role,
    story: story,
    level: level,
    partial_exp: 0,
    skills: skills_data[role],
  };
}

/**
 * @param {import("../types").CharacterRole} role
 * @param {number} level
 * @returns {import("../types").PartialInGameStatistics}
 */
export function getStatsByRoleAndLevel(role, level) {
  if (role === 'adventurer') {
    return {
      max_hp: 75 + level * 25,
      max_mp: 49 + level * 1,
      mp_regen: 0.1,
    };
  }
  if (role === 'mage') {
    return {
      max_hp: 50 + level * 10,
      max_mp: 125 + level * 35,
      mp_regen: 0.15,
    };
  }
  if (role === 'rogue') {
    return {
      max_hp: 85 + level * 15,
      max_mp: 49 + level * 1,
      mp_regen: 0.1,
    };
  }
  if (role === 'warrior') {
    return {
      max_hp: 115 + level * 35,
      max_mp: 49 + level * 1,
      mp_regen: 0.1,
    };
  }
  if (role === 'mirror') {
    return {
      max_hp: 1,
      max_mp: 49 + level * 1,
      mp_regen: 0.1,
    };
  }
  if (role === 'statue') {
    return {
      max_hp: 1,
      max_mp: 49 + level * 1,
      mp_regen: 0.1,
    };
  }
  assertTypeIsHandled(role, 'player role');
}

/**
 * @param {import("../types").StoredCharacterData} character
 * @param {import("../types").InGameCharacterEquipment} equipment
 * @returns {import("../types").InGameStatistics}
 */
export function getCharacterStats(character, equipment) {
  const base_stats = getStatsByRoleAndLevel(character.role, character.level);
  /** @type {import('../types').StatsModifierCharacteristics[]} */
  const modifiers = [];
  if (equipment.accessory) modifiers.push(equipment.accessory.stats);
  if (equipment.body) modifiers.push(equipment.body.stats);
  if (equipment.weapon) modifiers.push(equipment.weapon.stats);
  return getInGameStatisticsFromModifiers(base_stats, modifiers);
}

/**
 * @param {import("../types").StoredCharacterData} character
 * @param {import("../types").InGameCharacterEquipment} [equipment]
 * */
export function getImageFromCharacter(character, equipment) {
  if (character.role === 'statue') return 'character_body_statue';
  if (equipment && equipment.weapon) {
    if (equipment.weapon.image === 'item_staff') return 'character_body_staff';
    if (equipment.weapon.image === 'item_sword') return 'character_body_sword';
  }
  return 'character_body_base';
}

/** @param {import("../types").StoredCharacterData} character */
export function getFaceImageFromCharacter(character) {
  if (character.role === 'statue') return 'character_face_statue';
  if (character.role === 'mirror') return 'character_face_mirror_surprised';
  return 'character_face_base_smiling';
}

/**
 * @param {import("../types").StoredCharacterData} character
 * @param {import("../types").InGameCharacterEquipment} equipment
 * @returns {import("../types").InGameCharacterData}
 */
export function generateInGameCharacterFromData(character, equipment) {
  const stats = getCharacterStats(character, equipment);
  return {
    type: 'character',
    base: character,
    image: getImageFromCharacter(character, equipment),
    animation: { current: 'none' },
    base_stats: stats,
    current_stats: { hp: stats.max_hp, mp: stats.max_mp },
    status_effects: [],
    skills: getInGameSkillsFromList(character.skills),
    equipment: equipment,
  };
}

/**
 * @param {import("../types").InGameCharacterData} character
 * @param {number} level
 */
export function changeCharacterLevelTo(character, level) {
  character.base.level = level;
  character.base_stats = getCharacterStats(character.base, character.equipment);
  if (character.current_stats.hp > character.base_stats.max_hp)
    character.current_stats.hp = character.base_stats.max_hp;
  if (character.current_stats.mp > character.base_stats.max_mp)
    character.current_stats.mp = character.base_stats.max_mp;
}

/**
 * @param {import("../types").StoredCharacterData} character
 * @param {import("../types").CharacterRole} role
 */
export function changeStoredCharacterRoleTo(character, role) {
  character.role = role;
  character.skills = skills_data[role];
}

/**
 * @param { import("../types").InGameCharacterData } character
 * @param { import("../types").CharacterRole } role
 */
export function changeCharacterRoleTo(character, role) {
  changeStoredCharacterRoleTo(character.base, role);
  character.skills = getInGameSkillsFromList(character.base.skills);
  character.base_stats = getCharacterStats(character.base, character.equipment);
  if (character.current_stats.hp > character.base_stats.max_hp)
    character.current_stats.hp = character.base_stats.max_hp;
  if (character.current_stats.mp > character.base_stats.max_mp)
    character.current_stats.mp = character.base_stats.max_mp;
}

/** @param {number} level */
export function getRequiredExperienceForPlayerLevel(level) {
  return Math.ceil(Math.pow(level, 1.75) * 100);
}

/**
 * @param {import("../types").StoredCharacterData} character
 * @returns {string}
 */
export function getCharacterDescription(character) {
  let text = `lv. ${character.level} ${character.role}`;
  text += `\n${character.name}`;
  const base_stats = getStatsByRoleAndLevel(character.role, character.level);
  const character_stats = getInGameStatisticsFromModifiers(base_stats, []);
  text += `\nSkills:`;
  for (const skill_id of character.skills) {
    text += `\n- ${getSkillDescriptionForInGameCharacter(character_stats, skill_id)}`;
  }
  text += `\nExp: ${character.partial_exp} / ${getRequiredExperienceForPlayerLevel(character.level)}`;
  return text;
}
