import { assertTypeIsHandled } from '../utils.mjs';
import { getInGameStatisticsFromModifiers } from './stats-modifiers.mjs';

/** @type { Array<import("../types").SkillCharacteristics> } */
const skill_data = [
  {
    id: 'attack_first',
    name: 'Attack',
    target: { type: 'first_enemy' },
    action: { type: 'damage', damage_hp: 15 },
    cost_data: { mana: 0, mana_percent: 0, cooldown: 10 },
  },
  {
    id: 'attack_first_heavy',
    name: 'Heavy Attack',
    target: { type: 'first_enemy' },
    action: { type: 'damage', damage_hp: 60 },
    cost_data: { mana: 0, mana_percent: 0, cooldown: 30 },
  },
  {
    id: 'magic_attack_first',
    name: 'Magic Attack',
    target: { type: 'first_enemy' },
    action: { type: 'damage_scaling_magic_cost', ratio: 1 },
    cost_data: { mana: 0, mana_percent: 0.9, cooldown: 1 },
  },
  {
    id: 'attack_group',
    name: 'Attack Group',
    target: { type: 'all_enemies' },
    action: { type: 'damage', damage_hp: 5 },
    cost_data: { mana: 0, mana_percent: 0, cooldown: 15 * 2 },
  },
  {
    id: 'magic_attack_group',
    name: 'Attack Group',
    target: { type: 'all_enemies' },
    action: { type: 'damage', damage_hp: 10 },
    cost_data: { mana: 30, mana_percent: 0, cooldown: 20 * 2 },
  },
  {
    id: 'heal_self',
    name: 'Heal',
    target: { type: 'self' },
    action: { type: 'heal', heal_hp: 10 },
    cost_data: { mana: 40, mana_percent: 0, cooldown: 4 * 20 },
  },
  {
    id: 'heal_group',
    name: 'Heal Group',
    target: { type: 'all_allies' },
    action: { type: 'heal', heal_hp: 30 },
    cost_data: { mana: 40, mana_percent: 0, cooldown: 4 * 20 },
  },
  {
    id: 'dispel_group',
    name: 'Dispel Group',
    target: { type: 'all_allies' },
    action: { type: 'dispel_all_effects' },
    cost_data: { mana: 100, mana_percent: 0, cooldown: 12 * 20 },
  },
  {
    id: 'resist_petrification',
    name: 'Resist',
    target: { type: 'self' },
    action: { type: 'resist_effect', effect: 'petrification' },
    cost_data: { mana: 45, mana_percent: 0, cooldown: 4 * 20 },
  },
  {
    id: 'resist_mirror_trap',
    name: 'Resist',
    target: { type: 'self' },
    action: { type: 'resist_effect', effect: 'mirror_trap' },
    cost_data: { mana: 45, mana_percent: 0, cooldown: 4 * 20 },
  },
  {
    id: 'struggle_magic_bindings',
    name: 'Struggle',
    target: { type: 'self' },
    action: { type: 'resist_effect', effect: 'magic_bindings' },
    cost_data: { mana: 45, mana_percent: 0, cooldown: 4 * 20 },
  },
  {
    id: 'struggle_rope_bindings',
    name: 'Struggle',
    target: { type: 'self' },
    action: { type: 'resist_effect', effect: 'rope_bindings' },
    cost_data: { mana: 45, mana_percent: 0, cooldown: 4 * 20 },
  },
  {
    id: 'struggle_magic_cage',
    name: 'Struggle',
    target: { type: 'self' },
    action: { type: 'resist_effect', effect: 'magic_cage' },
    cost_data: { mana: 45, mana_percent: 0, cooldown: 4 * 20 },
  },
  {
    id: 'struggle_slime',
    name: 'Struggle',
    target: { type: 'self' },
    action: { type: 'resist_effect', effect: 'slimed' },
    cost_data: { mana: 45, mana_percent: 0, cooldown: 4 * 20 },
  },
];

/**
 * @param { import("../types").SkillId } id
 * @returns { import("../types").SkillCharacteristics }
 */
export function getSkillCharacteristics(id) {
  const data = skill_data.find((s) => s.id === id);
  if (!data) throw new Error(`Could not find characteristics of skill ${id}`);
  return data;
}

/**
 * @param { import("../types").SkillId[] } skills
 * @returns { import("../types").InGameCharacterSkill[] }
 */
export function getInGameSkillsFromList(skills) {
  return skills.map((skill) => ({
    id: skill,
    characteristics: getSkillCharacteristics(skill),
    cooldown: {
      remaining_ticks: 0,
    },
  }));
}

/**
 * @param {import('../types').InGameStatistics} character_stats
 * @param {import("../types").SkillCharacteristics} characteristics
 * @returns {number}
 */
export function getSkillActualManaCost(character_stats, characteristics) {
  let data =
    (characteristics.cost_data.mana + character_stats.max_mp * characteristics.cost_data.mana_percent) *
    character_stats.skill_mp_cost_divider;
  if (data < 0) return 0;
  return data;
}

/**
 * @param {import('../types').InGameStatistics} character_stats
 * @param {import("../types").SkillCharacteristics} characteristics
 * @returns {number}
 */
export function getSkillActualCooldown(character_stats, characteristics) {
  const data = characteristics.cost_data.cooldown / character_stats.skill_cooldown_divider;
  if (data < 0) return 0;
  return data;
}

/**
 * @param {import('../types').InGameStatistics} character_stats
 * @param { import("../types").SkillActionCharacteristics } action
 * @returns {string}
 */
function getSkillActionDescription(character_stats, action) {
  if (action.type === 'damage')
    return `hits the target for ${action.damage_hp + character_stats.attack_additional_hp} HP`;
  if (action.type === 'damage_scaling_magic_cost')
    return `hits the target for ${action.ratio * character_stats.max_mp + character_stats.attack_additional_hp} HP (scales with MP)`;
  if (action.type === 'heal') return `heals the target for ${action.heal_hp + character_stats.heal_additional_hp} HP`;
  if (action.type === 'drain_experience')
    return `drains ${action.damage_experience} exp from the target, absorbing ${
      action.damage_experience * action.drain_ratio
    } of that exp for yourself`;
  if (action.type === 'drain_hp')
    return `drains ${action.damage_hp + character_stats.attack_additional_hp} HP from the target, healing you for ${
      (action.damage_hp + character_stats.attack_additional_hp) * action.drain_ratio
    } HP`;
  if (action.type === 'drain_mana')
    return `drains ${action.damage_mp} MP from the character, absorbing ${
      action.damage_mp * action.drain_ratio
    } MP for yourself`;
  if (action.type === 'dispel_all_effects') return `dispels non-permanent effects`;
  if (action.type === 'apply_effect') return `apply a specific effect`;
  if (action.type === 'dispel_effect') return `dispels a specific effect`;
  if (action.type === 'resist_effect') return `resist the ongoing effect`;
  assertTypeIsHandled(action, 'skill action');
}

/**
 * @param {import('../types').InGameStatistics} character_stats
 * @param {import("../types").SkillId} skill_id
 * @returns {string}
 */
export function getSkillDescriptionForInGameCharacter(character_stats, skill_id) {
  const characteristics = getSkillCharacteristics(skill_id);
  let text = `${characteristics.name}`;
  text += ` - ${getSkillActionDescription(character_stats, characteristics.action)}`;
  text += ` (`;
  const mana_cost = getSkillActualManaCost(character_stats, characteristics);
  if (mana_cost > 0) text += `${Math.ceil(mana_cost)}MP, `;
  const cooldown = getSkillActualCooldown(character_stats, characteristics);
  text += `${Math.floor(cooldown / 2) / 10}s cooldown)`;
  return text;
}
