import { assertTypeIsHandled } from '../utils.mjs';

/**
 * @param {import("../types").StatsModifierCharacteristics} stats
 * @returns {string}
 */
export function getStatsModifierDescription(stats) {
  if (stats.type === 'hp_modifier') {
    if (stats.additional_hp > 0) return `Adds ${stats.additional_hp} HP`;
    else return `Loses ${stats.additional_hp} HP`;
  }
  if (stats.type === 'mp_modifier') {
    if (stats.additional_mp > 0) return `Adds ${stats.additional_mp} MP`;
    else return `Loses ${stats.additional_mp} MP`;
  }
  if (stats.type === 'mp_regen_modifier') {
    if (stats.additional_mp_regen > 0) return `Regens ${stats.additional_mp_regen * 20} MP per second`;
    else return `Loses ${stats.additional_mp_regen * 20} MP per second`;
  }
  if (stats.type === 'damage_modifier') {
    if (stats.additional_damage > 0) return `Adds ${stats.additional_damage} HP of damage to attacks`;
    else return `Adds ${stats.additional_damage} HP of damage to attacks`;
  }
  if (stats.type === 'heal_modifier') {
    if (stats.additional_heal > 0) return `Heals restore ${stats.additional_heal} more HP`;
    else return `Heals restore ${-stats.additional_heal} less HP`;
  }
  if (stats.type === 'mp_cost_modifier') {
    if (stats.mp_cost_divider > 0.0) return `Spells cost ${Math.floor(stats.mp_cost_divider * 100)}% less mana`;
    else return `Spells cost ${Math.floor(-stats.mp_cost_divider * 100)}% more mana`;
  }
  if (stats.type === 'cooldown_modifier') {
    if (stats.cooldown_divider > 0.0) return `Cooldowns are ${Math.floor(stats.cooldown_divider * 100)}% faster`;
    else return `Cooldowns are ${Math.floor(-stats.cooldown_divider * 100)}% slower`;
  }
  if (stats.type === 'counterspell_self_modifier') {
    if (stats.increase_percent > 0)
      return `${Math.floor(stats.increase_percent * 100)}% more chance to cancel spells targeting this character`;
    else return `${-Math.floor(stats.increase_percent * 100)}% less chance to cancel spells targeting this character`;
  }
  if (stats.type === 'counterspell_team_modifier') {
    if (stats.increase_percent > 0)
      return `${Math.floor(stats.increase_percent * 100)}% more chance to cancel spells targeting the team`;
    else return `${-Math.floor(stats.increase_percent * 100)}% less chance to cancel spells targeting the team`;
  }
  if (stats.type === 'evasion_modifier') {
    if (stats.increase_percent > 0)
      return `${Math.floor(stats.increase_percent * 100)}% more chance to evade an attack`;
    else return `${-Math.floor(stats.increase_percent * 100)}% less chance to evade an attack`;
  }
  if (stats.type === 'redirect_attack_to_character_modifier') {
    if (stats.increase_percent > 0)
      return `${Math.floor(stats.increase_percent * 100)}% more chance to redirect an attack to this character`;
    else return `${-Math.floor(stats.increase_percent * 100)}% less chance to redirect an attack to this character`;
  }
  if (stats.type === 'redirected_attack_damage_modifier') {
    if (stats.damage_multiplier > 0)
      return `Take ${Math.floor(100 / stats.damage_multiplier) - 100}% more damage from redirected attacks`;
    else return `Take ${Math.floor(stats.damage_multiplier * 100)}% less damage from redirected attacks`;
  }
  assertTypeIsHandled(stats, 'equipment stats');
}

/**
 * @param {import('../types').PartialInGameStatistics} base_stats
 * @param {import('../types').StatsModifierCharacteristics[]} list
 * @returns {import('../types').InGameStatistics}
 */
export function getInGameStatisticsFromModifiers(base_stats, list) {
  /** @type {import('../types').InGameStatistics} */
  const result = {
    max_hp: base_stats.max_hp !== undefined ? base_stats.max_hp : 0,
    max_mp: base_stats.max_mp !== undefined ? base_stats.max_mp : 0,
    hp_regen: base_stats.hp_regen !== undefined ? base_stats.hp_regen : 0,
    mp_regen: base_stats.mp_regen !== undefined ? base_stats.mp_regen : 0,
    skill_mp_cost_divider: base_stats.skill_mp_cost_divider !== undefined ? base_stats.skill_mp_cost_divider : 1.0,
    skill_cooldown_divider: base_stats.skill_cooldown_divider !== undefined ? base_stats.skill_cooldown_divider : 1.0,
    attack_additional_hp: base_stats.attack_additional_hp !== undefined ? base_stats.attack_additional_hp : 0,
    heal_additional_hp: base_stats.heal_additional_hp !== undefined ? base_stats.heal_additional_hp : 0,
    evasion_chance: base_stats.evasion_chance !== undefined ? base_stats.evasion_chance : 0,
    team_attack_redirect_chance:
      base_stats.team_attack_redirect_chance !== undefined ? base_stats.team_attack_redirect_chance : 0,
    redirected_attack_damage_modifier:
      base_stats.redirected_attack_damage_modifier !== undefined ? base_stats.redirected_attack_damage_modifier : 1.0,
    self_counterspell_chance:
      base_stats.self_counterspell_chance !== undefined ? base_stats.self_counterspell_chance : 0,
    team_counterspell_chance:
      base_stats.team_counterspell_chance !== undefined ? base_stats.team_counterspell_chance : 0,
  };
  for (const entry of list) {
    if (entry.type === 'cooldown_modifier') {
      result.skill_cooldown_divider += entry.cooldown_divider;
      continue;
    }
    if (entry.type === 'counterspell_self_modifier') {
      result.self_counterspell_chance += entry.increase_percent;
      continue;
    }
    if (entry.type === 'counterspell_team_modifier') {
      result.team_counterspell_chance += entry.increase_percent;
      continue;
    }
    if (entry.type === 'damage_modifier') {
      result.attack_additional_hp += entry.additional_damage;
      continue;
    }
    if (entry.type === 'evasion_modifier') {
      result.evasion_chance += entry.increase_percent;
      continue;
    }
    if (entry.type === 'heal_modifier') {
      result.heal_additional_hp += entry.additional_heal;
      continue;
    }
    if (entry.type === 'hp_modifier') {
      result.max_hp += entry.additional_hp;
      continue;
    }
    if (entry.type === 'mp_cost_modifier') {
      result.skill_mp_cost_divider += entry.mp_cost_divider;
      continue;
    }
    if (entry.type === 'mp_modifier') {
      result.max_mp += entry.additional_mp;
      continue;
    }
    if (entry.type === 'mp_regen_modifier') {
      result.mp_regen += entry.additional_mp_regen;
      continue;
    }
    if (entry.type === 'redirect_attack_to_character_modifier') {
      result.team_attack_redirect_chance += entry.increase_percent;
      continue;
    }
    if (entry.type === 'redirected_attack_damage_modifier') {
      result.redirected_attack_damage_modifier += entry.damage_multiplier;
      continue;
    }
    assertTypeIsHandled(entry, 'modifier entry type');
  }
  return result;
}
