/**
 * @typedef { import("./types").GameData } GameData
 * @typedef { import("./types").ControllerSystemMethod } ControllerSystemMethod
 * @typedef { import("./types").SystemMethod } SystemMethod
 */

import { getAudioEngine } from './audio/engine.mjs';

/**
 * Create a new game engine for the given context
 * @param {CanvasRenderingContext2D} renderContext The rendering context
 * @param {import("./types").AudioAssets} audio_assets The sound snippets available in-game
 * @param {GameData} game_data the entities to load
 * @param {ControllerSystemMethod[]} controller_systems The methods to use for controlling the entities
 * @param {SystemMethod[]} render_systems The methods to use for rendering the entities
 */
export function Game(renderContext, audio_assets, game_data, controller_systems, render_systems) {
  const virtualCanvas = document.createElement('canvas');
  virtualCanvas.width = renderContext.canvas.width;
  virtualCanvas.height = renderContext.canvas.height;
  const context = virtualCanvas.getContext('2d');
  if (!context) throw new Error('Could not create context');
  this.shouldQuit = false;
  this.showMetaText = true;
  this.fps = 0;
  this.context = context;
  this.virtualCanvas = virtualCanvas;
  this.renderContext = renderContext;
  /** @type {ReturnType<import('./audio/engine.mjs').getAudioEngine>} */
  this.audio = getAudioEngine(audio_assets);
  /** @type { Array<{ sound: "error" | "click" }> } */ this.sound_queue = [];
  /** @type {ControllerSystemMethod[]} */ this.controller_systems = controller_systems;
  /** @type {SystemMethod[]} */ this.renderer_systems = render_systems;
  /** @type {GameData} */ this.game_data = game_data;
  this.keyboard = {
    config: {
      up: ['z', 'ArrowUp'],
      down: ['s', 'ArrowDown'],
      left: ['q', 'ArrowLeft'],
      right: ['d', 'ArrowRight'],
      action: ['f', 'Spacebar', ' '],
      back: ['Escape'],
    },
    state: {
      up: false,
      down: false,
      left: false,
      right: false,
      action: false,
      back: false,
      shift: false,
    },
  };
  this.mouse = {
    x: -1,
    y: -1,
    primary_button: false,
    primary_button_clicked_frame: false,
    primary_button_held_clicked: false,
    secondary_button: false,
    scroll_up: false,
    scroll_down: false,
  };
  this.renderer = {
    /**
     * @param {ImageBitmap} image
     * @param {number} x
     * @param {number} y
     * @param {number} width
     * @param {number} height
     */
    renderRawImage: function (image, x, y, width, height) {
      context.drawImage(image, x, y, width, height);
    },
  };
  this.audio.set_volume('music', this.game_data.pause.music_volume.scroll_offset);
  this.audio.set_volume('effect', this.game_data.pause.effects_volume.scroll_offset);
}

/** @param {number} elapsedTicks */
Game.prototype.update = function (elapsedTicks) {
  if (this.mouse.primary_button) {
    if (!this.mouse.primary_button_held_clicked) {
      this.mouse.primary_button_held_clicked = true;
      this.mouse.primary_button_clicked_frame = true;
    } else {
      this.mouse.primary_button_clicked_frame = false;
    }
  } else {
    this.mouse.primary_button_clicked_frame = false;
    this.mouse.primary_button_held_clicked = false;
  }

  for (const system of this.controller_systems) {
    system(this, elapsedTicks);
  }
};

/** @param {number} fps */
Game.prototype.render = function (fps) {
  // Update the FPS counter
  this.fps = fps;

  for (const system of this.renderer_systems) {
    system(this);
  }

  // Swap the contexts for double buffering
  this.renderContext.drawImage(this.virtualCanvas, 0, 0);
};

Game.prototype.extraEvents = function () {
  for (const event of this.sound_queue) {
    const snippet = this.audio.play_once(event.sound);
  }

  this.sound_queue = [];
};

Game.prototype.showExitScreen = function () {
  this.audio.stop_all();
  this.context.fillStyle = 'black';
  this.context.strokeStyle = 'transparent';
  this.context.fillRect(0, 0, 800, 600);
  this.context.fillStyle = 'white';
  this.context.textAlign = 'center';
  this.context.font = '40px sans-serif';
  this.context.fillText('Stopped', 400, 320);
  // Swap the contexts for double buffering
  this.renderContext.drawImage(this.virtualCanvas, 0, 0);
};
