import { getRelationsFor } from '../data/backstory.mjs';
import { getEventCharacterInfo } from './add-backstory.mjs';

/**
 * @template T
 * @template {keyof T} K
 * @typedef {T extends any ? Omit<T, K> : never} DistributiveOmit
 */

/**
 * @param {import("../types").GameData} game_data
 * @param {import("../types").StoredCharacterData} capture
 * @param {import("../types").StoredCharacterData[]} captured_group
 */
export function createCapturedBroughtInDungeonBackstory(game_data, capture, captured_group) {
  const girlfriendIsKnown = game_data.game_state.dungeon_knowledge.girlfriends_known.includes(capture.id);
  if (girlfriendIsKnown) {
    const girlfriendStatus = capture.story.relations.find((r) => r.status.is_partner);
    const girlfriendInGroup = captured_group.find((c) => girlfriendStatus && c.id === girlfriendStatus.id);
    const girlfriendInDungeon = game_data.game_state.characters_outside_town.find(
      (c) => girlfriendStatus && c.id === girlfriendStatus.id,
    );

    if (girlfriendInGroup) {
      const nonGirlfriends = captured_group.filter((c) => c !== girlfriendInGroup);
      if (nonGirlfriends.length > 0) {
        addCapturedBackstoryEvent(game_data, capture, {
          type: 'group_capture_start_with_girlfriend',
          capturedGroup: nonGirlfriends.map((c) => getEventCharacterInfo(capture, c.id)),
          girlfriend: getEventCharacterInfo(capture, girlfriendInGroup.id),
        });
      } else {
        addCapturedBackstoryEvent(game_data, capture, {
          type: 'paired_capture_start_with_girlfriend',
          girlfriend: getEventCharacterInfo(capture, girlfriendInGroup.id),
        });
      }
    } else if (girlfriendInDungeon) {
      if (captured_group.length > 0) {
        addCapturedBackstoryEvent(game_data, capture, {
          type: 'group_capture_start_brought_to_girlfriend',
          capturedGroup: captured_group.map((c) => getEventCharacterInfo(capture, c.id)),
          girlfriend: getEventCharacterInfo(capture, girlfriendInDungeon.id),
        });
      } else {
        addCapturedBackstoryEvent(game_data, capture, {
          type: 'alone_capture_start_brought_to_girlfriend',
          girlfriend: getEventCharacterInfo(capture, girlfriendInDungeon.id),
        });
      }
      addCapturedBackstoryEvent(game_data, girlfriendInDungeon, {
        type: 'girlfriend_reunited_after_capture',
        girlfriend: getEventCharacterInfo(girlfriendInDungeon, capture.id),
      });
    } else {
      if (captured_group.length > 0) {
        addCapturedBackstoryEvent(game_data, capture, {
          type: 'group_capture_start',
          capturedGroup: captured_group.map((c) => getEventCharacterInfo(capture, c.id)),
        });
      } else {
        addCapturedBackstoryEvent(game_data, capture, { type: 'alone_capture_start' });
      }
    }
  } else {
    if (captured_group.length > 0) {
      addCapturedBackstoryEvent(game_data, capture, {
        type: 'group_capture_start',
        capturedGroup: captured_group.map((c) => getEventCharacterInfo(capture, c.id)),
      });
    } else {
      addCapturedBackstoryEvent(game_data, capture, { type: 'alone_capture_start' });
    }
  }
}

/**
 * @param {import("../types").GameData} game_data
 * @param {import("../types").StoredCharacterData} character
 * @param {DistributiveOmit<import("../types").CharacterBackstoryCaptureEvent, "cycle">} story
 */
export function addCapturedBackstoryEvent(game_data, character, story) {
  const entry = (() => {
    const last_backstory = character.story.events[character.story.events.length - 1];
    if (last_backstory.location === 'captured') {
      return last_backstory;
    }
    /** @type {import("../types").CharacterBackstoryEvent} */
    const event_group = {
      location: 'captured',
      cycle_start: game_data.game_state.current_cycle,
      content: [],
    };
    character.story.events.push(event_group);
    return event_group;
  })();
  entry.content.push({ ...story, cycle: game_data.game_state.current_cycle });
  if (story.type === 'companion_play_with_event') {
    const relation = getRelationsFor(character, story.captured.character);
    relation.status.is_hardship_companion = true;
    relation.interactions.hardship_interactions++;
  } else if (story.type === 'companion_interrogate_event_stayed_silent') {
    const relation = getRelationsFor(character, story.captured.character);
    relation.status.is_hardship_companion = true;
    relation.interactions.hardship_interactions++;
  } else if (story.type === 'companion_interrogate_event_self_denounced_other_girlfriend') {
    const relation = getRelationsFor(character, story.captured.character);
    relation.interactions.guilt_interactions++;
    const relationGirlfriend = getRelationsFor(character, story.targetGirlfriend.character);
    relationGirlfriend.interactions.guilt_interactions++;
  } else if (story.type === 'companion_magic_event') {
    const relation = getRelationsFor(character, story.captured.character);
    relation.status.is_hardship_companion = true;
    relation.interactions.hardship_interactions++;
  } else if (story.type === 'alone_capture_start_brought_to_girlfriend') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.status.is_hardship_companion = true;
    relation.interactions.hardship_interactions++;
    relation.interactions.romance_interactions++;
  } else if (story.type === 'group_capture_start_brought_to_girlfriend') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.status.is_hardship_companion = true;
    relation.interactions.hardship_interactions++;
    relation.interactions.romance_interactions++;
  } else if (story.type === 'paired_capture_start_with_girlfriend') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.status.is_hardship_companion = true;
    relation.interactions.hardship_interactions++;
    relation.interactions.romance_interactions++;
  } else if (story.type === 'girlfriend_reunited_after_capture') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.status.is_hardship_companion = true;
    relation.interactions.hardship_interactions++;
    relation.interactions.romance_interactions++;
  } else if (story.type === 'girlfriend_reunited_after_talked') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.status.is_hardship_companion = true;
    relation.interactions.hardship_interactions++;
    relation.interactions.romance_interactions++;
  } else if (story.type === 'girlfriend_reunited_after_other_talked') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.status.is_hardship_companion = true;
    relation.interactions.hardship_interactions++;
    relation.interactions.romance_interactions++;
  } else if (story.type === 'girlfriend_wait_event') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.status.is_hardship_companion = true;
    relation.interactions.hardship_interactions++;
    relation.interactions.romance_interactions++;
  } else if (story.type === 'girlfriend_play_with_event') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.status.is_hardship_companion = true;
    relation.interactions.hardship_interactions++;
  } else if (story.type === 'transmutation_prepare_1_in_front_of_girlfriend') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.status.is_hardship_companion = true;
    relation.interactions.hardship_interactions++;
  } else if (story.type === 'transmutation_girlfriend_is_prepared_1') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.status.is_hardship_companion = true;
    relation.interactions.hardship_interactions++;
  } else if (story.type === 'transmutation_prepare_2_in_front_of_girlfriend') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.interactions.hardship_interactions++;
  } else if (story.type === 'transmutation_girlfriend_is_prepared_2') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.interactions.hardship_interactions++;
  } else if (story.type === 'transmutation_prepare_3_in_front_of_girlfriend') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.interactions.hardship_interactions++;
  } else if (story.type === 'transmutation_girlfriend_is_prepared_3') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.interactions.hardship_interactions++;
  } else if (story.type === 'transmutation_girlfriend_is_executed') {
    const relation = getRelationsFor(character, story.girlfriend.character);
    relation.interactions.hardship_interactions++;
  }
}
