import { getRelationsFor } from '../data/backstory.mjs';
import { createCapturedBroughtInDungeonBackstory } from './add-backstory-captured.mjs';
import { addTownBackstoryEvent } from './add-backstory-town.mjs';
import { getEventCharacterInfo } from './add-backstory.mjs';

/**
 * @template T
 * @template {keyof T} K
 * @typedef {T extends any ? Omit<T, K> : never} DistributiveOmit
 */

/**
 * @param {import("../types").GameData} game_data
 * @param {import("../types").StoredCharacterData[]} team
 * @param {import("../types").StoredCharacterData[]} lost_characters
 * @param {"exploration" | "exit"} type
 */
export function addExplorationBackstoryCompanionsLost(game_data, team, lost_characters, type) {
  /** @type {import('../types').EventEnemyInfo[]} */
  const enemy = [{ kind: 'mage' }];
  if (lost_characters.length === 0) return; // This should not be called if there's no lost characters!
  if (team.length === 0) {
    if (lost_characters.length === 1) {
      const capture = lost_characters[0];
      createExplorationCapturedBackstory(game_data, capture, [], [], enemy);
      createCapturedBroughtInDungeonBackstory(game_data, capture, []);
    } else {
      for (const capture of lost_characters) {
        const captured_group = lost_characters.filter((c) => c !== capture);
        createExplorationCapturedBackstory(game_data, capture, [], captured_group, enemy);
        createCapturedBroughtInDungeonBackstory(game_data, capture, captured_group);
      }
    }

    return;
  }
  if (team.length === 1) {
    const sole_member = team[0];
    if (lost_characters.length === 1) {
      const capture = lost_characters[0];
      addExplorationBackstoryEvent(game_data, sole_member, {
        type: type === 'exit' ? 'alone_has_lost_companion_during_exit' : 'alone_has_lost_companion_during_exploration',
        enemies: enemy,
        captured: getEventCharacterInfo(sole_member, capture.id),
      });
      createExplorationCapturedBackstory(game_data, capture, [sole_member], [], enemy);
      createCapturedBroughtInDungeonBackstory(game_data, capture, []);
      return;
    } else {
      addExplorationBackstoryEvent(game_data, sole_member, {
        type:
          type === 'exit' ? 'alone_has_lost_companions_during_exit' : 'alone_has_lost_companions_during_exploration',
        enemies: enemy,
        capturedGroup: lost_characters.map((c) => getEventCharacterInfo(sole_member, c.id)),
      });
      for (const capture of lost_characters) {
        const others = lost_characters.filter((c) => c !== capture);
        createExplorationCapturedBackstory(game_data, capture, [sole_member], others, enemy);
        createCapturedBroughtInDungeonBackstory(game_data, capture, others);
      }
    }
    return;
  }

  // There's a team of survivors
  if (lost_characters.length === 1) {
    const capture = lost_characters[0];
    for (const teammate of team) {
      addExplorationBackstoryEvent(game_data, teammate, {
        type: type === 'exit' ? 'team_has_lost_companion_during_exit' : 'team_has_lost_companion_during_exploration',
        enemies: enemy,
        group: team.filter((t) => t !== teammate).map((t) => getEventCharacterInfo(teammate, t.id)),
        captured: getEventCharacterInfo(teammate, capture.id),
      });
    }
    createExplorationCapturedBackstory(game_data, capture, team, [], enemy);
    createCapturedBroughtInDungeonBackstory(game_data, capture, []);
    return;
  }
  for (const teammate of team) {
    addExplorationBackstoryEvent(game_data, teammate, {
      type: type === 'exit' ? 'team_has_lost_companions_during_exit' : 'team_has_lost_companions_during_exploration',
      enemies: enemy,
      group: team.filter((t) => t !== teammate).map((t) => getEventCharacterInfo(teammate, t.id)),
      capturedGroup: lost_characters.map((c) => getEventCharacterInfo(teammate, c.id)),
    });
  }
  for (const capture of lost_characters) {
    createExplorationCapturedBackstory(game_data, capture, team, [], enemy);
    createCapturedBroughtInDungeonBackstory(game_data, capture, []);
  }
}

/**
 * @param {import("../types").GameData} game_data
 * @param {import("../types").StoredCharacterData} capture
 * @param {import("../types").StoredCharacterData[]} still_free_group
 * @param {import("../types").StoredCharacterData[]} captured_group
 * @param {import("../types").EventEnemyInfo[]} enemies
 */
function createExplorationCapturedBackstory(game_data, capture, still_free_group, captured_group, enemies) {
  if (still_free_group.length === 0) {
    if (captured_group.length === 0) {
      addExplorationBackstoryEvent(game_data, capture, {
        type: 'alone_becomes_captured',
        enemies: enemies,
      });
    } else {
      addExplorationBackstoryEvent(game_data, capture, {
        type: 'alone_becomes_captured_with_companions',
        capturedGroup: captured_group.map((c) => getEventCharacterInfo(capture, c.id)),
        enemies: enemies,
      });
    }
  } else {
    if (captured_group.length === 0) {
      addExplorationBackstoryEvent(game_data, capture, {
        type: 'team_becomes_captured_alone',
        group: still_free_group.map((g) => getEventCharacterInfo(capture, g.id)),
        enemies: enemies,
      });
    } else {
      addExplorationBackstoryEvent(game_data, capture, {
        type: 'team_becomes_captured_with_companions',
        enemies: enemies,
        group: still_free_group.map((g) => getEventCharacterInfo(capture, g.id)),
        capturedGroup: captured_group.map((c) => getEventCharacterInfo(capture, c.id)),
      });
    }
  }
}

/**
 * @param {import("../types").GameData} game_data
 * @param {import("../types").StoredCharacterData[]} team
 * @param {import("../types").StoredCharacterData[]} saved_characters
 */
export function addExplorationBackstoryBackToTown(game_data, team, saved_characters) {
  if (saved_characters.length === 0) {
    if (team.length === 1) {
      addTownBackstoryEvent(game_data, team[0], { type: 'arrival_alone_dungeon' });
    } else {
      for (const character of team) {
        const group = team.filter((t) => t !== character).map((t) => getEventCharacterInfo(character, t.id));
        addTownBackstoryEvent(game_data, character, { type: 'arrival_team_dungeon', group: group });
      }
    }
    return;
  }

  if (team.length === 1) {
    const sole_rescuer = team[0];
    if (saved_characters.length === 1) {
      addExplorationBackstoryEvent(game_data, sole_rescuer, {
        type: 'alone_has_saved_companion',
        rescuee: getEventCharacterInfo(sole_rescuer, saved_characters[0].id),
      });
      addTownBackstoryEvent(game_data, sole_rescuer, {
        type: 'arrival_alone_rescued_people',
        rescuees: saved_characters.map((s) => getEventCharacterInfo(sole_rescuer, s.id)),
      });
      addTownBackstoryEvent(game_data, saved_characters[0], {
        type: 'arrival_team_was_rescued_alone',
        group: team.map((r) => getEventCharacterInfo(saved_characters[0], r.id)),
      });
      return;
    }
    addExplorationBackstoryEvent(game_data, sole_rescuer, {
      type: 'alone_has_saved_companions',
      rescuees: saved_characters.map((s) => getEventCharacterInfo(sole_rescuer, s.id)),
    });
    addTownBackstoryEvent(game_data, sole_rescuer, {
      type: 'arrival_alone_rescued_people',
      rescuees: saved_characters.map((s) => getEventCharacterInfo(sole_rescuer, s.id)),
    });
    for (const saved_character of saved_characters) {
      addTownBackstoryEvent(game_data, saved_character, {
        type: 'arrival_team_was_rescued_group',
        group: team.map((r) => getEventCharacterInfo(saved_character, r.id)),
        rescuees: saved_characters
          .filter((s) => s !== saved_character)
          .map((s) => getEventCharacterInfo(saved_character, s.id)),
      });
    }
    return;
  }
  if (saved_characters.length === 1) {
    for (const rescuer of team) {
      addExplorationBackstoryEvent(game_data, rescuer, {
        type: 'team_has_saved_companion',
        group: team.filter((r) => r !== rescuer).map((r) => getEventCharacterInfo(rescuer, r.id)),
        rescuee: getEventCharacterInfo(rescuer, saved_characters[0].id),
      });
      addTownBackstoryEvent(game_data, rescuer, {
        type: 'arrival_team_rescued_people',
        group: team.filter((r) => r !== rescuer).map((r) => getEventCharacterInfo(rescuer, r.id)),
        rescuees: saved_characters.map((s) => getEventCharacterInfo(rescuer, s.id)),
      });
    }
    addTownBackstoryEvent(game_data, saved_characters[0], {
      type: 'arrival_team_was_rescued_alone',
      group: team.map((r) => getEventCharacterInfo(saved_characters[0], r.id)),
    });
    return;
  }

  for (const rescuer of team) {
    addExplorationBackstoryEvent(game_data, rescuer, {
      type: 'team_has_saved_companions',
      group: team.filter((r) => r !== rescuer).map((r) => getEventCharacterInfo(rescuer, r.id)),
      rescuees: saved_characters.map((s) => getEventCharacterInfo(rescuer, s.id)),
    });
    addTownBackstoryEvent(game_data, rescuer, {
      type: 'arrival_team_rescued_people',
      group: team.filter((r) => r !== rescuer).map((r) => getEventCharacterInfo(rescuer, r.id)),
      rescuees: saved_characters.map((s) => getEventCharacterInfo(rescuer, s.id)),
    });
  }
  for (const saved_character of saved_characters) {
    addTownBackstoryEvent(game_data, saved_character, {
      type: 'arrival_team_was_rescued_group',
      group: team.map((r) => getEventCharacterInfo(saved_character, r.id)),
      rescuees: saved_characters
        .filter((s) => s !== saved_character)
        .map((s) => getEventCharacterInfo(saved_character, s.id)),
    });
  }
}

/**
 * @param {import("../types").GameData} game_data
 * @param {import("../types").StoredCharacterData} character
 * @param {DistributiveOmit<import("../types").CharacterBackstoryExplorationEvent, "cycle">} story
 */
export function addExplorationBackstoryEvent(game_data, character, story) {
  const entry = (() => {
    const last_backstory = character.story.events[character.story.events.length - 1];
    if (last_backstory.location === 'exploration') {
      return last_backstory;
    }
    /** @type {import("../types").CharacterBackstoryEvent} */
    const event_group = {
      location: 'exploration',
      cycle_start: game_data.game_state.current_cycle,
      content: [],
    };
    character.story.events.push(event_group);
    return event_group;
  })();
  entry.content.push({ ...story, cycle: game_data.game_state.current_cycle });
  if (story.type === 'team_loot') {
    for (const teammate of story.group) {
      const relation = getRelationsFor(character, teammate.character);
      relation.status.is_team_member = true;
    }
  } else if (story.type === 'alone_rescue') {
    const relation = getRelationsFor(character, story.rescuee.character);
    relation.interactions.did_protect_interactions++;
  } else if (story.type === 'team_rescue') {
    const relation = getRelationsFor(character, story.rescuee.character);
    relation.interactions.did_protect_interactions++;
  } else if (story.type === 'alone_becomes_rescued') {
    const relation = getRelationsFor(character, story.rescuer.character);
    relation.interactions.protected_by_interactions++;
  } else if (story.type === 'team_becomes_rescued') {
    for (const teammate of story.group) {
      const relation = getRelationsFor(character, teammate.character);
      relation.interactions.protected_by_interactions++;
    }
  } else if (story.type === 'alone_has_lost_companion_during_exploration') {
    const relation = getRelationsFor(character, story.captured.character);
    relation.interactions.guilt_interactions++;
  } else if (story.type === 'alone_has_lost_companion_during_exit') {
    const relation = getRelationsFor(character, story.captured.character);
    relation.interactions.guilt_interactions++;
  } else if (story.type === 'alone_has_lost_companions_during_exploration') {
    for (const capture of story.capturedGroup) {
      const relation = getRelationsFor(character, capture.character);
      relation.interactions.guilt_interactions++;
    }
  } else if (story.type === 'alone_has_lost_companions_during_exit') {
    for (const capture of story.capturedGroup) {
      const relation = getRelationsFor(character, capture.character);
      relation.interactions.guilt_interactions++;
    }
  } else if (story.type === 'team_has_lost_companion_during_exploration') {
    const relation = getRelationsFor(character, story.captured.character);
    relation.interactions.guilt_interactions++;
  } else if (story.type === 'team_has_lost_companion_during_exit') {
    const relation = getRelationsFor(character, story.captured.character);
    relation.interactions.guilt_interactions++;
  } else if (story.type === 'team_has_lost_companions_during_exploration') {
    for (const capture of story.capturedGroup) {
      const relation = getRelationsFor(character, capture.character);
      relation.interactions.guilt_interactions++;
    }
  } else if (story.type === 'team_has_lost_companions_during_exit') {
    for (const capture of story.capturedGroup) {
      const relation = getRelationsFor(character, capture.character);
      relation.interactions.guilt_interactions++;
    }
  } else if (story.type === 'alone_becomes_captured_with_companions') {
    for (const capture of story.capturedGroup) {
      const relation = getRelationsFor(character, capture.character);
      relation.interactions.hardship_interactions++;
    }
  } else if (story.type === 'team_becomes_captured_with_companions') {
    for (const capture of story.capturedGroup) {
      const relation = getRelationsFor(character, capture.character);
      relation.interactions.hardship_interactions++;
    }
  }
}
