import { getRelationsFor } from '../data/backstory.mjs';
import { drainLevelIntoEquipment, getEquipmentFromCharacter } from '../data/equipment.mjs';
import { getAllLiveCharacters } from '../data/utils.mjs';
import { binaryChoiceResult, getRandomInArray, getRandomInWeightedArray } from '../random.mjs';
import { addCapturedBackstoryEvent } from './add-backstory-captured.mjs';
import { addTransmutedBackstoryEvent } from './add-backstory-transmuted.mjs';
import { getEventCharacterInfo } from './add-backstory.mjs';

const CAPTURE_EVENT_DUAL_INTERROGATION_CHANCE = 0.2;
const CAPTURE_EVENT_INTERROGATION_CHANCE = 0.5;
const CAPTURE_EVENT_INTERROGATION_SUCCESS_CHANCE = 0.2;
const CAPTURE_EVENT_DUAL_INTERROGATION_SUCCESS_CHANCE = 0.4;
const CAPTURE_EVENT_INTERROGATION_SUCCESS_GIRLFRIEND_CHANCE = 0.9;
const CAPTURE_EVENT_CHANCE = 0.2;
const CAPTURE_DUAL_EVENT_CHANCE = 0.2;

const DRAIN_GIRLFRIEND_EXP_RATIO = 0.65;
const DRAIN_STANDARD_EXP_RATIO = 0.95;

/**
 * @typedef {[import('../types').StoredCharacterData] | [import('../types').StoredCharacterData, import('../types').StoredCharacterData]} InterrogationGroup
 * @typedef {[import('../types').StoredCharacterData, import('../types').StoredCharacterData]} PairedGirlfriends
 * @typedef {{ type: "single", c: import('../types').StoredCharacterData, isValidTransmuteTarget: boolean }
 *   | { type: "dual", c1: import('../types').StoredCharacterData, c2: import('../types').StoredCharacterData }
 * } DungeonRandomEvents
 */

/**
 * @param {import("../types").GameData} game_data The current game
 */
export function handleOutsideCharacterUpdate(game_data) {
  const characters_outside = game_data.game_state.characters_outside_town.length;
  if (characters_outside === 0) return;
  const items_outside_count = game_data.game_state.items_outside_town.length;

  const sorted_characters = sort_outside_town_characters(game_data);

  // Transform outside characters into items with a percent chance for each character
  // Note: The percent chance depends of how many items are outside the town,
  // going from 25% if there's no items and two characters, to 0% if there's two times more items than characters.
  const transformation_chance_ratio = Math.max(0, Math.min(0.25, (2 * characters_outside - items_outside_count) / 16));

  createCharacterInterrogationBackstories(game_data, sorted_characters.interrogation_groups);
  createLoneCharacterTransmutationBackstories(
    game_data,
    sorted_characters.interrogated_non_girlfriended_characters,
    transformation_chance_ratio,
  );
  createCharacterPairedGirlfriendsBackstories(
    game_data,
    sorted_characters.paired_girlfriend_characters,
    transformation_chance_ratio,
  );

  createCharacterForceDrainIntoBackstories(game_data, sorted_characters.force_drain_into_characters);

  // Apply capture events to the remaining characters
  const remaining_character_with_events = getCharactersForRemainingEvents(game_data, sorted_characters);
  createCharacterRandomDungeonEventBackstories(game_data, remaining_character_with_events, transformation_chance_ratio);
}

/**
 * @param {import("../types").GameData} game_data The current game
 * @returns {{
 *   force_drain_into_characters: ReadonlyArray<{ character: import('../types').StoredCharacterData, item: import('../types').StoredEquipmentData }>,
 *   not_interrogated_characters: ReadonlyArray<import('../types').StoredCharacterData>,
 *   interrogation_groups: ReadonlyArray<InterrogationGroup>,
 *   interrogated_non_girlfriended_characters: ReadonlyArray<import('../types').StoredCharacterData>,
 *   interrogated_girlfriend_unavailable_characters: ReadonlyArray<import('../types').StoredCharacterData>,
 *   paired_girlfriend_characters: ReadonlyArray<PairedGirlfriends>,
 * }}
 */
function sort_outside_town_characters(game_data) {
  /** @type {Array<{ character: import('../types').StoredCharacterData, item: import('../types').StoredEquipmentData }>} */
  const force_drain_into_characters = [];
  /** @type {Array<import('../types').StoredCharacterData>} */
  const known_girlfriend_characters = [];
  /** @type {Array<import('../types').StoredCharacterData>} */
  let not_interrogated_characters = [];
  /** @type {Array<PairedGirlfriends>} */
  let paired_girlfriend_characters = [];
  /** @type {Array<import('../types').StoredCharacterData>} */
  const interrogated_girlfriend_unavailable_characters = [];
  /** @type {Array<import('../types').StoredCharacterData>} */
  const interrogated_non_girlfriended_characters = [];
  // Interrogate characters that are not known yet
  /** @type {Array<InterrogationGroup>} */
  const interrogation_groups = [];

  for (const character of game_data.game_state.characters_outside_town) {
    const force_drain = game_data.game_state.dungeon_knowledge.forced_drain_into.find(
      (d) => d.character === character.id,
    );
    if (force_drain) {
      const item = game_data.game_state.items_outside_town.find(
        (i) => i.character && i.character.data.id === force_drain.item_character,
      );
      if (item) {
        force_drain_into_characters.push({ character, item });
        continue;
      } else {
        game_data.game_state.dungeon_knowledge.forced_drain_into =
          game_data.game_state.dungeon_knowledge.forced_drain_into.filter((i) => i.character !== character.id);
      }
    }
    if (game_data.game_state.dungeon_knowledge.girlfriends_known.includes(character.id)) {
      known_girlfriend_characters.push(character);
      continue;
    }
    if (game_data.game_state.dungeon_knowledge.interrogations_completed.includes(character.id)) {
      interrogated_non_girlfriended_characters.push(character);
      continue;
    }
    not_interrogated_characters.push(character);
  }

  for (const character of known_girlfriend_characters) {
    if (paired_girlfriend_characters.some((c) => c[0] === character || c[1] === character)) continue;
    const girlfriendRelation = character.story.relations.find((e) => e.status.is_partner);
    const girlfriend = known_girlfriend_characters.find((c) => girlfriendRelation && c.id === girlfriendRelation.id);
    if (girlfriend) {
      paired_girlfriend_characters.push([character, girlfriend]);
    } else {
      interrogated_girlfriend_unavailable_characters.push(character);
    }
  }

  for (const character of not_interrogated_characters) {
    if (interrogation_groups.some((g) => g[0] === character || g[1] === character)) continue;
    const hasGirlfriend = character.story.relations.some((e) => e.status.is_partner);
    if (hasGirlfriend && binaryChoiceResult(CAPTURE_EVENT_DUAL_INTERROGATION_CHANCE)) {
      const others = not_interrogated_characters.filter(
        (c) =>
          c !== character &&
          !interrogation_groups.some((g) => g[0] === character || g[1] === character) &&
          !c.story.relations.some((e) => e.status.is_partner),
      );
      if (others.length > 0) {
        const target = getRandomInArray(others);
        interrogation_groups.push([character, target]);
        not_interrogated_characters = not_interrogated_characters.filter((c) => c !== character);
        not_interrogated_characters = not_interrogated_characters.filter((c) => c !== target);
        continue;
      }
    }
    if (binaryChoiceResult(CAPTURE_EVENT_INTERROGATION_CHANCE)) {
      interrogation_groups.push([character]);
      not_interrogated_characters = not_interrogated_characters.filter((c) => c !== character);
    }
  }
  return {
    force_drain_into_characters,
    not_interrogated_characters,
    interrogation_groups,
    interrogated_non_girlfriended_characters,
    interrogated_girlfriend_unavailable_characters,
    paired_girlfriend_characters,
  };
}

/**
 * @param {import("../types").GameData} game_data The current game
 * @param {{
 *   not_interrogated_characters: ReadonlyArray<import('../types').StoredCharacterData>,
 *   interrogated_non_girlfriended_characters: ReadonlyArray<import('../types').StoredCharacterData>,
 *   interrogated_girlfriend_unavailable_characters: ReadonlyArray<import('../types').StoredCharacterData>,
 *   interrogation_groups: ReadonlyArray<InterrogationGroup>,
 * }} sorted_characters
 * @returns {ReadonlyArray<DungeonRandomEvents>}
 */
function getCharactersForRemainingEvents(game_data, sorted_characters) {
  /** @type {Array<{ character: import('../types').StoredCharacterData, isValidTransmuteTarget: boolean }>} */
  const alone_characters_for_events = [];
  for (const character of sorted_characters.not_interrogated_characters) {
    if (sorted_characters.interrogation_groups.some((c) => c[0] === character || c[1] === character)) continue;
    alone_characters_for_events.push({ character, isValidTransmuteTarget: false });
  }
  for (const character of sorted_characters.interrogated_non_girlfriended_characters) {
    if (!game_data.game_state.characters_outside_town.includes(character)) continue;
    alone_characters_for_events.push({ character, isValidTransmuteTarget: true });
  }
  for (const character of sorted_characters.interrogated_girlfriend_unavailable_characters) {
    alone_characters_for_events.push({ character, isValidTransmuteTarget: false });
  }

  /** @type {Array<DungeonRandomEvents>} */
  const alone_character_events_to_execute = [];
  for (const c of alone_characters_for_events) {
    if (
      alone_character_events_to_execute.some(
        (g) =>
          (g.type === 'single' && g.c === c.character) ||
          (g.type === 'dual' && (g.c1 === c.character || g.c2 === c.character)),
      )
    )
      continue;
    if (binaryChoiceResult(CAPTURE_DUAL_EVENT_CHANCE)) {
      const others = alone_characters_for_events.filter(
        (o) =>
          o !== c &&
          !alone_character_events_to_execute.some(
            (g) =>
              (g.type === 'single' && g.c === o.character) ||
              (g.type === 'dual' && (g.c1 === o.character || g.c2 === o.character)),
          ),
      );
      if (others.length > 0) {
        const target = getRandomInArray(others);
        alone_character_events_to_execute.push({ type: 'dual', c1: c.character, c2: target.character });
        continue;
      }
    }
    if (binaryChoiceResult(CAPTURE_EVENT_CHANCE)) {
      alone_character_events_to_execute.push({
        type: 'single',
        c: c.character,
        isValidTransmuteTarget: c.isValidTransmuteTarget,
      });
    }
  }
  return alone_character_events_to_execute;
}

/**
 * @param {import("../types").GameData} game_data The current game
 * @param {ReadonlyArray<InterrogationGroup>} interrogation_groups
 */
function createCharacterInterrogationBackstories(game_data, interrogation_groups) {
  for (const group of interrogation_groups) {
    if (group.length === 2) {
      const character1 = group[0];
      const character2 = group[1];
      if (game_data.game_state.dungeon_knowledge.interrogations_completed.includes(character1.id)) continue;
      if (game_data.game_state.dungeon_knowledge.interrogations_completed.includes(character2.id)) continue;
      const girlfriendRelation = character1.story.relations.find((e) => e.status.is_partner);
      const girlfriend = getAllLiveCharacters(game_data).find(
        (c) => girlfriendRelation && c.id === girlfriendRelation.id,
      );
      if (girlfriend === character2 && binaryChoiceResult(CAPTURE_EVENT_INTERROGATION_SUCCESS_GIRLFRIEND_CHANCE)) {
        addCapturedBackstoryEvent(game_data, character1, {
          type: 'girlfriend_interrogate_event_broke',
          girlfriend: getEventCharacterInfo(character1, character2.id),
        });
        addCapturedBackstoryEvent(game_data, character2, {
          type: 'girlfriend_interrogate_event_broke',
          girlfriend: getEventCharacterInfo(character2, character1.id),
        });
        game_data.game_state.dungeon_knowledge.interrogations_completed.push(character1.id);
        game_data.game_state.dungeon_knowledge.interrogations_completed.push(character2.id);
        game_data.game_state.dungeon_knowledge.girlfriends_known.push(character1.id);
        game_data.game_state.dungeon_knowledge.girlfriends_known.push(character2.id);
      } else if (girlfriend === character2) {
        addCapturedBackstoryEvent(game_data, character1, {
          type: 'girlfriend_interrogate_event_stayed_silent',
          girlfriend: getEventCharacterInfo(character1, character2.id),
        });
        addCapturedBackstoryEvent(game_data, character2, {
          type: 'girlfriend_interrogate_event_stayed_silent',
          girlfriend: getEventCharacterInfo(character2, character1.id),
        });
      } else if (girlfriend && binaryChoiceResult(CAPTURE_EVENT_DUAL_INTERROGATION_SUCCESS_CHANCE)) {
        if (binaryChoiceResult(0.5)) {
          addCapturedBackstoryEvent(game_data, character1, {
            type: 'companion_interrogate_event_self_denounced_own_girlfriend',
            captured: getEventCharacterInfo(character1, character2.id),
            girlfriend: getEventCharacterInfo(character1, girlfriend.id),
          });
          addCapturedBackstoryEvent(game_data, character2, {
            type: 'companion_interrogate_event_other_denounced_other_girlfriend',
            captured: getEventCharacterInfo(character2, character1.id),
            targetGirlfriend: getEventCharacterInfo(character2, girlfriend.id),
          });
          if (game_data.game_state.characters_outside_town.some((c) => c.id === girlfriend.id)) {
            addCapturedBackstoryEvent(game_data, character1, {
              type: 'girlfriend_reunited_after_talked',
              girlfriend: getEventCharacterInfo(character1, girlfriend.id),
            });
            addCapturedBackstoryEvent(game_data, girlfriend, {
              type: 'girlfriend_reunited_after_other_talked',
              girlfriend: getEventCharacterInfo(girlfriend, character1.id),
            });
          }
        } else {
          addCapturedBackstoryEvent(game_data, character1, {
            type: 'companion_interrogate_event_other_denounced_self_girlfriend',
            captured: getEventCharacterInfo(character1, character2.id),
            girlfriend: getEventCharacterInfo(character1, girlfriend.id),
          });
          addCapturedBackstoryEvent(game_data, character2, {
            type: 'companion_interrogate_event_self_denounced_other_girlfriend',
            captured: getEventCharacterInfo(character2, character1.id),
            targetGirlfriend: getEventCharacterInfo(character2, girlfriend.id),
          });
          if (game_data.game_state.characters_outside_town.some((c) => c.id === girlfriend.id)) {
            addCapturedBackstoryEvent(game_data, character1, {
              type: 'girlfriend_reunited_after_talked',
              girlfriend: getEventCharacterInfo(character1, girlfriend.id),
            });
            addCapturedBackstoryEvent(game_data, girlfriend, {
              type: 'girlfriend_reunited_after_other_talked',
              girlfriend: getEventCharacterInfo(girlfriend, character1.id),
            });
          }
        }
        game_data.game_state.dungeon_knowledge.interrogations_completed.push(character1.id);
        game_data.game_state.dungeon_knowledge.interrogations_completed.push(character2.id);
        game_data.game_state.dungeon_knowledge.interrogations_completed.push(girlfriend.id);
        game_data.game_state.dungeon_knowledge.girlfriends_known.push(character1.id);
        game_data.game_state.dungeon_knowledge.girlfriends_known.push(girlfriend.id);
      } else {
        addCapturedBackstoryEvent(game_data, character1, {
          type: 'companion_interrogate_event_stayed_silent',
          captured: getEventCharacterInfo(character1, character2.id),
        });
        addCapturedBackstoryEvent(game_data, character2, {
          type: 'companion_interrogate_event_stayed_silent',
          captured: getEventCharacterInfo(character2, character1.id),
        });
      }
    } else {
      const character = group[0];
      if (game_data.game_state.dungeon_knowledge.interrogations_completed.includes(character.id)) continue;
      if (binaryChoiceResult(CAPTURE_EVENT_INTERROGATION_SUCCESS_CHANCE)) {
        const girlfriendRelation = character.story.relations.find((e) => e.status.is_partner);
        const girlfriend = getAllLiveCharacters(game_data).find(
          (c) => girlfriendRelation && c.id === girlfriendRelation.id,
        );
        if (girlfriend) {
          addCapturedBackstoryEvent(game_data, character, {
            type: 'alone_interrogate_event_denounced_girlfriend',
            girlfriend: getEventCharacterInfo(character, girlfriend.id),
          });
          game_data.game_state.dungeon_knowledge.girlfriends_known.push(character.id);
          game_data.game_state.dungeon_knowledge.girlfriends_known.push(girlfriend.id);
          game_data.game_state.dungeon_knowledge.interrogations_completed.push(girlfriend.id);
          if (game_data.game_state.characters_outside_town.some((c) => c.id === girlfriend.id)) {
            addCapturedBackstoryEvent(game_data, character, {
              type: 'girlfriend_reunited_after_talked',
              girlfriend: getEventCharacterInfo(character, girlfriend.id),
            });
            addCapturedBackstoryEvent(game_data, girlfriend, {
              type: 'girlfriend_reunited_after_other_talked',
              girlfriend: getEventCharacterInfo(girlfriend, character.id),
            });
          }
        } else {
          addCapturedBackstoryEvent(game_data, character, {
            type: 'alone_interrogate_event_talked',
          });
        }
        game_data.game_state.dungeon_knowledge.interrogations_completed.push(character.id);
      } else {
        addCapturedBackstoryEvent(game_data, character, {
          type: 'alone_interrogate_event_stayed_silent',
        });
      }
    }
  }
}

/**
 * @param {import("../types").GameData} game_data The current game
 * @param {ReadonlyArray<import('../types').StoredCharacterData>} possible_characters
 * @param {number} transformation_chance_ratio
 */
function createLoneCharacterTransmutationBackstories(game_data, possible_characters, transformation_chance_ratio) {
  const targetable_lone_characters = game_data.game_state.dungeon_knowledge.transmutation_states
    .filter((s) => s.step === 3)
    .map((s) => possible_characters.find((c) => c.id === s.id))
    .filter((c) => !!c);

  for (const character of targetable_lone_characters) {
    if (!binaryChoiceResult(transformation_chance_ratio)) continue;
    const equipment = getEquipmentFromCharacter(character, DRAIN_STANDARD_EXP_RATIO);
    game_data.game_state.items_outside_town.push(equipment);
    game_data.game_state.characters_outside_town = game_data.game_state.characters_outside_town.filter(
      (c) => c !== character,
    );
    game_data.game_state.dungeon_knowledge.transmutation_states =
      game_data.game_state.dungeon_knowledge.transmutation_states.filter((c) => c.id !== character.id);
    addTransmutedBackstoryEvent(game_data, character, {
      type: 'transmutation_execute_alone',
      equipment: { name: equipment.name, level: equipment.level },
    });
  }
}

/**
 * @param {import("../types").GameData} game_data The current game
 * @param {ReadonlyArray<PairedGirlfriends>} paired_girlfriends
 * @param {number} transformation_chance_ratio
 */
function createCharacterPairedGirlfriendsBackstories(game_data, paired_girlfriends, transformation_chance_ratio) {
  for (const pair of paired_girlfriends) {
    const character1 = pair[0];
    const character2 = pair[1];
    const c1State = game_data.game_state.dungeon_knowledge.transmutation_states.find((c) => c.id === character1.id);
    const c2State = game_data.game_state.dungeon_knowledge.transmutation_states.find((c) => c.id === character2.id);
    if (binaryChoiceResult(transformation_chance_ratio)) {
      let target = character1;
      let targetState = c1State;
      let witness = character2;
      /** @type { 0 | 1 | 2 | 3 } */
      let step = 0;
      if (c1State) {
        target = character1;
        witness = character2;
        step = c1State.step;
        targetState = c1State;
      } else if (c2State) {
        target = character2;
        witness = character1;
        step = c2State.step;
        targetState = c2State;
      } else {
        if (character1.level > character2.level) {
          target = character1;
          witness = character2;
          step = 0;
        } else {
          target = character2;
          witness = character1;
          step = 0;
        }
      }
      if (step === 0) {
        addCapturedBackstoryEvent(game_data, target, {
          type: 'transmutation_prepare_1_in_front_of_girlfriend',
          girlfriend: getEventCharacterInfo(target, witness.id),
        });
        addCapturedBackstoryEvent(game_data, witness, {
          type: 'transmutation_girlfriend_is_prepared_1',
          girlfriend: getEventCharacterInfo(witness, target.id),
        });
        game_data.game_state.dungeon_knowledge.transmutation_states.push({ id: target.id, step: 1 });
        continue;
      } else if (step === 1) {
        addCapturedBackstoryEvent(game_data, target, {
          type: 'transmutation_prepare_2_in_front_of_girlfriend',
          girlfriend: getEventCharacterInfo(target, witness.id),
        });
        addCapturedBackstoryEvent(game_data, witness, {
          type: 'transmutation_girlfriend_is_prepared_2',
          girlfriend: getEventCharacterInfo(witness, target.id),
        });
        if (targetState) targetState.step = 2;
        continue;
      } else if (step === 2) {
        addCapturedBackstoryEvent(game_data, target, {
          type: 'transmutation_prepare_3_in_front_of_girlfriend',
          girlfriend: getEventCharacterInfo(target, witness.id),
        });
        addCapturedBackstoryEvent(game_data, witness, {
          type: 'transmutation_girlfriend_is_prepared_3',
          girlfriend: getEventCharacterInfo(witness, target.id),
        });
        if (targetState) targetState.step = 3;
        continue;
      } else if (step === 3) {
        const equipment = getEquipmentFromCharacter(target, 0.75);
        drainLevelIntoEquipment(equipment, witness, 0.75);
        game_data.game_state.items_outside_town.push(equipment);
        game_data.game_state.characters_outside_town = game_data.game_state.characters_outside_town.filter(
          (c) => c !== target,
        );
        game_data.game_state.dungeon_knowledge.girlfriends_known =
          game_data.game_state.dungeon_knowledge.girlfriends_known.filter((c) => c !== target.id);
        game_data.game_state.dungeon_knowledge.girlfriends_known =
          game_data.game_state.dungeon_knowledge.girlfriends_known.filter((c) => c !== witness.id);
        game_data.game_state.dungeon_knowledge.transmutation_states =
          game_data.game_state.dungeon_knowledge.transmutation_states.filter((c) => c.id !== target.id);
        game_data.game_state.dungeon_knowledge.forced_drain_into.push({
          character: witness.id,
          item_character: target.id,
        });
        addTransmutedBackstoryEvent(game_data, target, {
          type: 'transmutation_execute_in_front_of_girlfriend',
          girlfriend: getEventCharacterInfo(target, witness.id),
          equipment: { name: equipment.name, level: equipment.level },
        });
        addCapturedBackstoryEvent(game_data, witness, {
          type: 'transmutation_girlfriend_is_executed',
          girlfriend: getEventCharacterInfo(target, target.id),
          equipment: { name: equipment.name, level: equipment.level },
        });
        continue;
      }
    }

    if (binaryChoiceResult(CAPTURE_EVENT_CHANCE)) {
      const event_action = getRandomInArray(/** @type {const} */ (['wait', 'magic', 'play_with']));
      if (event_action === 'wait') {
        addCapturedBackstoryEvent(game_data, character1, {
          type: 'girlfriend_wait_event',
          girlfriend: getEventCharacterInfo(character1, character2.id),
        });
        addCapturedBackstoryEvent(game_data, character2, {
          type: 'girlfriend_wait_event',
          girlfriend: getEventCharacterInfo(character2, character1.id),
        });
      } else if (event_action === 'magic') {
        addCapturedBackstoryEvent(game_data, character1, {
          type: 'girlfriend_magic_event',
          girlfriend: getEventCharacterInfo(character1, character2.id),
          state: 'known',
        });
        addCapturedBackstoryEvent(game_data, character2, {
          type: 'girlfriend_magic_event',
          girlfriend: getEventCharacterInfo(character2, character1.id),
          state: 'known',
        });
      } else if (event_action === 'play_with') {
        /** @type {import('../types').EventEnemyInfo[]} */
        const event_enemies = [{ kind: 'mage' }];
        addCapturedBackstoryEvent(game_data, character1, {
          type: 'girlfriend_play_with_event',
          enemies: event_enemies,
          girlfriend: getEventCharacterInfo(character1, character2.id),
          state: 'known',
        });
        addCapturedBackstoryEvent(game_data, character2, {
          type: 'girlfriend_play_with_event',
          enemies: event_enemies,
          girlfriend: getEventCharacterInfo(character2, character1.id),
          state: 'known',
        });
      }
    }
  }
}

/**
 * @param {import("../types").GameData} game_data The current game
 * @param {ReadonlyArray<{ character: import('../types').StoredCharacterData, item: import('../types').StoredEquipmentData }>} force_drain_into_characters
 */
function createCharacterForceDrainIntoBackstories(game_data, force_drain_into_characters) {
  for (const data of force_drain_into_characters) {
    if (!data.item.character) continue;
    const relations = getRelationsFor(data.character, data.item.character.data.id);
    if (relations.status.is_partner) {
      drainLevelIntoEquipment(data.item, data.character, DRAIN_GIRLFRIEND_EXP_RATIO);
      addCapturedBackstoryEvent(game_data, data.character, {
        type: 'alone_drain_into_girlfriend_event',
        girlfriend: getEventCharacterInfo(data.character, data.item.character.data.id),
        equipment: { name: data.item.name, level: data.item.level },
      });
      addTransmutedBackstoryEvent(game_data, data.item.character.data, {
        type: 'drain_upgrade_event',
        teammate: getEventCharacterInfo(data.item.character.data, data.character.id),
        equipment: { name: data.item.name, level: data.item.level },
      });
    } else {
      drainLevelIntoEquipment(data.item, data.character, DRAIN_STANDARD_EXP_RATIO);
      addCapturedBackstoryEvent(game_data, data.character, {
        type: 'alone_drain_event',
        equipment: { name: data.item.name, level: data.item.level },
      });
      addTransmutedBackstoryEvent(game_data, data.item.character.data, {
        type: 'drain_upgrade_event',
        teammate: getEventCharacterInfo(data.item.character.data, data.character.id),
        equipment: { name: data.item.name, level: data.item.level },
      });
    }
    if (data.character.level === 1 && data.character.partial_exp === 0) {
      game_data.game_state.dungeon_knowledge.forced_drain_into =
        game_data.game_state.dungeon_knowledge.forced_drain_into.filter((i) => i.character !== data.character.id);
    }
  }
}

/**
 * @param {import("../types").GameData} game_data The current game
 * @param {ReadonlyArray<DungeonRandomEvents>} character_with_events
 * @param {number} transformation_chance_ratio
 */
function createCharacterRandomDungeonEventBackstories(game_data, character_with_events, transformation_chance_ratio) {
  for (const group of character_with_events) {
    if (group.type === 'dual') {
      const character1 = group.c1;
      const character2 = group.c2;
      const actions = /** @type {const} */ (['magic', 'play_with']);
      const event_action = getRandomInArray(actions);
      if (event_action == 'play_with') {
        const relations = character1.story.relations.find((r) => r.id === character2.id);
        /** @type {import('../types').EventEnemyInfo[]} */
        const event_enemies = [{ kind: 'mage' }];
        if (relations && relations.status.is_partner) {
          addCapturedBackstoryEvent(game_data, character1, {
            type: 'girlfriend_play_with_event',
            girlfriend: getEventCharacterInfo(character1, character2.id),
            enemies: event_enemies,
            state: 'secret',
          });
          addCapturedBackstoryEvent(game_data, character2, {
            type: 'girlfriend_play_with_event',
            girlfriend: getEventCharacterInfo(character2, character1.id),
            enemies: event_enemies,
            state: 'secret',
          });
        } else {
          addCapturedBackstoryEvent(game_data, character1, {
            type: 'companion_play_with_event',
            captured: getEventCharacterInfo(character1, character2.id),
            enemies: event_enemies,
          });
          addCapturedBackstoryEvent(game_data, character2, {
            type: 'companion_play_with_event',
            captured: getEventCharacterInfo(character2, character1.id),
            enemies: event_enemies,
          });
        }
      } else if (event_action == 'magic') {
        const relations = character1.story.relations.find((r) => r.id === character2.id);
        if (relations && relations.status.is_partner) {
          addCapturedBackstoryEvent(game_data, character1, {
            type: 'girlfriend_magic_event',
            girlfriend: getEventCharacterInfo(character1, character2.id),
            state: 'secret',
          });
          addCapturedBackstoryEvent(game_data, character2, {
            type: 'girlfriend_magic_event',
            girlfriend: getEventCharacterInfo(character2, character1.id),
            state: 'secret',
          });
        } else {
          addCapturedBackstoryEvent(game_data, character1, {
            type: 'companion_magic_event',
            captured: getEventCharacterInfo(character1, character2.id),
          });
          addCapturedBackstoryEvent(game_data, character2, {
            type: 'companion_magic_event',
            captured: getEventCharacterInfo(character2, character1.id),
          });
        }
      }
    } else {
      const character = group.c;
      if (group.isValidTransmuteTarget && binaryChoiceResult(transformation_chance_ratio)) {
        const step = game_data.game_state.dungeon_knowledge.transmutation_states.find((c) => c.id === character.id);
        if (!step) {
          game_data.game_state.dungeon_knowledge.transmutation_states.push({ id: character.id, step: 1 });
          addCapturedBackstoryEvent(game_data, character, { type: 'transmutation_prepare_1_alone' });
          continue;
        } else if (step.step === 1) {
          step.step = 2;
          addCapturedBackstoryEvent(game_data, character, { type: 'transmutation_prepare_2_alone' });
          continue;
        } else if (step.step === 2) {
          step.step = 3;
          addCapturedBackstoryEvent(game_data, character, { type: 'transmutation_prepare_3_alone' });
          continue;
        }
      }

      /** @type {Array<'wait' | 'magic' | 'play_with' | 'exp_drain' | 'transmutation_step'>} */
      const actions = ['wait', 'magic', 'play_with'];
      const target_item_for_drain = getRandomInWeightedArray(
        game_data.game_state.items_outside_town.map((i) => ({ data: i, weight: i.level + (i.character ? 50 : 0) })),
      );
      if ((character.level > 1 || character.partial_exp > 0) && target_item_for_drain) {
        actions.push('exp_drain');
      }

      const event_action = getRandomInArray(actions);
      if (event_action === 'wait') {
        addCapturedBackstoryEvent(game_data, character, {
          type: 'alone_wait_event',
        });
      } else if (event_action == 'play_with') {
        /** @type {import('../types').EventEnemyInfo[]} */
        const event_enemies = [{ kind: 'mage' }];
        addCapturedBackstoryEvent(game_data, character, {
          type: 'alone_play_with_event',
          enemies: event_enemies,
        });
      } else if (event_action == 'magic') {
        addCapturedBackstoryEvent(game_data, character, {
          type: 'alone_magic_event',
        });
      } else if (event_action == 'exp_drain') {
        drainLevelIntoEquipment(target_item_for_drain, character, DRAIN_STANDARD_EXP_RATIO);
        addCapturedBackstoryEvent(game_data, character, {
          type: 'alone_drain_event',
          equipment: { name: target_item_for_drain.name, level: target_item_for_drain.level },
        });
        if (target_item_for_drain.character) {
          addTransmutedBackstoryEvent(game_data, target_item_for_drain.character.data, {
            type: 'drain_upgrade_event',
            teammate: getEventCharacterInfo(target_item_for_drain.character.data, character.id),
            equipment: { name: target_item_for_drain.name, level: target_item_for_drain.level },
          });
        }
      }
    }
  }
}
