import { getRelationsFor } from '../data/backstory.mjs';
import { getAllCharactersInTown } from '../data/utils.mjs';
import { binaryChoiceResult, getRandomInWeightedArray } from '../random.mjs';
import { assertTypeIsHandled } from '../utils.mjs';
import { addTownBackstoryEvent } from './add-backstory-town.mjs';
import { getEventCharacterInfo } from './add-backstory.mjs';

const TOWN_EVENT_CHANCE = 0.25;
const FRIENDSHIP_START_CHANCE = 0.3;
const FRIENDSHIP_TO_ROMANCE_START_CHANCE = 0.3;
const MIN_FRIENDSHIP_ADVANTAGE_FOR_FRIENDSHIP_START = 5;
const MIN_FRIENDSHIP_ADVANTAGE_FOR_ROMANCE_START = 10;
const FRIENDSHIP_TO_RIVALRY_RATIO_FOR_RIVALRY_START = 1.5;
const FRIENDSHIP_TO_RIVALRY_ADVANTAGE_FOR_RIVALRY_START = 3;
const RIVALRY_EVENT_DEFAULT_OFFSET = 2;
const RIVALRY_EVENT_UPPER_OFFSET = 7;
const RIVALRY_EVENT_PROBABILITY_BETWEEN = 0.6;
const RIVALRY_START_CHANCE = 0.8;
const RIVALRY_TO_ROMANCE_START_CHANCE = 0.8;
const RIVALRY_END_CHANCE = 0.3;

// Note: this uses a logistic function whose parameters are created from the two constants above.
// Logistic function: x => 1 / ( 1 + e^( -k * (x - x0) ) )
// If you have a p% probability != 0.5 for value x !== x0 => k = -ln(1/p - 1) / (x - x0)
const rivalry_event_x0_value =
  RIVALRY_EVENT_DEFAULT_OFFSET + (RIVALRY_EVENT_UPPER_OFFSET - RIVALRY_EVENT_DEFAULT_OFFSET) / 2;
const rivalry_event_probability_at_default = (1 - RIVALRY_EVENT_PROBABILITY_BETWEEN) / 2;
const rivalry_event_k_value =
  -Math.log(1 / rivalry_event_probability_at_default - 1) / (RIVALRY_EVENT_DEFAULT_OFFSET - rivalry_event_x0_value);

/**
 * @param {import("../types").GameData} game_data The current game
 */
export function handleTownCharacterUpdate(game_data) {
  const characters_in_town = getAllCharactersInTown(game_data);
  /** @type {Array<{ character: import("../types").StoredCharacterData, job?: "training" | "upgrading" }>} */
  const characters_with_events = [];
  for (const character of characters_in_town) {
    if (binaryChoiceResult(TOWN_EVENT_CHANCE)) {
      if (game_data.town.change_role.active && game_data.town.change_role.target === character) {
        characters_with_events.push({ character: character, job: 'training' });
      } else if (game_data.town.upgrade_equipment.active && game_data.town.upgrade_equipment.handler === character) {
        characters_with_events.push({ character: character, job: 'upgrading' });
      } else {
        characters_with_events.push({ character: character });
      }
    }
  }

  // Shuffle the characters
  characters_in_town.sort(() => 0.5 - Math.random());

  // Pair the characters

  /**
   * @type {Array<{
   *   character1: import("../types").StoredCharacterData,
   *   character2: import("../types").StoredCharacterData,
   *   job?: "c1training" | "c1upgrading" | "c1training_and_c2upgrading",
   * }>}
   */
  const character_pairs = [];
  while (characters_with_events.length > 0) {
    const character1Data = characters_with_events.pop();
    if (!character1Data) break;
    const character2Data = characters_with_events.pop();
    if (!character2Data) break;
    if (character1Data.job === 'training') {
      if (character2Data.job === 'upgrading') {
        character_pairs.push({
          character1: character1Data.character,
          character2: character2Data.character,
          job: 'c1training_and_c2upgrading',
        });
      } else {
        character_pairs.push({
          character1: character1Data.character,
          character2: character2Data.character,
          job: 'c1training',
        });
      }
    } else if (character2Data.job === 'training') {
      // Note: we swap the characters for these two options
      if (character1Data.job === 'upgrading') {
        character_pairs.push({
          character1: character2Data.character,
          character2: character1Data.character,
          job: 'c1training_and_c2upgrading',
        });
      } else {
        character_pairs.push({
          character1: character2Data.character,
          character2: character1Data.character,
          job: 'c1training',
        });
      }
    } else if (character1Data.job === 'upgrading') {
      character_pairs.push({
        character1: character1Data.character,
        character2: character2Data.character,
        job: 'c1upgrading',
      });
    } else if (character2Data.job === 'upgrading') {
      // Note: we swap the characters
      character_pairs.push({
        character1: character2Data.character,
        character2: character1Data.character,
        job: 'c1upgrading',
      });
    } else {
      character_pairs.push({ character1: character1Data.character, character2: character2Data.character });
    }
  }

  // Create events for the paired characters
  for (const pair of character_pairs) {
    const character1 = pair.character1;
    const character2 = pair.character2;
    const c1_for_c2_relations = getRelationsFor(character1, character2.id);
    const c2_for_c1_relations = getRelationsFor(character2, character1.id);

    const are_partners = c1_for_c2_relations.status.is_partner;
    const c1_has_other_partner = character1.story.relations.some((r) => r.id !== character2.id && r.status.is_partner);
    const c2_has_other_partner = character2.story.relations.some((r) => r.id !== character1.id && r.status.is_partner);
    const has_other_partner = c1_has_other_partner || c2_has_other_partner;
    const is_team_member = c1_for_c2_relations.status.is_team_member;
    const is_hardship_companion = c1_for_c2_relations.status.is_hardship_companion;
    const is_rivalry = c1_for_c2_relations.status.is_rival_bully || c1_for_c2_relations.status.is_rival_target;
    const c1_guilt = c1_for_c2_relations.interactions.guilt_interactions;
    const c2_guilt = c2_for_c1_relations.interactions.guilt_interactions;
    const c1_protect = c1_for_c2_relations.interactions.did_protect_interactions;
    const c2_protect = c2_for_c1_relations.interactions.did_protect_interactions;
    const total_hardship = c1_for_c2_relations.interactions.hardship_interactions;
    const total_friendship = c1_for_c2_relations.interactions.friendship_interactions;
    const total_rivalry =
      c1_for_c2_relations.interactions.rivalry_bullies_interactions +
      c1_for_c2_relations.interactions.rivalry_is_bullied_interactions;

    const can_start_friendship =
      !c1_for_c2_relations.status.is_friend &&
      total_hardship + total_friendship - total_rivalry > MIN_FRIENDSHIP_ADVANTAGE_FOR_FRIENDSHIP_START;
    const can_start_friendship_to_romance =
      !has_other_partner &&
      c1_for_c2_relations.status.is_friend &&
      total_friendship > MIN_FRIENDSHIP_ADVANTAGE_FOR_ROMANCE_START;

    /** @type {'friendship' | 'rivalry' | 'romance'} */
    let interaction_type = 'friendship';
    if (!c1_for_c2_relations.status.is_friend) {
      if (is_rivalry) {
        interaction_type = 'rivalry';
      } else {
        const rivalry_skew = total_friendship - total_rivalry;
        const rivalry_chance = 1 / (1 + Math.exp(-rivalry_event_k_value * (rivalry_skew - rivalry_event_x0_value)));
        if (binaryChoiceResult(rivalry_chance)) interaction_type = 'rivalry';
      }
    }
    if (are_partners) interaction_type = 'romance';

    let rivalry_c1_would_bully = false;
    if (c1_for_c2_relations.status.is_rival_target) {
      rivalry_c1_would_bully = true;
    } else {
      const c1_bully_event_ratio =
        c1_for_c2_relations.interactions.rivalry_bullies_interactions -
        c1_for_c2_relations.interactions.rivalry_is_bullied_interactions;
      const probability_of_contrary = Math.exp((-c1_bully_event_ratio * c1_bully_event_ratio) / 10) / 2;
      const probability = c1_bully_event_ratio > 0 ? 1 - probability_of_contrary : probability_of_contrary;
      if (binaryChoiceResult(probability)) {
        rivalry_c1_would_bully = true;
      }
    }

    // Special case: handle custom interactions through jobs
    if (pair.job) {
      if (pair.job === 'c1training') {
        if (interaction_type === 'romance') {
          addTownBackstoryEvent(game_data, character1, {
            type: 'romance_event',
            kind: 'training',
            girlfriend: getEventCharacterInfo(character1, character2.id),
          });
          addTownBackstoryEvent(game_data, character2, {
            type: 'romance_event',
            kind: 'watched_training',
            girlfriend: getEventCharacterInfo(character2, character1.id),
          });
          continue;
        } else if (interaction_type === 'rivalry') {
          if (rivalry_c1_would_bully) {
            addTownBackstoryEvent(game_data, character1, {
              type: 'rivalry_event_bullies',
              kind: 'training',
              rival: getEventCharacterInfo(character1, character2.id),
            });
            addTownBackstoryEvent(game_data, character2, {
              type: 'rivalry_event_is_bullied',
              kind: 'watched_training',
              rival: getEventCharacterInfo(character2, character1.id),
            });
          } else {
            addTownBackstoryEvent(game_data, character1, {
              type: 'rivalry_event_is_bullied',
              kind: 'training',
              rival: getEventCharacterInfo(character1, character2.id),
            });
            addTownBackstoryEvent(game_data, character2, {
              type: 'rivalry_event_bullies',
              kind: 'watched_training',
              rival: getEventCharacterInfo(character2, character1.id),
            });
          }
          continue;
        } else if (interaction_type === 'friendship') {
          addTownBackstoryEvent(game_data, character1, {
            type: 'friendship_event',
            kind: 'training',
            friend: getEventCharacterInfo(character1, character2.id),
          });
          addTownBackstoryEvent(game_data, character2, {
            type: 'friendship_event',
            kind: 'watched_training',
            friend: getEventCharacterInfo(character2, character1.id),
          });
          continue;
        }
        assertTypeIsHandled(interaction_type, 'interaction kind');
      } else if (pair.job === 'c1upgrading') {
        if (interaction_type === 'romance') {
          addTownBackstoryEvent(game_data, character1, {
            type: 'romance_event',
            kind: 'upgrading_item',
            girlfriend: getEventCharacterInfo(character1, character2.id),
          });
          addTownBackstoryEvent(game_data, character2, {
            type: 'romance_event',
            kind: 'watched_upgrade_item',
            girlfriend: getEventCharacterInfo(character2, character1.id),
          });
          continue;
        } else if (interaction_type === 'rivalry') {
          if (rivalry_c1_would_bully) {
            addTownBackstoryEvent(game_data, character1, {
              type: 'rivalry_event_bullies',
              kind: 'upgrading_item',
              rival: getEventCharacterInfo(character1, character2.id),
            });
            addTownBackstoryEvent(game_data, character2, {
              type: 'rivalry_event_is_bullied',
              kind: 'watched_upgrade_item',
              rival: getEventCharacterInfo(character2, character1.id),
            });
          } else {
            addTownBackstoryEvent(game_data, character1, {
              type: 'rivalry_event_is_bullied',
              kind: 'upgrading_item',
              rival: getEventCharacterInfo(character1, character2.id),
            });
            addTownBackstoryEvent(game_data, character2, {
              type: 'rivalry_event_bullies',
              kind: 'watched_upgrade_item',
              rival: getEventCharacterInfo(character2, character1.id),
            });
          }
          continue;
        } else if (interaction_type === 'friendship') {
          addTownBackstoryEvent(game_data, character1, {
            type: 'friendship_event',
            kind: 'upgrading_item',
            friend: getEventCharacterInfo(character1, character2.id),
          });
          addTownBackstoryEvent(game_data, character2, {
            type: 'friendship_event',
            kind: 'watched_upgrade_item',
            friend: getEventCharacterInfo(character2, character1.id),
          });
          continue;
        }
        assertTypeIsHandled(interaction_type, 'interaction kind');
      } else if (pair.job === 'c1training_and_c2upgrading') {
        if (interaction_type === 'romance') {
          addTownBackstoryEvent(game_data, character1, {
            type: 'romance_event',
            kind: 'training_watching_upgrading_item',
            girlfriend: getEventCharacterInfo(character1, character2.id),
          });
          addTownBackstoryEvent(game_data, character2, {
            type: 'romance_event',
            kind: 'upgrading_item_watching_training',
            girlfriend: getEventCharacterInfo(character2, character1.id),
          });
          continue;
        } else if (interaction_type === 'rivalry') {
          if (rivalry_c1_would_bully) {
            addTownBackstoryEvent(game_data, character1, {
              type: 'rivalry_event_bullies',
              kind: 'training_watching_upgrading_item',
              rival: getEventCharacterInfo(character1, character2.id),
            });
            addTownBackstoryEvent(game_data, character2, {
              type: 'rivalry_event_is_bullied',
              kind: 'upgrading_item_watching_training',
              rival: getEventCharacterInfo(character2, character1.id),
            });
          } else {
            addTownBackstoryEvent(game_data, character1, {
              type: 'rivalry_event_is_bullied',
              kind: 'training_watching_upgrading_item',
              rival: getEventCharacterInfo(character1, character2.id),
            });
            addTownBackstoryEvent(game_data, character2, {
              type: 'rivalry_event_bullies',
              kind: 'upgrading_item_watching_training',
              rival: getEventCharacterInfo(character2, character1.id),
            });
          }
          continue;
        } else if (interaction_type === 'friendship') {
          addTownBackstoryEvent(game_data, character1, {
            type: 'friendship_event',
            kind: 'training_watching_upgrading_item',
            friend: getEventCharacterInfo(character1, character2.id),
          });
          addTownBackstoryEvent(game_data, character2, {
            type: 'friendship_event',
            kind: 'upgrading_item_watching_training',
            friend: getEventCharacterInfo(character2, character1.id),
          });
          continue;
        }
        assertTypeIsHandled(interaction_type, 'interaction kind');
      }
      assertTypeIsHandled(pair.job, 'interaction job');
    }

    /** @type {Array<{ data: import("../types").CharacterBackstoryTownSharedEvent, weight: number }>} */
    const default_interaction_kinds = [
      { data: 'patrol_round', weight: 4 },
      { data: 'evening_guard_round', weight: 4 },
      { data: 'gathering_food', weight: 2 },
      { data: 'town_cleaning', weight: 1 },
    ];

    if (interaction_type === 'romance') {
      const interaction_kind = getRandomInWeightedArray(
        /** @type {const} */ ([
          ...default_interaction_kinds,
          { data: 'eating_lively_conversation', weight: 5 },
          { data: 'friendly_spar', weight: 5 },
          { data: 'late_afternoon_stroll', weight: 4 },
          { data: 'friendly_spar_with_fun', weight: 1 },
        ]),
      );
      if (interaction_kind === 'friendly_spar_with_fun') {
        addTownBackstoryEvent(game_data, character1, {
          type: 'romance_event',
          kind: 'friendly_spar_with_fun_tops',
          girlfriend: getEventCharacterInfo(character1, character2.id),
        });
        addTownBackstoryEvent(game_data, character2, {
          type: 'romance_event',
          kind: 'friendly_spar_with_fun_bots',
          girlfriend: getEventCharacterInfo(character2, character1.id),
        });
      } else {
        addTownBackstoryEvent(game_data, character1, {
          type: 'romance_event',
          kind: interaction_kind,
          girlfriend: getEventCharacterInfo(character1, character2.id),
        });
        addTownBackstoryEvent(game_data, character2, {
          type: 'romance_event',
          kind: interaction_kind,
          girlfriend: getEventCharacterInfo(character2, character1.id),
        });
      }
      continue;
    }

    if (is_rivalry) {
      const can_end_rivalry =
        can_start_friendship &&
        c1_guilt + c1_protect > c1_for_c2_relations.interactions.rivalry_bullies_interactions &&
        c2_guilt + c2_protect > c2_for_c1_relations.interactions.rivalry_bullies_interactions;
      if (!has_other_partner && can_end_rivalry && binaryChoiceResult(RIVALRY_TO_ROMANCE_START_CHANCE)) {
        addTownBackstoryEvent(game_data, character1, {
          type: 'romance_start',
          mode: 'rivalry_ends',
          girlfriend: getEventCharacterInfo(character1, character2.id),
        });
        addTownBackstoryEvent(game_data, character2, {
          type: 'romance_start',
          mode: 'rivalry_ends',
          girlfriend: getEventCharacterInfo(character2, character1.id),
        });
        continue;
      }
      if (can_end_rivalry && binaryChoiceResult(RIVALRY_END_CHANCE)) {
        addTownBackstoryEvent(game_data, character1, {
          type: 'rivalry_end',
          rival: getEventCharacterInfo(character1, character2.id),
        });
        addTownBackstoryEvent(game_data, character2, {
          type: 'rivalry_end',
          rival: getEventCharacterInfo(character2, character1.id),
        });
        continue;
      }
    }

    if (interaction_type === 'rivalry') {
      const can_start_rivalry =
        !is_rivalry &&
        total_rivalry - total_friendship > FRIENDSHIP_TO_RIVALRY_ADVANTAGE_FOR_RIVALRY_START &&
        total_rivalry > total_friendship * FRIENDSHIP_TO_RIVALRY_RATIO_FOR_RIVALRY_START;
      if (can_start_rivalry && binaryChoiceResult(RIVALRY_START_CHANCE)) {
        const c1_is_bully =
          c1_for_c2_relations.interactions.rivalry_bullies_interactions >
          c1_for_c2_relations.interactions.rivalry_is_bullied_interactions;
        if (c1_is_bully) {
          addTownBackstoryEvent(game_data, character1, {
            type: 'rivalry_start_bullies',
            rival: getEventCharacterInfo(character1, character2.id),
          });
          addTownBackstoryEvent(game_data, character2, {
            type: 'rivalry_start_is_bullied',
            rival: getEventCharacterInfo(character2, character1.id),
          });
        } else {
          addTownBackstoryEvent(game_data, character1, {
            type: 'rivalry_start_is_bullied',
            rival: getEventCharacterInfo(character1, character2.id),
          });
          addTownBackstoryEvent(game_data, character2, {
            type: 'rivalry_start_bullies',
            rival: getEventCharacterInfo(character2, character1.id),
          });
        }
        continue;
      }

      const raw_viciousness = (c1_for_c2_relations.status.is_rival_target ? 10 : 1) * total_rivalry;
      const viciousness_percent = 1 - Math.exp(-Math.sqrt(raw_viciousness / 10));

      /** @type {Array<{ data: import("../types").CharacterBackstoryRivalryEvent, weight: number }>} */
      let possible_weighted_interactions = [...default_interaction_kinds];
      if (viciousness_percent > 0.75) {
        possible_weighted_interactions.push({ data: 'fight_then_prank_ties_up', weight: 3 });
        possible_weighted_interactions.push({ data: 'fight_cheats', weight: 2 });
        possible_weighted_interactions.push({ data: 'fight_taunts', weight: 2 });
        possible_weighted_interactions.push({ data: 'evening_guard_round_prank_ties_up', weight: 3 });
        possible_weighted_interactions.push({ data: 'town_cleaning_prank_ties_up', weight: 3 });
        possible_weighted_interactions.push({ data: 'prank_hides_clothes', weight: 1 });
      } else if (viciousness_percent > 0.4) {
        possible_weighted_interactions.push({ data: 'fight_taunts', weight: 3 });
        possible_weighted_interactions.push({ data: 'fight_then_prank_hides_bag', weight: 2 });
        possible_weighted_interactions.push({ data: 'prank_hides_bag', weight: 2 });
        possible_weighted_interactions.push({ data: 'fight_cheats', weight: 1 });
      } else {
        possible_weighted_interactions.push({ data: 'fight_fair_win', weight: 3 });
        possible_weighted_interactions.push({ data: 'fight_taunts', weight: 2 });
        possible_weighted_interactions.push({ data: 'prank_hides_bag', weight: 1 });
      }

      const interaction_kind = getRandomInWeightedArray(possible_weighted_interactions);
      if (rivalry_c1_would_bully) {
        addTownBackstoryEvent(game_data, character1, {
          type: 'rivalry_event_bullies',
          kind: interaction_kind,
          rival: getEventCharacterInfo(character1, character2.id),
        });
        addTownBackstoryEvent(game_data, character2, {
          type: 'rivalry_event_is_bullied',
          kind: interaction_kind,
          rival: getEventCharacterInfo(character2, character1.id),
        });
      } else {
        addTownBackstoryEvent(game_data, character1, {
          type: 'rivalry_event_is_bullied',
          kind: interaction_kind,
          rival: getEventCharacterInfo(character1, character2.id),
        });
        addTownBackstoryEvent(game_data, character2, {
          type: 'rivalry_event_bullies',
          kind: interaction_kind,
          rival: getEventCharacterInfo(character2, character1.id),
        });
      }
      continue;
    }

    if (interaction_type === 'friendship') {
      if (can_start_friendship_to_romance && binaryChoiceResult(FRIENDSHIP_TO_ROMANCE_START_CHANCE)) {
        addTownBackstoryEvent(game_data, character1, {
          type: 'romance_start',
          mode: 'standard',
          girlfriend: getEventCharacterInfo(character1, character2.id),
        });
        addTownBackstoryEvent(game_data, character2, {
          type: 'romance_start',
          mode: 'standard',
          girlfriend: getEventCharacterInfo(character2, character1.id),
        });
        continue;
      }
      if (can_start_friendship && binaryChoiceResult(FRIENDSHIP_START_CHANCE)) {
        addTownBackstoryEvent(game_data, character1, {
          type: 'friendship_start',
          friend: getEventCharacterInfo(character1, character2.id),
        });
        addTownBackstoryEvent(game_data, character2, {
          type: 'friendship_start',
          friend: getEventCharacterInfo(character2, character1.id),
        });
        continue;
      }

      const interaction_kind = getRandomInWeightedArray(
        /** @type {const} */ ([
          ...default_interaction_kinds,
          { data: 'eating_lively_conversation', weight: 5 },
          { data: 'friendly_spar', weight: 3 },
        ]),
      );

      addTownBackstoryEvent(game_data, character1, {
        type: 'friendship_event',
        kind: interaction_kind,
        friend: getEventCharacterInfo(character1, character2.id),
      });
      addTownBackstoryEvent(game_data, character2, {
        type: 'friendship_event',
        kind: interaction_kind,
        friend: getEventCharacterInfo(character2, character1.id),
      });
      continue;
    }

    assertTypeIsHandled(interaction_type, 'interaction type');
  }
}
