import { generateRandomStoredEquipment, getEquipmentFromCharacter } from '../data/equipment.mjs';
import {
  addCaptureAloneOriginStory,
  addCaptureTeamOriginStory,
  addFriendsOriginBackstory,
  addGirlfriendsOriginBackstory,
  addRandomOriginStory,
  addRivalsOriginBackstory,
  generateHumanName,
  generateStoredCharacter,
} from '../data/player-character.mjs';
import { getAllLiveTransmutedAndDustedCharacters, getAllLiveCharacters } from '../data/utils.mjs';
import { binaryChoiceResult, getRandomInArray } from '../random.mjs';
import { addTransmutedBackstoryEvent } from './add-backstory-transmuted.mjs';
import { getEventCharacterInfo } from './add-backstory.mjs';
import { handleOutsideCharacterUpdate } from './advance-cycle-outside-characters.mjs';
import { handleTownCharacterUpdate } from './advance-cycle-town-backstory.mjs';
import { handleTownUpgradeEvents } from './advance-cycle-town-upgrades.mjs';
import { forceFullGameOver } from './force-full-game-over.mjs';

const MAX_LIVE_CHARACTERS = 8;
const MIN_AVAILABLE_ITEMS = 1;

const TRANSMUTED_ALONE_CHANCE = 0.02;
const TRANSMUTED_ENEMY_USE_CHANCE = 0.02;
const TRANSMUTED_ENEMY_USE_GROUP_CHANCE = 0.3;

/**
 * @param {import("../types").GameData} game_data
 * @param {{ advance_mode: "skip_in_town" }
 *  | { advance_mode: "dungeon_fight", fight_mode: "small_group" | "large_group" | "ambush" }
 * } mode
 */
export function advanceCycle(game_data, mode) {
  game_data.game_state.current_cycle++;

  handleCorruptionUpdate(game_data, mode);
  handleRandomItemsCreation(game_data);
  handleTownCharacterUpdate(game_data);
  handleTownUpgradeEvents(game_data);
  handleOutsideItemsUpdate(game_data);
  handleOutsideCharacterUpdate(game_data);
  handleNewCharacterArrivals(game_data);

  if (mode.advance_mode === 'skip_in_town' && game_data.game_state.current_corruption_ratio >= 1.0) {
    forceFullGameOver(game_data);
  }
}

/**
 * @param {import("../types").GameData} game_data
 * @param {{ advance_mode: "skip_in_town" }
 *  | { advance_mode: "dungeon_fight", fight_mode: "small_group" | "large_group" | "ambush" }
 * } mode
 */
function handleCorruptionUpdate(game_data, mode) {
  if (mode.advance_mode === 'skip_in_town') {
    game_data.game_state.current_corruption_ratio += 0.01;
    if (game_data.game_state.current_corruption_ratio > 1.0) game_data.game_state.current_corruption_ratio = 1.0;
  }
  if (mode.advance_mode === 'dungeon_fight') {
    if (mode.fight_mode === 'small_group') {
      game_data.game_state.current_corruption_ratio -= 0.01;
    }
    if (mode.fight_mode === 'large_group') {
      game_data.game_state.current_corruption_ratio -= 0.05;
    }
    if (mode.fight_mode === 'ambush') {
      game_data.game_state.current_corruption_ratio += 0.01;
    }
    if (game_data.game_state.current_corruption_ratio < 0) game_data.game_state.current_corruption_ratio = 0;
    if (game_data.game_state.current_corruption_ratio > 1.0) game_data.game_state.current_corruption_ratio = 1.0;
  }
}

/**
 * @param {import("../types").GameData} game_data The current game
 */
function handleOutsideItemsUpdate(game_data) {
  let accessory_characters_outside = game_data.game_state.items_outside_town
    .filter((c) => c.type === 'accessory')
    .map((e) => e.character)
    .filter((e) => !!e)
    .map((e) => e.data);
  let body_characters_outside = game_data.game_state.items_outside_town
    .filter((c) => c.type === 'body')
    .map((e) => e.character)
    .filter((e) => !!e)
    .map((e) => e.data);
  let weapon_characters_outside = game_data.game_state.items_outside_town
    .filter((c) => c.type === 'weapon')
    .map((e) => e.character)
    .filter((e) => !!e)
    .map((e) => e.data);

  /** @type {import("../types").StoredCharacterData[]} */
  const transmuted_group = [];
  if (binaryChoiceResult(TRANSMUTED_ENEMY_USE_GROUP_CHANCE)) {
    const accessory = getRandomInArray(accessory_characters_outside);
    if (accessory) {
      accessory_characters_outside = accessory_characters_outside.filter((a) => a !== accessory);
      transmuted_group.push(accessory);
    }
  }
  if (binaryChoiceResult(TRANSMUTED_ENEMY_USE_GROUP_CHANCE)) {
    const body = getRandomInArray(body_characters_outside);
    if (body) {
      body_characters_outside = body_characters_outside.filter((a) => a !== body);
      transmuted_group.push(body);
    }
  }
  if (binaryChoiceResult(TRANSMUTED_ENEMY_USE_GROUP_CHANCE)) {
    const weapon = getRandomInArray(weapon_characters_outside);
    if (weapon) {
      weapon_characters_outside = weapon_characters_outside.filter((a) => a !== weapon);
      transmuted_group.push(weapon);
    }
  }

  if (transmuted_group.length > 1) {
    for (const character of transmuted_group) {
      const others = transmuted_group.filter((g) => g !== character);
      addTransmutedBackstoryEvent(game_data, character, {
        type: 'enemy_group_use_event',
        capturedGroup: others.map((c) => getEventCharacterInfo(character, c.id)),
        enemies: [{ kind: 'mage' }],
      });
    }
  }

  const remaining_transmuted_characters = body_characters_outside
    .concat(accessory_characters_outside)
    .concat(weapon_characters_outside);
  for (const character of remaining_transmuted_characters) {
    if (binaryChoiceResult(TRANSMUTED_ALONE_CHANCE)) {
      addTransmutedBackstoryEvent(game_data, character, { type: 'alone_event' });
    } else if (binaryChoiceResult(TRANSMUTED_ENEMY_USE_CHANCE)) {
      addTransmutedBackstoryEvent(game_data, character, {
        type: 'enemy_use_event',
        enemies: [{ kind: 'mage' }],
      });
    }
  }
}

/**
 * @param {import("../types").GameData} game_data The current game
 */
function handleNewCharacterArrivals(game_data) {
  const live_character_count = getAllLiveCharacters(game_data).length;
  if (live_character_count < MAX_LIVE_CHARACTERS) {
    if (binaryChoiceResult(0.01 * game_data.dungeon.level.exploration_cycles)) {
      const used_names = getAllLiveTransmutedAndDustedCharacters(game_data).map((c) => c.name);
      if (live_character_count - MAX_LIVE_CHARACTERS >= 2 && binaryChoiceResult(0.2)) {
        const c1Name = generateHumanName(used_names);
        const c2Name = generateHumanName(used_names.concat(c1Name));
        const character1 = generateStoredCharacter('adventurer', 1, { events: [], relations: [] }, c1Name);
        addRandomOriginStory(character1);
        const character2 = generateStoredCharacter('adventurer', 1, { events: [], relations: [] }, c2Name);
        addRandomOriginStory(character2);
        if (binaryChoiceResult(0.2)) {
          addGirlfriendsOriginBackstory(character1, character2);
        } else if (binaryChoiceResult(0.2)) {
          addRivalsOriginBackstory(character1, character2);
        } else {
          addFriendsOriginBackstory(character1, character2);
        }
        addCaptureTeamOriginStory(character1, [character2], game_data.game_state.current_cycle);
        addCaptureTeamOriginStory(character2, [character1], game_data.game_state.current_cycle);
        game_data.game_state.characters_outside_town.push(character1);
      } else {
        const new_character_name = generateHumanName(used_names);
        const character = generateStoredCharacter('adventurer', 1, { events: [], relations: [] }, new_character_name);
        addRandomOriginStory(character);
        addCaptureAloneOriginStory(character, game_data.game_state.current_cycle);
        game_data.game_state.characters_outside_town.push(character);
      }
    }
  }
}

/**
 * @param {import("../types").GameData} game_data The current game
 */
function handleRandomItemsCreation(game_data) {
  const items_outside = game_data.game_state.items_outside_town.length;
  if (items_outside >= MIN_AVAILABLE_ITEMS) return;
  game_data.game_state.items_outside_town.push(generateRandomStoredEquipment(1));
}
