import { binaryChoiceResult } from '../random.mjs';
import {
  addExplorationBackstoryCompanionsLost,
  addExplorationBackstoryBackToTown,
} from './add-backstory-exploration.mjs';

/**
 * @param {import("../types").GameData} game_data
 * @param {number} team_lost_ratio The base ratio for game-overed characters to not be recovered (all-or-nothing chance)
 * @param {number} items_lost_ratio The base ratio for game-overed items to not be recovered (applies to each one individually)
 */
export function leaveDungeonForTown(game_data, team_lost_ratio, items_lost_ratio) {
  const exploration_team = game_data.dungeon.team
    .filter((f) => !f.status_effects.some((s) => s.characteristics.behavior.type === 'game_over'))
    .map((t) => t.base);
  transferGatheredEquipments(game_data, items_lost_ratio);
  const transferResult = transferNonTeamCharacters(game_data, team_lost_ratio);
  transferTeam(game_data);

  if (transferResult.captured_characters.length > 0) {
    addExplorationBackstoryCompanionsLost(game_data, exploration_team, transferResult.captured_characters, 'exit');
  }

  for (const character of transferResult.saved_characters) {
    game_data.game_state.dungeon_knowledge.interrogations_completed =
      game_data.game_state.dungeon_knowledge.interrogations_completed.filter((c) => c !== character.id);
  }

  addExplorationBackstoryBackToTown(game_data, exploration_team, transferResult.saved_characters);

  game_data.dungeon.level.exploration_cycles = 0;
  game_data.dungeon.spawn.ambush_cooldown = -1;
  game_data.game_state.screen = 'town';
  game_data.town.screen = 'town';
}

/**
 * @param {import("../types").GameData} game_data
 * @param {number} lost_chance_ratio The base ratio for game-overed characters to not be recovered (whatever their current class and stuff)
 */
function transferGatheredEquipments(game_data, lost_chance_ratio) {
  for (const equipment of game_data.dungeon.rewards.gathered_equipments) {
    if (binaryChoiceResult(lost_chance_ratio)) {
      game_data.game_state.items_outside_town.push(equipment);
    } else {
      game_data.town.unused_equipment.push(equipment);
    }
  }
  game_data.dungeon.rewards.gathered_equipments = [];
}

/**
 * @param {import("../types").GameData} game_data
 * @param {number} lost_chance_ratio The base ratio for game-overed characters to not be recovered (whatever their current class and stuff)
 * @returns {{ captured_characters: import('../types').StoredCharacterData[], saved_characters: import('../types').StoredCharacterData[] }}
 */
function transferNonTeamCharacters(game_data, lost_chance_ratio) {
  /** @type {import('../types').StoredCharacterData[]} */
  const captured_characters = [];
  /** @type {import('../types').StoredCharacterData[]} */
  const saved_characters = [];
  if (binaryChoiceResult(lost_chance_ratio)) {
    for (const character of game_data.dungeon.rewards.saved_characters) {
      game_data.game_state.characters_outside_town.push(character);
      captured_characters.push(character);
    }
    for (const character of game_data.dungeon.team) {
      if (!character.status_effects.some((s) => s.characteristics.behavior.type === 'game_over')) continue;
      game_data.game_state.characters_outside_town.push(character.base);
      captured_characters.push(character.base);
    }
  } else {
    for (const character of game_data.dungeon.rewards.saved_characters) {
      game_data.town.available_team.push(character);
      saved_characters.push(character);
    }
    for (const character of game_data.dungeon.team) {
      if (!character.status_effects.some((s) => s.characteristics.behavior.type === 'game_over')) continue;
      saved_characters.push(character.base);
    }
  }

  game_data.dungeon.team = game_data.dungeon.team.filter((t) => !captured_characters.some((c) => c === t.base));
  game_data.dungeon.rewards.saved_characters = [];

  return { captured_characters: captured_characters, saved_characters: saved_characters };
}

/**
 * @param {import("../types").GameData} game_data
 */
function transferTeam(game_data) {
  for (let i = 0; i < 4; ++i) {
    const slot = game_data.town.expedition_slots[i];
    const player = game_data.dungeon.team[i];
    if (player && slot) {
      slot.character = player.base;
      slot.equipment = player.equipment;
    }
  }
  game_data.dungeon.team = [];
}
