/**
 * @typedef { import("../types").ImageAssetPartialInformation } ImageAssetPartialInformation
 */

/**
 * Load a list of images from their base URL and their source information.
 * @param {string} image_url The base URL of the image
 * @param {ImageAssetPartialInformation[]} partial_information The partial informations for the subimages to extract out of the image
 * @returns {Promise<ImageBitmap>[]} A list of promises for the image bitmaps.
 */
function loadImage(image_url, partial_information) {
  /** @type {Promise<Blob> | null} */
  let image_fetcher = null;
  /** @return {Promise<Blob>} */
  function getImage() {
    if (!image_fetcher) {
      image_fetcher = fetch(image_url).then((response) => response.blob());
    }
    return image_fetcher;
  }

  return partial_information.map((source_info) => {
    return getImage().then((image) => {
      return createImageBitmap(
        image,
        source_info.origin.x,
        source_info.origin.y,
        source_info.origin.w,
        source_info.origin.h,
        {
          resizeWidth: source_info.target.w,
          resizeHeight: source_info.target.h,
        },
      );
    });
  });
}

/**
 * Load an image from a source path.
 * @param {{ [file_url: string]: { [key_name: string]: ImageAssetPartialInformation } }} image_request_list The list of key to source path of the images to load
 * @param { (count: number) => void } onImageCountReady The method to execute when the total count of images to load has been detected. Will only be called once.
 * @param { () => void } onSingleImageLoaded The method to execute each time an image has been loaded. Will be called several times. Will only be called after `onImageCountReady`.
 * @returns {Promise<{ [key: string]: ImageBitmap }>} A promise for the image itself.
 */
export function loadImageObject(image_request_list, onImageCountReady, onSingleImageLoaded) {
  /** @type {string[]} */
  const key_list = [];
  const load_list = [];
  for (const file_name in image_request_list) {
    const file_data = image_request_list[file_name];
    const request_list = [];
    for (const key_name in file_data) {
      const request_data = file_data[key_name];
      key_list.push(key_name);
      request_list.push(request_data);
    }
    load_list.push(...loadImage(file_name, request_list));
  }
  onImageCountReady(load_list.length);
  return Promise.all(
    load_list.map((i) =>
      i.then((i) => {
        onSingleImageLoaded();
        return i;
      }),
    ),
  ).then(function (loadedImages) {
    /** @type {{[key: string]: ImageBitmap}} */
    const result = {};
    for (let i = 0; i < key_list.length; ++i) {
      const key = key_list[i];
      const loadedImage = loadedImages[i];
      result[key] = loadedImage;
    }
    return result;
  });
}

/**
 * @param {{ [file_url: string]: import("../types").AudioAssetPartialInformation }} assets
 * @returns {Promise<import("../types").AudioAssets>}
 */
export function loadAudioObject(assets) {
  return new Promise((resolve) => {
    /** @type {import("../types").AudioAssets["snippets"]} */
    const snippets = {};
    for (const file_url in assets) {
      snippets[assets[file_url].name] = {
        kind: assets[file_url].kind,
        media: new Audio(file_url),
      };
      snippets[assets[file_url].name].media.load();
    }
    resolve({ snippets: snippets });
  });
}
