import { Game } from './engine.mjs';

import { image_descriptions } from './loader/images.mjs';
import { audio_descriptions } from './loader/audio.mjs';
import { loadAudioObject, loadImageObject } from './loader/resources.mjs';
import { EXPECTED_VERSION, getGameData } from './game-data.mjs';

import { getDungeonSystem } from './system/dungeon/index.mjs';
import { getParticlesSystem } from './system/particles.mjs';
import { getTownSystem } from './system/town/index.mjs';
import { getGameOverSystem } from './system/game-over.mjs';
import { getPopupSystem } from './system/popup.mjs';
import { getAllLiveTransmutedAndDustedCharacters } from './data/utils.mjs';
import { generateReadableBackstory } from './data/backstory-readable.mjs';
import { getPauseSystem } from './system/pause.mjs';
import { getRandomNumber } from './random.mjs';

/** @type {HTMLCanvasElement} */
// @ts-ignore (for some reason, this next line isn't cast in the proper type automatically)
const canvas = document.getElementById('game');
const context = (() => {
  const context = canvas.getContext('2d');
  if (!context) throw new Error('Could not create 2D context');
  return context;
})();

const launchWarningMessage = document.getElementById('launch_warning');
if (launchWarningMessage) {
  launchWarningMessage.remove();
}

const MS_BETWEEN_TICKS = 1000 / 20;
const MAX_MS_BETWEEN_FRAMES = 1000 / 60;

let total_image_count = 0;
let loaded_image_count = 0;

for (let i = 0; i < Math.random() * 500; ++i) {
  getRandomNumber();
}

/**
 * @param {number} count The total count of images
 */
function onImageCountReady(count) {
  total_image_count = count;
  loaded_image_count = 0;
  displayImageLoadingBar();
}

function onSingleImageLoaded() {
  loaded_image_count++;
  displayImageLoadingBar();
}

function displayImageLoadingBar() {
  context.fillStyle = 'black';
  context.fillRect(0, 0, 800, 600);
  context.fillStyle = 'grey';
  context.fillRect(198, 498, 400, 24);
  context.fillStyle = 'white';
  context.fillRect(200, 500, 400, 20);
  context.font = '50px sans-serif';
  context.textBaseline = 'middle';
  context.fillText(`Loading images... ${loaded_image_count} / ${total_image_count}`, 100, 400, 600);
  context.fillStyle = 'purple';
  context.fillRect(200, 500, (400 * loaded_image_count) / total_image_count, 20);
}

Promise.all([
  loadImageObject(image_descriptions, onImageCountReady, onSingleImageLoaded),
  loadAudioObject(audio_descriptions),
]).then((loading_result) => {
  total_image_count = 0;
  const images = loading_result[0];
  const audio = loading_result[1];

  (function gameLoop() {
    const game_data = (() => {
      const encoded_game_data_str = window.localStorage.getItem('adventurers-journey-save');
      if (encoded_game_data_str) {
        try {
          const decoded_game_data_str = atob(encoded_game_data_str);
          /** @type {import('./types').GameData} */
          const decoded_game_data = JSON.parse(decoded_game_data_str);
          if (decoded_game_data.version === EXPECTED_VERSION) return decoded_game_data;
        } catch {
          /* NO OP */
        }
        window.localStorage.removeItem('adventurers-journey-save');
      }
      return getGameData();
    })();

    const game = new Game(
      context,
      audio,
      game_data,
      [
        getTownSystem(images),
        getDungeonSystem(images),
        getParticlesSystem(),
        getGameOverSystem(images),
        getPopupSystem(),
        getPauseSystem(),
      ],
      [],
    );
    /** @type {any} */
    const modifiable_window = window;
    modifiable_window.clearSavedSession = function () {
      window.localStorage.removeItem('adventurers-journey-save');
    };
    modifiable_window.saveCurrentSession = function () {
      window.localStorage.setItem('adventurers-journey-save', btoa(JSON.stringify(game_data)));
    };
    modifiable_window.showAllBackstories = function () {
      const characters = getAllLiveTransmutedAndDustedCharacters(game_data);
      for (const c of characters) {
        console.log(c.name, generateReadableBackstory(game_data, c));
      }
    };
    modifiable_window.getGame = function () {
      return game_data;
    };

    window.addEventListener('keydown', function (event) {
      game.keyboard.state.shift = event.shiftKey;
      for (const keyName in game.keyboard.config) {
        // Typescript doesn't seem to understand (maybe with reason?) that "keyName" is not a generic string.
        //@ts-ignore
        if (game.keyboard.config[keyName] && game.keyboard.config[keyName].indexOf(event.key) >= 0) {
          //@ts-ignore
          game.keyboard.state[keyName] = true;
        }
      }
    });

    window.addEventListener('keyup', function (event) {
      game.keyboard.state.shift = event.shiftKey;
      for (const keyName in game.keyboard.config) {
        //@ts-ignore
        if (game.keyboard.config[keyName] && game.keyboard.config[keyName].indexOf(event.key) >= 0) {
          //@ts-ignore
          game.keyboard.state[keyName] = false;
        }
      }
    });

    canvas.addEventListener('mousedown', function (event) {
      game.mouse.x = event.offsetX;
      game.mouse.y = event.offsetY;
      if (event.button === 0) game.mouse.primary_button = true;
      if (event.button === 2) game.mouse.secondary_button = true;
    });

    canvas.addEventListener('wheel', function (event) {
      if (event.deltaY > 0) game.mouse.scroll_up = true;
      if (event.deltaY < 0) game.mouse.scroll_down = true;
    });

    canvas.addEventListener('mousemove', function (event) {
      game.mouse.x = event.offsetX;
      game.mouse.y = event.offsetY;
    });

    canvas.addEventListener('mouseup', function (event) {
      game.mouse.x = event.offsetX;
      game.mouse.y = event.offsetY;
      if (event.button === 0) game.mouse.primary_button = false;
      if (event.button === 2) game.mouse.secondary_button = false;
    });

    let lastLoop = Date.now();
    let lastTick = Date.now();
    function singleLoop() {
      const currentLoop = Date.now();
      if (currentLoop > lastTick + MS_BETWEEN_TICKS) {
        const elapsedMsBetweenNowAndThen = currentLoop - lastTick;
        const remainingMs = elapsedMsBetweenNowAndThen % MS_BETWEEN_TICKS;
        const elapsedTicks = (elapsedMsBetweenNowAndThen - remainingMs) / MS_BETWEEN_TICKS;
        game.update(elapsedTicks);
        lastTick = currentLoop - remainingMs;
        game.mouse.scroll_up = false;
        game.mouse.scroll_down = false;
      }
      if (currentLoop > lastLoop + MAX_MS_BETWEEN_FRAMES) {
        const fps = 1000 / (currentLoop - lastLoop);
        game.render(fps);
        lastLoop = currentLoop;
      }
      game.extraEvents();
    }
    function insideLoop() {
      singleLoop();
      if (!game.shouldQuit) {
        window.requestAnimationFrame(insideLoop);
      } else {
        game.showExitScreen();
      }
    }
    window.requestAnimationFrame(insideLoop);
  })();
});
