import { changeCharacterRoleTo, getImageFromCharacter } from '../../data/player-character.mjs';
import { getSkillCharacteristics } from '../../data/player-skills.mjs';
import { getStatusEffectCharacteristics } from '../../data/status-effects.mjs';
import { addExplorationBackstoryEvent } from '../../events/add-backstory-exploration.mjs';
import { getEventCharacterInfo } from '../../events/add-backstory.mjs';
import { generatePlayerParticle } from './particles.mjs';

/**
 * @param {import("../../engine.mjs").Game} game The current game
 * @param {import("../../types").DungeonGameData} data
 * @param {import("../../types").InGameCharacterData} character
 * @param {import("../../types").StatusEffectId} effect
 */
export function applyEffectToPlayer(game, data, character, effect) {
  const characteristics = getStatusEffectCharacteristics(effect);
  if (characteristics.behavior.type === 'game_over') {
    character.status_effects.push({
      id: effect,
      characteristics: getStatusEffectCharacteristics(effect),
      lifetime: { type: 'fixed' },
      current_stack: 1,
    });
    if (characteristics.behavior.new_role) {
      changeCharacterRoleTo(character, characteristics.behavior.new_role);
      const role_change_message = `Class change: ${character.base.role}`;
      generatePlayerParticle(game, character, role_change_message, { r: 128, g: 128, b: 128 });
      if (data.team.length === 1) {
        addExplorationBackstoryEvent(game.game_data, character.base, {
          type: 'alone_class_changed',
          enemies: data.enemy.map((e) => ({ kind: e.kind })),
          new_role: characteristics.behavior.new_role,
        });
      } else {
        addExplorationBackstoryEvent(game.game_data, character.base, {
          type: 'team_class_changed',
          enemies: data.enemy.map((e) => ({ kind: e.kind })),
          group: data.team.filter((t) => t !== character).map((t) => getEventCharacterInfo(character.base, t.base.id)),
          new_role: characteristics.behavior.new_role,
        });
      }
    }
    if (data.team.length === 1) {
      addExplorationBackstoryEvent(game.game_data, character.base, {
        type: 'alone_becomes_critical',
        enemies: data.enemy.map((e) => ({ kind: e.kind })),
        critical: { name: characteristics.name },
      });
    } else {
      const others = data.team.filter((c) => c !== character);
      addExplorationBackstoryEvent(game.game_data, character.base, {
        type: 'team_becomes_critical',
        critical: { name: characteristics.name },
        enemies: data.enemy.map((e) => ({ kind: e.kind })),
        group: data.team.filter((t) => t !== character).map((t) => getEventCharacterInfo(character.base, t.base.id)),
      });
      if (others.length === 1) {
        addExplorationBackstoryEvent(game.game_data, others[0].base, {
          type: 'alone_has_critical_companion_event',
          critical: { name: characteristics.name },
          enemies: data.enemy.map((e) => ({ kind: e.kind })),
          teammate: getEventCharacterInfo(others[0].base, character.base.id),
        });
      } else {
        for (const other of others) {
          addExplorationBackstoryEvent(game.game_data, other.base, {
            type: 'team_has_critical_companion_event',
            critical: { name: characteristics.name },
            enemies: data.enemy.map((e) => ({ kind: e.kind })),
            group: data.team
              .filter((t) => t !== character)
              .map((t) => getEventCharacterInfo(character.base, t.base.id)),
            teammate: getEventCharacterInfo(other.base, character.base.id),
          });
        }
      }
    }
  } else if (characteristics.behavior.type === 'staged_progression') {
    if (characteristics.behavior.resist_skill) {
      const skill_characteristics = getSkillCharacteristics(characteristics.behavior.resist_skill);
      const skill_addition_message = `Skill obtained: ${skill_characteristics.name}`;
      generatePlayerParticle(game, character, skill_addition_message, { r: 128, g: 128, b: 128 });
      character.skills.unshift({
        id: characteristics.behavior.resist_skill,
        characteristics: skill_characteristics,
        cooldown: { remaining_ticks: 0 },
      });
    }
    const first_stack = characteristics.behavior.stages.find((s) => s.stack === 1);
    if (first_stack) {
      character.status_effects.push({
        id: effect,
        characteristics: characteristics,
        current_stack: 1,
        lifetime: {
          type: 'coolup',
          ongoing_ticks: 0,
          max_ticks: first_stack.tick_count,
        },
      });
    } else {
      character.status_effects.push({
        id: effect,
        characteristics: characteristics,
        current_stack: 1,
        lifetime: {
          type: 'cooldown',
          remaining_ticks: 100,
          total_ticks: 100,
        },
      });
    }
  } else if (characteristics.behavior.type === 'cooldown') {
    character.status_effects.push({
      id: effect,
      characteristics: characteristics,
      current_stack: 1,
      lifetime: {
        type: 'cooldown',
        remaining_ticks: characteristics.behavior.tick_count,
        total_ticks: characteristics.behavior.tick_count,
      },
    });
  }
  setImageEffectsToPlayer(game, data, character);
}

/**
 * @param {import("../../engine.mjs").Game} game The current game
 * @param {import("../../types").DungeonGameData} data
 * @param {import("../../types").InGameCharacterData} character
 */
export function setImageEffectsToPlayer(game, data, character) {
  character.image = getImageFromCharacter(character.base, character.equipment);
  for (const effect of character.status_effects) {
    if (effect.lifetime.type === 'to_remove') continue;
    if (effect.characteristics.modify_image.type === 'none') continue;
    if (effect.characteristics.modify_image.type === 'static') {
      character.image = effect.characteristics.modify_image.image;
    }
    if (
      effect.characteristics.modify_image.type === 'per_stack' &&
      effect.characteristics.modify_image.stacks[effect.current_stack]
    ) {
      character.image = effect.characteristics.modify_image.stacks[effect.current_stack];
    }
  }
}

/**
 * @param {import("../../engine.mjs").Game} game The current game
 * @param {import('../../types').DungeonGameData} data
 * @param {import('../../types').InGameCharacterData} character
 * @param {number} elapsedTicks The elapsed ticks since the last update
 */
export function updatePlayerEffectLogic(game, data, character, elapsedTicks) {
  for (const effect of character.status_effects) {
    if (effect.lifetime.type === 'cooldown') {
      effect.lifetime.remaining_ticks -= elapsedTicks;
      if (effect.lifetime.remaining_ticks <= 0) {
        if (effect.current_stack > 1) {
          effect.current_stack--;
          effect.lifetime.remaining_ticks = effect.lifetime.total_ticks;
        } else {
          effect.lifetime = { type: 'to_remove' };
        }
      }
    } else if (effect.lifetime.type === 'coolup') {
      effect.lifetime.ongoing_ticks += elapsedTicks;
      if (effect.lifetime.ongoing_ticks >= effect.lifetime.max_ticks) {
        if (effect.characteristics.behavior.type === 'staged_progression') {
          const stage_information = effect.characteristics.behavior.stages.find(
            (s) => s.stack === effect.current_stack,
          );
          if (stage_information) {
            for (const message of stage_information.next_messages) {
              generatePlayerParticle(game, character, message, { r: 255, g: 0, b: 255 });
            }
            if (stage_information.next.type === 'stack_upgrade') {
              effect.current_stack = stage_information.next.next_stack;
              const next_stage_information = effect.characteristics.behavior.stages.find(
                (s) => s.stack === effect.current_stack,
              );
              if (next_stage_information) {
                effect.lifetime = {
                  type: 'coolup',
                  ongoing_ticks: 0,
                  max_ticks: next_stage_information.tick_count,
                };
              } else {
                effect.lifetime = { type: 'fixed' };
              }
              if (effect.characteristics.modify_image.type === 'static') {
                character.image = effect.characteristics.modify_image.image;
              } else if (
                effect.characteristics.modify_image.type === 'per_stack' &&
                effect.characteristics.modify_image.stacks[effect.current_stack]
              ) {
                character.image = effect.characteristics.modify_image.stacks[effect.current_stack];
              }
            } else if (stage_information.next.type === 'change_status') {
              effect.lifetime = { type: 'to_remove' };
              applyEffectToPlayer(game, data, character, stage_information.next.next_status_effect_id);
            }
          } else {
            effect.lifetime = { type: 'fixed' };
          }
        } else {
          effect.lifetime = { type: 'fixed' };
        }
      }
    }
  }

  for (const effect of character.status_effects) {
    if (effect.lifetime.type !== 'to_remove') continue;
    if (effect.characteristics.behavior.type === 'staged_progression' && effect.characteristics.behavior.resist_skill) {
      const skill_id = effect.characteristics.behavior.resist_skill;
      character.skills = character.skills.filter((s) => s.id !== skill_id);
    }
  }
  character.status_effects = character.status_effects.filter((e) => e.lifetime.type !== 'to_remove');
  setImageEffectsToPlayer(game, data, character);
}
