import { renderButton } from '../../components/button.mjs';
import { showPopupOnMousePosition } from '../../components/popup.mjs';
import { generateEnemiesForDungeon } from '../../data/enemy-generate.mjs';
import { getEventCharacterInfo } from '../../events/add-backstory.mjs';
import {
  addExplorationBackstoryEvent,
  addExplorationBackstoryCompanionsLost,
} from '../../events/add-backstory-exploration.mjs';
import { advanceCycle } from '../../events/advance-cycle.mjs';
import { leaveDungeonForTown } from '../../events/leave-dungeon-for-town.mjs';
import { binaryChoiceResult, getRandomInteger } from '../../random.mjs';
import { display_character, display_equipment } from '../town/utils.mjs';
import { drawCooldownCircle } from './utils.mjs';

const EXPLORATION_CAPTURE_CHANCE = 0.1;

/**
 * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
 */
export function getEnemyRespawnCooldownSystem(images) {
  return {
    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {import('../../types').DungeonGameData} data
     * @param {number} elapsedTicks The elapsed ticks since the last update
     */
    handleEnemyRespawnCooldown: function (game, data, elapsedTicks) {
      if (data.level.is_round_active) {
        handleCharacterCaptureDuringExploration(game, data);
        getRewardIfNeeded(game, data);
        data.level.is_round_active = false;
      }

      if (data.rewards.last_reward_message) {
        game.context.fillStyle = '#AAA';
        game.context.fillRect(540, 50, 220, 60);
        game.context.fillStyle = 'green';
        game.context.font = `20px sans-serif`;
        game.context.fillText('Reward obtained!', 550, 75, 200);
        game.context.fillText(data.rewards.last_reward_message, 550, 100, 200);
      }

      if (getRewardChance(game, data, 'large_group') <= 0) {
        const spawn_small_button = renderButton(game, !game.game_data.pause.is_paused, {
          x: 550,
          y: 150,
          background: { base: '#AAA', hover: '#CCC' },
          content: {
            type: 'fixed_text',
            text: 'Go towards the dungeon',
            font_size: 20,
            width: 170,
            height: 60,
          },
        });
        displayRewardChancePercent(game, { x: 740, y: 155 }, getRewardChance(game, data, 'small_group'));
        if (
          !game.game_data.pause.is_paused &&
          game.mouse.primary_button_clicked_frame &&
          spawn_small_button.is_hovered
        ) {
          startNewRound(game, data, 'small_group');
          return;
        }
      } else {
        const spawn_small_button = renderButton(game, !game.game_data.pause.is_paused, {
          x: 550,
          y: 150,
          background: { base: '#AAA', hover: '#CCC' },
          content: {
            type: 'fixed_text',
            text: 'Stay on the outskirts',
            font_size: 20,
            width: 170,
            height: 40,
          },
        });
        displayRewardChancePercent(game, { x: 730, y: 155 }, getRewardChance(game, data, 'small_group'));
        if (
          !game.game_data.pause.is_paused &&
          game.mouse.primary_button_clicked_frame &&
          spawn_small_button.is_hovered
        ) {
          startNewRound(game, data, 'small_group');
          return;
        }

        const spawn_large_button = renderButton(game, !game.game_data.pause.is_paused, {
          x: 550,
          y: 250,
          background: { base: '#AAA', hover: '#CCC' },
          content: {
            type: 'fixed_text',
            text: 'Explore deep',
            font_size: 20,
            width: 170,
            height: 40,
          },
        });
        displayRewardChancePercent(game, { x: 730, y: 255 }, getRewardChance(game, data, 'large_group'));
        if (
          !game.game_data.pause.is_paused &&
          game.mouse.primary_button_clicked_frame &&
          spawn_large_button.is_hovered
        ) {
          startNewRound(game, data, 'large_group');
          return;
        }
      }

      const leave_button = renderButton(game, !game.game_data.pause.is_paused, {
        x: 550,
        y: 350,
        background: { base: '#AAA', hover: '#CCC' },
        content: {
          type: 'fixed_text',
          text: 'Leave the dungeon',
          font_size: 20,
          width: 170,
          height: 40,
        },
      });
      if (leave_button.is_hovered) {
        showPopupOnMousePosition(
          game,
          'By leaving the dungeon now,\nyou have a 75% chance\nof saving characters\nunder game over effects.',
        );
      }
      if (!game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame && leave_button.is_hovered) {
        leaveDungeonForTown(game.game_data, 0.25, 0.05);
        return;
      }

      if (data.spawn.ambush_cooldown < 0) data.spawn.ambush_cooldown = data.spawn.ambush_time;
      data.spawn.ambush_cooldown -= elapsedTicks;
      drawCooldownCircle(game, 600, 450, 30, data.spawn.ambush_cooldown, data.spawn.ambush_time);
      if (data.spawn.ambush_cooldown <= 0) {
        // Trigger an ambush
        startNewRound(game, data, 'ambush');
        return;
      }

      if (data.rewards.gathered_equipments.length > 0) {
        game.context.fillStyle = 'black';
        game.context.font = `20px sans-serif`;
        game.context.fillText('Gathered loot:', 450, 520);
        displayGatheredLoot(game, data);
      }

      if (data.rewards.saved_characters.length > 0) {
        game.context.fillStyle = 'black';
        game.context.font = `20px sans-serif`;
        game.context.fillText('Saved Characters:', 600, 520);
        displaySavedCharacters(game, data);
      }
    },
  };

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {{x: number, y: number }} position
   * @param {number} reward_chance
   */
  function displayRewardChancePercent(game, position, reward_chance) {
    if (reward_chance <= 0) return;
    game.context.fillStyle = '#777';
    game.context.fillRect(position.x, position.y, 60, 30);
    game.context.fillStyle = 'black';
    game.context.font = `10px sans-serif`;
    game.context.fillText(`Reward chance:`, position.x + 5, position.y + 10, 50);
    game.context.font = `15px sans-serif`;
    game.context.fillStyle = '#7C7';
    const chance_percent_text = `${Math.floor(reward_chance * 100)}%`;
    const chance_percent_text_size = game.context.measureText(chance_percent_text);
    if (chance_percent_text_size.width < 50) {
      game.context.fillText(
        chance_percent_text,
        position.x + 5 + (50 - chance_percent_text_size.width) / 2,
        position.y + 25,
        50,
      );
    } else {
      game.context.fillText(chance_percent_text, position.x + 5, position.y + 25, 50);
    }
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import("../../types").DungeonGameData} data
   */
  function displayGatheredLoot(game, data) {
    let start_x = 450;
    let start_y = 530;
    for (const equipment of data.rewards.gathered_equipments) {
      const size = display_equipment.simplified_slot.content(game, images, equipment, { x: start_x, y: start_y });
      start_x += size.width + 5;
      if (start_x + size.width > 590) {
        start_x = 450;
        start_y += size.height + 5;
      }
    }
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import("../../types").DungeonGameData} data
   */
  function displaySavedCharacters(game, data) {
    let start_x = 600;
    let start_y = 530;
    for (const character of data.rewards.saved_characters) {
      const size = display_character.simplified_slot.content(game, images, character, { x: start_x, y: start_y });
      start_x += size.width + 5;
      if (start_x + size.width > 750) {
        start_x = 600;
        start_y += size.height + 5;
      }
    }
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import("../../types").DungeonGameData} data
   */
  function handleCharacterCaptureDuringExploration(game, data) {
    if (!binaryChoiceResult(EXPLORATION_CAPTURE_CHANCE)) return;
    /** @type {import('../../types').StoredCharacterData[]} */
    const lost_characters = [];
    for (const character of data.team) {
      if (!character.status_effects.some((s) => s.characteristics.behavior.type === 'game_over')) continue;
      lost_characters.push(character.base);
    }
    for (const character of data.rewards.saved_characters) {
      lost_characters.push(character);
    }

    data.team = data.team.filter((c) => !lost_characters.some((o) => o === c.base));
    data.rewards.saved_characters = data.rewards.saved_characters.filter((c) => !lost_characters.some((o) => o === c));

    for (const character of lost_characters) {
      game.game_data.game_state.characters_outside_town.push(character);
    }

    addExplorationBackstoryCompanionsLost(
      game.game_data,
      data.team.map((t) => t.base),
      lost_characters,
      'exploration',
    );
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import("../../types").DungeonGameData} data
   */
  function getRewardIfNeeded(game, data) {
    if (!binaryChoiceResult(data.rewards.current_reward_chance)) return;

    const available_choices =
      game.game_data.game_state.items_outside_town.length + game.game_data.game_state.characters_outside_town.length;
    if (available_choices === 0) return;
    const random_choice = getRandomInteger(0, available_choices);
    if (random_choice < game.game_data.game_state.items_outside_town.length) {
      const equipment = game.game_data.game_state.items_outside_town[random_choice];
      data.rewards.gathered_equipments.push(equipment);
      data.rewards.last_reward_message = `Equipment: ${equipment.name}`;
      game.game_data.game_state.items_outside_town = game.game_data.game_state.items_outside_town.filter(
        (e) => e !== equipment,
      );

      if (data.team.length === 1) {
        addExplorationBackstoryEvent(game.game_data, data.team[0].base, {
          type: 'alone_loot',
          loot: { name: equipment.name, level: equipment.level },
        });
      } else {
        for (const character of data.team) {
          addExplorationBackstoryEvent(game.game_data, character.base, {
            type: 'team_loot',
            group: data.team
              .filter((c) => c !== character)
              .map((c) => getEventCharacterInfo(character.base, c.base.id)),
            loot: { name: equipment.name, level: equipment.level },
          });
        }
      }
    } else {
      const saved_character =
        game.game_data.game_state.characters_outside_town[
          random_choice - game.game_data.game_state.items_outside_town.length
        ];
      data.rewards.saved_characters.push(saved_character);
      data.rewards.last_reward_message = `Freed: ${saved_character.name} (lv. ${saved_character.level})`;
      game.game_data.game_state.characters_outside_town = game.game_data.game_state.characters_outside_town.filter(
        (c) => c !== saved_character,
      );

      if (data.team.length === 1) {
        addExplorationBackstoryEvent(game.game_data, saved_character, {
          type: 'alone_becomes_rescued',
          rescuer: getEventCharacterInfo(saved_character, data.team[0].base.id),
        });
        addExplorationBackstoryEvent(game.game_data, data.team[0].base, {
          type: 'alone_rescue',
          rescuee: getEventCharacterInfo(data.team[0].base, saved_character.id),
        });
      } else {
        addExplorationBackstoryEvent(game.game_data, saved_character, {
          type: 'team_becomes_rescued',
          group: data.team.map((t) => getEventCharacterInfo(saved_character, t.base.id)),
        });
        for (const teammmate of data.team) {
          addExplorationBackstoryEvent(game.game_data, teammmate.base, {
            type: 'team_rescue',
            group: data.team
              .filter((t) => t !== teammmate)
              .map((t) => getEventCharacterInfo(teammmate.base, t.base.id)),
            rescuee: getEventCharacterInfo(teammmate.base, saved_character.id),
          });
        }
      }
    }
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import("../../types").DungeonGameData} data
   * @param {"small_group" | "large_group" | "ambush"} mode
   */
  function startNewRound(game, data, mode) {
    const level = getLevelForRound(game, data, mode);
    data.enemy = generateEnemiesForDungeon(data, level, mode);
    data.rewards.current_reward_chance = getRewardChance(game, data, mode);
    data.rewards.last_reward_message = '';
    data.spawn.ambush_cooldown = -1;
    data.level.is_round_active = true;
    data.level.exploration_cycles++;
    advanceCycle(game.game_data, { advance_mode: 'dungeon_fight', fight_mode: mode });
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import("../../types").DungeonGameData} data
   * @param {"small_group" | "large_group" | "ambush"} mode
   */
  function getLevelForRound(game, data, mode) {
    const max_team_level = Math.max(0, ...data.team.map((p) => p.base.level));
    if (mode === 'small_group') {
      const level_multiplier = Math.min(0.75 + data.level.exploration_cycles * 0.05, 1.25);
      return Math.min(Math.ceil(max_team_level * level_multiplier), max_team_level + 5);
    }
    if (mode === 'large_group') {
      const level_multiplier = 0.75 + (data.level.exploration_cycles + 2) * 0.06;
      return Math.min(Math.ceil(max_team_level * level_multiplier), max_team_level + 10);
    }
    return max_team_level + 2;
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import("../../types").DungeonGameData} data
   * @param {"small_group" | "large_group" | "ambush"} mode
   * @returns {number}
   */
  function getRewardChance(game, data, mode) {
    if (data.level.exploration_cycles > 2) {
      const level_ratio = Math.min(4, Math.pow(data.level.exploration_cycles - 2, 0.5));
      if (mode === 'small_group') return 0.05 * level_ratio;
      if (mode === 'large_group') return 0.2 * level_ratio;
    }
    return 0;
  }
}
