import { changeCharacterLevelTo, getRequiredExperienceForPlayerLevel } from '../../data/player-character.mjs';
import { getRandomInArray } from '../../random.mjs';
import { assertTypeIsHandled } from '../../utils.mjs';
import { generatePlayerParticle } from './particles.mjs';
import { activateSkill } from './skills.mjs';
import { drawBar, getEnemyPosition } from './utils.mjs';

/**
 * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
 */
export function getEnemyDungeonSystem(images) {
  return {
    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {import("../../types").DungeonGameData} data
     * @param {import("../../types").InGameEnemyData} enemy
     * @param {number} elapsedTicks The elapsed ticks since the last update
     */
    displayEnemy: function (game, data, enemy, elapsedTicks) {
      const position = getEnemyPosition(game, enemy);
      if (enemy.animation.current === 'attack' && images[enemy.image + '_attack']) {
        game.renderer.renderRawImage(images[enemy.image + '_attack'], position.x - 100, position.y + 50, 200, 150);
        enemy.animation.remaining_frames--;
        if (enemy.animation.remaining_frames <= 0) enemy.animation = { current: 'none' };
      } else {
        game.renderer.renderRawImage(images[enemy.image], position.x - 100, position.y + 50, 200, 150);
      }
      {
        // Enemy's name and level
        const text = `${enemy.name} lv. ${enemy.level}`;
        game.context.fillStyle = 'black';
        game.context.font = `15px sans-serif`;
        game.context.fillText(text, position.x, position.y + 30);
      }
      {
        // Enemy's HP Bar
        const text = `HP: ${Math.ceil(enemy.current_stats.hp)} / ${Math.ceil(enemy.base_stats.max_hp)}`;
        const percent = enemy.current_stats.hp / enemy.base_stats.max_hp;
        drawBar(game, { x: position.x, y: position.y + 40, width: 140, height: 20 }, percent, '#E33', 'grey', text);
      }
    },

    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {import("../../types").DungeonGameData} data
     * @param {import("../../types").InGameEnemyData} enemy
     * @param {number} elapsedTicks The elapsed ticks since the last update
     */
    handleEnemyLogic: function (game, data, enemy, elapsedTicks) {
      // Logic: Reduce skill cooldowns
      for (const skill of enemy.skills) {
        if (skill.cooldown.remaining_ticks <= 0) continue;
        skill.cooldown.remaining_ticks -= elapsedTicks;
        if (skill.cooldown.remaining_ticks <= 0) skill.cooldown.remaining_ticks = 0;
      }

      // Logic: Attack player
      const possible_attacks = getPossibleAttacks(game, data, enemy);
      if (possible_attacks.length > 0) {
        const decision = getRandomInArray(possible_attacks);
        if (decision.attack.cooldown.remaining_ticks <= 0) {
          activateSkill(game, data, enemy, decision.targets, decision.attack);
        }
      }
    },

    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {import("../../types").DungeonGameData} data
     * @param {import("../../types").InGameEnemyData} enemy
     * @param {number} elapsedTicks The elapsed ticks since the last update
     */
    handleEnemyDeath: function (game, data, enemy, elapsedTicks) {
      // Logic: Handle enemy death.
      if (enemy.current_stats.hp === 0) destroyEnemy(game, data, enemy);
    },
  };

  /**
   *
   * @param {Array<import('../../types').InGameCharacterData>} raw_targets
   * @param {import('../../types').InGameEnemySkill} skill
   * @returns {Array<import('../../types').InGameCharacterData>}
   */
  function getValidTargetsForSkill(raw_targets, skill) {
    const result = [];
    for (const target of raw_targets) {
      if (skill.characteristics.action.type === 'apply_effect') {
        // HACK: prevent stacking of multiple status effects
        if (target.status_effects.length > 0) continue;
      }
      if (skill.condition.type === 'no_condition') {
        result.push(target);
        continue;
      }
      if (skill.condition.type === 'target_not_critical') {
        if (target.type !== 'character') continue;
        if (target.status_effects.some((s) => s.characteristics.behavior.type === 'game_over')) continue;
        if (target.current_stats.hp <= 1) continue;
        result.push(target);
        continue;
      }
      if (skill.condition.type === 'target_vulnerable') {
        if (target.type !== 'character') continue;
        if (target.status_effects.some((s) => s.characteristics.behavior.type === 'game_over')) continue;
        if (target.current_stats.hp >= skill.condition.vulnerability_ratio * target.base_stats.max_hp) continue;
        result.push(target);
        continue;
      }
      if (skill.condition.type === 'target_critical') {
        if (target.type !== 'character') continue;
        if (target.status_effects.some((s) => s.characteristics.behavior.type === 'game_over')) continue;
        if (target.current_stats.hp > 1) continue;
        result.push(target);
        continue;
      }
      if (skill.condition.type === 'target_game_over') {
        if (target.type !== 'character') continue;
        if (!target.status_effects.some((s) => s.characteristics.behavior.type === 'game_over')) continue;
        result.push(target);
        continue;
      }
      assertTypeIsHandled(skill.condition, 'Target condition');
    }

    return result;
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import("../../types").DungeonGameData} data
   * @param {import("../../types").InGameEnemyData} enemy
   * @param {import("../../types").InGameEnemySkill} skill
   * @returns {Array<import('../../types').InGameCharacterData | import('../../types').InGameEnemyData>}
   */
  function getEnemyTargetsFromSkill(game, data, enemy, skill) {
    if (skill.characteristics.target.type === 'all_allies') {
      return data.enemy;
    }
    if (skill.characteristics.target.type === 'self') {
      return [enemy];
    }
    const available_targets = getValidTargetsForSkill(data.team, skill);
    if (skill.characteristics.target.type === 'all_enemies') {
      return available_targets;
    }
    if (skill.characteristics.target.type === 'random_enemy') {
      if (available_targets.length === 0) return [];
      return [getRandomInArray(available_targets)];
    }
    if (skill.characteristics.target.type === 'first_enemy') {
      if (available_targets.length === 0) return [];
      return [available_targets[0]];
    }
    if (skill.characteristics.target.type === 'last_enemy') {
      if (available_targets.length === 0) return [];
      return [available_targets[available_targets.length - 1]];
    }
    if (skill.characteristics.target.type === 'lowest_hp_enemy') {
      const lowest_hp = available_targets.reduce((p, e) => Math.min(p, e.current_stats.hp), 10000);
      const target = available_targets.find((e) => e.current_stats.hp <= lowest_hp);
      if (target) return [target];
      return [];
    }
    if (skill.characteristics.target.type === 'highest_hp_enemy') {
      const highest_hp = available_targets.reduce((p, e) => Math.max(p, e.current_stats.hp), 0);
      const target = available_targets.find((e) => e.current_stats.hp >= highest_hp);
      if (target) return [target];
      return [];
    }
    assertTypeIsHandled(skill.characteristics.target, 'skill target');
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import("../../types").DungeonGameData} data
   * @param {import("../../types").InGameEnemyData} enemy
   * @returns { Array<{ attack: import('../../types').InGameEnemySkill, targets: Array<import('../../types').InGameCharacterData | import('../../types').InGameEnemyData> }> }
   */
  function getPossibleAttacks(game, data, enemy) {
    /** @type { Array<{ attack: import('../../types').InGameEnemySkill, targets: Array<import('../../types').InGameCharacterData | import('../../types').InGameEnemyData> }> } */
    const possible_attacks = [];
    for (const attack of enemy.skills) {
      const possible_targets = getEnemyTargetsFromSkill(game, data, enemy, attack);
      if (possible_targets.length === 0) continue;
      possible_attacks.push({ attack: attack, targets: possible_targets });
    }
    return possible_attacks;
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import("../../types").DungeonGameData} data
   * @param {import("../../types").InGameEnemyData} enemy
   */
  function destroyEnemy(game, data, enemy) {
    for (const player of data.team) {
      player.base.partial_exp += enemy.exp_value;
      let has_leveled_up = false;
      while (player.base.partial_exp >= getRequiredExperienceForPlayerLevel(player.base.level)) {
        player.base.partial_exp -= getRequiredExperienceForPlayerLevel(player.base.level);
        changeCharacterLevelTo(player, player.base.level + 1);
        has_leveled_up = true;
      }
      if (has_leveled_up) {
        generatePlayerParticle(game, player, `Level up! (level ${player.base.level})`, { r: 128, g: 128, b: 128 });
      } else {
        generatePlayerParticle(game, player, `Exp +${enemy.exp_value}`, { r: 0, g: 128, b: 255 });
      }
    }
    data.enemy = data.enemy.filter((e) => e !== enemy);
  }
}
