import { showPopupOnMousePosition } from '../../components/popup.mjs';
import { getEquipmentDescription } from '../../data/equipment.mjs';
import { getRequiredExperienceForPlayerLevel } from '../../data/player-character.mjs';
import { getSkillDescriptionForInGameCharacter } from '../../data/player-skills.mjs';
import { getStatusEffectDescription } from '../../data/status-effects.mjs';
import { getRandomInArray } from '../../random.mjs';
import { assertTypeIsHandled } from '../../utils.mjs';
import { applyEffectToPlayer, updatePlayerEffectLogic } from './effects.mjs';
import { generatePlayerParticle, generateRisingParticleAt } from './particles.mjs';
import { activateSkill } from './skills.mjs';
import { drawBar, drawCooldownCircle, drawIndicatorCircle, getCharacterPosition, pointIsInArea } from './utils.mjs';

/**
 * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
 */
export function getPlayerDungeonSystem(images) {
  return {
    /**
     * @param { import("../../engine.mjs").Game } game The current game
     * @param { import("../../types").DungeonGameData } data
     * @param { import("../../types").InGameCharacterData } character
     * @param {number} elapsedTicks The elapsed ticks since the last update
     */
    handleCharacterLogic: function (game, data, character, elapsedTicks) {
      // Logic: Regen player MP
      if (character.current_stats.hp > 0) {
        character.current_stats.mp += elapsedTicks * character.base_stats.mp_regen;
        if (character.current_stats.mp > character.base_stats.max_mp) {
          character.current_stats.mp = character.base_stats.max_mp;
        }
      }

      updatePlayerEffectLogic(game, data, character, elapsedTicks);
    },
    /**
     * @param { import("../../engine.mjs").Game } game The current game
     * @param { import("../../types").DungeonGameData } data
     * @param { import("../../types").InGameCharacterData } character
     * @param { number } elapsedTicks The elapsed ticks since the last update
     */
    displayCharacter: function (game, data, character, elapsedTicks) {
      const position = getCharacterPosition(game, character);
      if (character.animation.current === 'attack' && images[character.image + '_angry_attack']) {
        game.renderer.renderRawImage(images[character.image + '_angry_attack'], position.x, position.y + 150, 200, 150);
        character.animation.remaining_frames--;
        if (character.animation.remaining_frames <= 0) character.animation = { current: 'none' };
      } else if (game.game_data.dungeon.enemy.length && images[character.image + '_angry']) {
        game.renderer.renderRawImage(images[character.image + '_angry'], position.x, position.y + 150, 200, 150);
      } else {
        game.renderer.renderRawImage(images[character.image], position.x, position.y + 150, 200, 150);
      }
      {
        // Player's name and level
        game.context.fillStyle = 'black';
        game.context.font = `12px sans-serif`;
        game.context.fillText(`lv. ${character.base.level} ${character.base.role}`, position.x + 10, position.y + 15);
        game.context.font = `15px sans-serif`;
        game.context.fillText(`${character.base.name}`, position.x + 10, position.y + 30);
      }
      {
        // Player's HP Bar
        const text = `HP ${Math.floor(character.current_stats.hp)} / ${Math.ceil(character.base_stats.max_hp)}`;
        const percent = character.current_stats.hp / character.base_stats.max_hp;
        drawBar(
          game,
          { x: position.x + 10, y: position.y + 40, width: 100, height: 16 },
          percent,
          '#E33',
          'grey',
          text,
        );
      }
      {
        // Player's MP Bar
        const text = `MP ${Math.floor(character.current_stats.mp)} / ${Math.ceil(character.base_stats.max_mp)}`;
        const percent = character.current_stats.mp / character.base_stats.max_mp;
        drawBar(
          game,
          { x: position.x + 10, y: position.y + 60, width: 100, height: 16 },
          percent,
          'blue',
          'grey',
          text,
        );
      }
      {
        // Player's Exp Bar
        const next_level_exp = getRequiredExperienceForPlayerLevel(character.base.level);
        const text = `exp ${Math.ceil(character.base.partial_exp)} / ${next_level_exp}`;
        const percent = character.base.partial_exp / next_level_exp;
        drawBar(game, { x: position.x + 10, y: position.y + 80, width: 75, height: 8 }, percent, '#0AA', 'grey', text);
      }
      handlePlayerEquipment(game, data, character, position, character.equipment.body, 'body', elapsedTicks);
      handlePlayerEquipment(game, data, character, position, character.equipment.accessory, 'accessory', elapsedTicks);
      handlePlayerEquipment(game, data, character, position, character.equipment.weapon, 'weapon', elapsedTicks);
      handlePlayerEffects(game, data, character, position, elapsedTicks);
      handlePlayerSkills(game, data, character, position, elapsedTicks);
    },
  };

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').DungeonGameData} data
   * @param {import('../../types').InGameCharacterData} character
   * @param {{ x: number, y: number }} position
   * @param {import('../../types').StoredEquipmentData | undefined} equipment
   * @param {import('../../types').StoredEquipmentData["type"]} equipment_type
   * @param {number} elapsedTicks The elapsed ticks since the last update
   */
  function handlePlayerEquipment(game, data, character, position, equipment, equipment_type, elapsedTicks) {
    const start_x =
      position.x +
      (() => {
        if (equipment_type === 'body') return 5;
        if (equipment_type === 'accessory') return 40;
        return 75;
      })();
    const start_y = position.y + 90;
    game.context.fillStyle = '#AAA';
    game.context.fillRect(start_x, start_y, 30, 30);
    if (equipment) {
      game.renderer.renderRawImage(images[equipment.image], start_x, start_y, 30, 30);

      let is_hovering = pointIsInArea(
        { x: start_x, y: start_y, width: 30, height: 30 },
        { x: game.mouse.x, y: game.mouse.y },
      );
      if (is_hovering) showPopupOnMousePosition(game, getEquipmentDescription(equipment));
    }
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').DungeonGameData} data
   * @param {import('../../types').InGameCharacterData} character
   * @param {{ x: number, y: number }} position
   * @param {number} elapsedTicks The elapsed ticks since the last update
   */
  function handlePlayerEffects(game, data, character, position, elapsedTicks) {
    let start_x = position.x;
    let start_y = 230;
    for (const effect of character.status_effects) {
      const current_y = start_y;
      const current_x = start_x + 10;
      game.context.font = `12px sans-serif`;
      const text_size = game.context.measureText(effect.characteristics.name);
      const width = text_size.width + 20;
      const height = 20;
      let is_hovering = pointIsInArea(
        { x: current_x, y: current_y, width: width, height: height },
        { x: game.mouse.x, y: game.mouse.y },
      );
      game.context.fillStyle = '#AAA';
      game.context.fillRect(current_x, current_y, width, height);
      if (effect.lifetime.type === 'cooldown') {
        drawCooldownCircle(
          game,
          current_x + width,
          current_y + height,
          8,
          effect.lifetime.remaining_ticks,
          effect.lifetime.total_ticks,
        );
      }
      if (effect.lifetime.type === 'coolup') {
        drawCooldownCircle(
          game,
          current_x + width,
          current_y + height,
          8,
          effect.lifetime.max_ticks - effect.lifetime.ongoing_ticks,
          effect.lifetime.max_ticks,
        );
      }

      if (effect.current_stack > 1) {
        drawIndicatorCircle(game, current_x + width, current_y, 8, `${effect.current_stack}`, '#CCC', 'white');
      }

      game.context.font = `12px sans-serif`;
      game.context.fillStyle = 'black';
      game.context.fillText(effect.characteristics.name, current_x + 10, current_y + 14);

      if (is_hovering) showPopupOnMousePosition(game, getStatusEffectDescription(effect));

      start_x += width + 20;
    }
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').DungeonGameData} data
   * @param {import('../../types').InGameCharacterData} character
   * @param {{ x: number, y: number }} position
   * @param {number} elapsedTicks The elapsed ticks since the last update
   */
  function handlePlayerSkills(game, data, character, position, elapsedTicks) {
    let start_x = position.x;
    let start_y = 400;
    for (const skill of character.skills) {
      const current_x = start_x + 10;
      const current_y = start_y;
      game.context.font = `14px sans-serif`;
      const text_size = game.context.measureText(skill.characteristics.name);
      const width = text_size.width + 30;
      const height = 20;
      /** @type {{ state: "valid" } | { state: "blocked", reason: "mp" | "effect" | "cooldown", message: string }} */
      const skill_status = (() => {
        for (const effect of character.status_effects) {
          const blockers = (() => {
            if (effect.characteristics.block_skills.type === 'static') {
              return effect.characteristics.block_skills.blocks;
            }
            return effect.characteristics.block_skills.stacks[effect.current_stack];
          })();
          const effect_blocks_skill =
            blockers.type === 'all' ||
            (blockers.type === 'exact' && blockers.skills.includes(skill.id)) ||
            (blockers.type === 'all_except_exact' && !blockers.skills.includes(skill.id));
          if (effect_blocks_skill)
            return { state: 'blocked', reason: 'effect', message: `Skill blocked by ${effect.characteristics.name}` };
        }
        const mana_cost =
          skill.characteristics.cost_data.mana +
          character.base_stats.max_mp * skill.characteristics.cost_data.mana_percent;
        if (character.current_stats.mp < mana_cost)
          return {
            state: 'blocked',
            reason: 'mp',
            message: `Not enough mana! ${mana_cost} required`,
          };
        if (skill.cooldown.remaining_ticks > 0)
          return { state: 'blocked', reason: 'cooldown', message: `On cooldown!` };
        return { state: 'valid' };
      })();

      const is_hovering = pointIsInArea(
        { x: current_x, y: current_y, width: width, height: height },
        { x: game.mouse.x, y: game.mouse.y },
      );

      const background_color = (() => {
        if (skill_status.state === 'blocked' && skill_status.reason === 'mp') return '#77C';
        if (skill_status.state === 'blocked' && skill_status.reason === 'effect') return '#C77';
        if (skill_status.state === 'blocked') return '#777';
        if (is_hovering) return '#CCC';
        return '#AAA';
      })();

      game.context.fillStyle = background_color;
      game.context.fillRect(current_x, current_y, width, height);
      if (skill.cooldown.remaining_ticks > 0) {
        drawCooldownCircle(
          game,
          current_x + width,
          current_y + height,
          10,
          skill.cooldown.remaining_ticks,
          skill.characteristics.cost_data.cooldown,
        );
      }
      game.context.font = `14px sans-serif`;
      game.context.fillStyle = 'black';
      game.context.fillText(skill.characteristics.name, current_x + 16, current_y + 16);
      start_y += height + 5;

      if (is_hovering)
        showPopupOnMousePosition(game, getSkillDescriptionForInGameCharacter(character.base_stats, skill.id));

      if (skill.cooldown.remaining_ticks > 0) {
        skill.cooldown.remaining_ticks = Math.max(0, skill.cooldown.remaining_ticks - elapsedTicks);
      }

      const user_clicked_on_spell =
        is_hovering && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame;
      if (user_clicked_on_spell) {
        if (skill_status.state === 'valid') {
          if (
            skill.cooldown.remaining_ticks === 0 &&
            character.current_stats.mp > skill.characteristics.cost_data.mana
          ) {
            const targets = getPlayerTargetsFromSkill(game, data, character, skill);
            activateSkill(game, data, character, targets, skill);
          }
          game.sound_queue.push({ sound: 'click' });
        } else {
          generateRisingParticleAt(
            game,
            { x: current_x + width / 2, y: current_y + height / 2 },
            skill_status.message,
            { r: 255, g: 0, b: 0 },
            'centered',
          );
          game.sound_queue.push({ sound: 'error' });
        }
      }
    }
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import("../../types").DungeonGameData} data
   * @param {import("../../types").InGameCharacterData} character
   * @param {import("../../types").InGameCharacterSkill} skill
   * @returns {Array<import('../../types').InGameCharacterData | import('../../types').InGameEnemyData>}
   */
  function getPlayerTargetsFromSkill(game, data, character, skill) {
    if (skill.characteristics.target.type === 'all_allies') {
      return data.team;
    }
    if (skill.characteristics.target.type === 'self') {
      return [character];
    }
    if (skill.characteristics.target.type === 'all_enemies') {
      return data.enemy;
    }
    if (skill.characteristics.target.type === 'random_enemy') {
      if (data.enemy.length === 0) return [];
      return [getRandomInArray(data.enemy)];
    }
    if (skill.characteristics.target.type === 'first_enemy') {
      if (data.enemy.length === 0) return [];
      return [data.enemy[0]];
    }
    if (skill.characteristics.target.type === 'last_enemy') {
      if (data.enemy.length === 0) return [];
      return [data.enemy[data.enemy.length - 1]];
    }
    if (skill.characteristics.target.type === 'lowest_hp_enemy') {
      const lowest_hp = data.enemy.reduce((p, e) => Math.min(p, e.current_stats.hp), 10000);
      const target = data.enemy.find((e) => e.current_stats.hp <= lowest_hp);
      if (target) return [target];
      return [];
    }
    if (skill.characteristics.target.type === 'highest_hp_enemy') {
      const highest_hp = data.enemy.reduce((p, e) => Math.max(p, e.current_stats.hp), 0);
      const target = data.enemy.find((e) => e.current_stats.hp >= highest_hp);
      if (target) return [target];
      return [];
    }
    assertTypeIsHandled(skill.characteristics.target, 'skill target');
  }
}
