import { changeCharacterLevelTo, getRequiredExperienceForPlayerLevel } from '../../data/player-character.mjs';
import { getSkillActualCooldown, getSkillActualManaCost } from '../../data/player-skills.mjs';
import { assertTypeIsHandled } from '../../utils.mjs';
import { applyEffectToPlayer } from './effects.mjs';
import { generateEnemyParticle, generatePlayerParticle } from './particles.mjs';

/**
 * @param {import("../../engine.mjs").Game} game The current game
 * @param {import("../../types").DungeonGameData} data
 * @param {import("../../types").InGameCharacterData | import('../../types').InGameEnemyData} caster
 * @param {Array<import('../../types').InGameCharacterData | import('../../types').InGameEnemyData>} targets
 * @param {import("../../types").InGameCharacterSkill} skill
 */
export function activateSkill(game, data, caster, targets, skill) {
  const skill_effect = skill.characteristics.action;
  const mana_cost = getSkillActualManaCost(caster.base_stats, skill.characteristics);
  caster.current_stats.mp = Math.max(0, caster.current_stats.mp - mana_cost);
  const cooldown = getSkillActualCooldown(caster.base_stats, skill.characteristics);
  skill.cooldown.remaining_ticks = cooldown;
  if (skill_effect.type === 'damage') {
    const damage_hp = skill_effect.damage_hp + caster.base_stats.attack_additional_hp;
    for (const target of targets) {
      target.current_stats.hp = Math.max(0, target.current_stats.hp - damage_hp);
      caster.animation = { current: 'attack', remaining_frames: 5 };
      generateTargetParticle(game, target, `HP -${damage_hp}`, { r: 255, g: 0, b: 0 });
    }
    return;
  }
  if (skill_effect.type === 'damage_scaling_magic_cost') {
    const damage_hp = skill_effect.ratio * caster.base_stats.max_mp + caster.base_stats.attack_additional_hp;
    for (const target of targets) {
      target.current_stats.hp = Math.max(0, target.current_stats.hp - damage_hp);
      caster.animation = { current: 'attack', remaining_frames: 5 };
      generateTargetParticle(game, target, `HP -${damage_hp}`, { r: 255, g: 0, b: 0 });
    }
    return;
  }
  if (skill_effect.type === 'heal') {
    const heal_hp = skill_effect.heal_hp + caster.base_stats.heal_additional_hp;
    for (const target of targets) {
      target.current_stats.hp = Math.min(target.current_stats.hp + heal_hp, target.base_stats.max_hp);
      generateTargetParticle(game, target, `HP +${heal_hp}`, { r: 128, g: 255, b: 128 });
    }
    return;
  }
  if (skill_effect.type === 'drain_hp') {
    const damage_hp = skill_effect.damage_hp + caster.base_stats.attack_additional_hp;
    let healed_hp = 0;
    for (const target of targets) {
      const hp_after_attack = Math.max(0, target.current_stats.hp - damage_hp);
      healed_hp += target.current_stats.hp - hp_after_attack;
      target.current_stats.hp = hp_after_attack;
      caster.animation = { current: 'attack', remaining_frames: 5 };
      generateTargetParticle(game, target, `HP -${damage_hp}`, { r: 255, g: 0, b: 0 });
    }
    healed_hp *= skill_effect.drain_ratio;
    if (healed_hp <= 0) return;
    const hp_after = Math.min(
      caster.current_stats.hp + healed_hp + caster.base_stats.heal_additional_hp,
      caster.base_stats.max_hp,
    );
    if (hp_after <= caster.current_stats.hp) return;
    generateTargetParticle(game, caster, `HP +${Math.floor(healed_hp)}`, { r: 128, g: 255, b: 128 });
    caster.current_stats.hp = hp_after;
    return;
  }
  if (skill_effect.type === 'drain_mana') {
    const damage_mp = skill_effect.damage_mp;
    let healed_mp = 0;
    for (const target of targets) {
      const mp_after_attack = Math.max(0, target.current_stats.mp - damage_mp);
      healed_mp += target.current_stats.mp - mp_after_attack;
      target.current_stats.mp = mp_after_attack;
      caster.animation = { current: 'attack', remaining_frames: 5 };
      generateTargetParticle(game, target, `MP -${damage_mp}`, { r: 0, g: 0, b: 255 });
    }
    healed_mp *= skill_effect.drain_ratio;
    if (healed_mp <= 0) return;
    const mp_after = Math.min(caster.current_stats.mp + healed_mp, caster.base_stats.max_mp);
    if (mp_after <= caster.current_stats.mp) return;
    generateTargetParticle(game, caster, `MP +${Math.floor(healed_mp)}`, { r: 128, g: 128, b: 255 });
    caster.current_stats.mp = mp_after;
    return;
  }
  if (skill_effect.type === 'drain_experience') {
    for (const target of targets) {
      if (target.type !== 'character') continue;
      const message = `Exp drained! -${skill_effect.damage_experience}XP`;
      generatePlayerParticle(game, target, message, { r: 0, g: 127, b: 255 });
      target.base.partial_exp -= skill_effect.damage_experience;
      while (target.base.partial_exp < 0) {
        if (target.base.level === 1) break;
        changeCharacterLevelTo(target, target.base.level - 1);
        target.base.partial_exp += getRequiredExperienceForPlayerLevel(target.base.level);
        const leveldown_message = `Level lost! (level ${target.base.level})`;
        generatePlayerParticle(game, target, leveldown_message, { r: 128, g: 128, b: 128 });
      }
      if (target.base.partial_exp <= 0) target.base.partial_exp = 0;
    }
    return;
  }
  if (skill_effect.type === 'apply_effect') {
    for (const target of targets) {
      if (target.type !== 'character') continue;
      generateTargetParticle(game, target, 'Limbs turning into stone!', { r: 255, g: 0, b: 255 });
      applyEffectToPlayer(game, data, target, skill_effect.effect);
    }
    return;
  }
  if (skill_effect.type === 'dispel_all_effects') {
    for (const target of targets) {
      if (target.type !== 'character') continue;
      let has_dispelled = false;
      for (const effect of target.status_effects) {
        if (effect.lifetime.type === 'coolup' || effect.lifetime.type === 'cooldown') {
          effect.lifetime = { type: 'to_remove' };
          has_dispelled = true;
        }
      }
      if (has_dispelled) {
        generatePlayerParticle(game, target, `Dispel!`, { r: 128, g: 255, b: 128 });
      }
    }
    return;
  }
  if (skill_effect.type === 'dispel_effect') {
    for (const target of targets) {
      if (target.type !== 'character') continue;
      let has_dispelled = false;
      for (const effect of target.status_effects) {
        if (effect.id !== skill_effect.effect) continue;
        if (effect.lifetime.type === 'coolup' || effect.lifetime.type === 'cooldown') {
          effect.lifetime = { type: 'to_remove' };
          has_dispelled = true;
        }
      }
      if (has_dispelled) {
        generatePlayerParticle(game, target, `Dispel!`, { r: 128, g: 255, b: 128 });
      }
    }
    return;
  }
  if (skill_effect.type === 'resist_effect') {
    for (const target of targets) {
      if (target.type !== 'character') continue;
      for (const effect of target.status_effects) {
        if (effect.id !== skill_effect.effect) continue;
        if (effect.lifetime.type !== 'coolup') continue;
        effect.lifetime = {
          type: 'cooldown',
          total_ticks: effect.lifetime.max_ticks + effect.lifetime.ongoing_ticks * 2,
          remaining_ticks: effect.lifetime.max_ticks + effect.lifetime.ongoing_ticks * 2,
        };
        generatePlayerParticle(game, target, `Resist!`, { r: 0, g: 255, b: 255 });
      }
    }
    return;
  }
  assertTypeIsHandled(skill_effect, 'Skill effect');
}

/**
 * @param {import("../../engine.mjs").Game} game
 * @param {import("../../types").InGameCharacterData | import('../../types').InGameEnemyData} target
 * @param {string} text
 * @param {{ r: number, g: number, b: number }} color
 */
function generateTargetParticle(game, target, text, color) {
  if (target.type === 'enemy') {
    generateEnemyParticle(game, target, text, color);
    return;
  }
  if (target.type === 'character') {
    generatePlayerParticle(game, target, text, color);
    return;
  }
  assertTypeIsHandled(target, 'skill target type');
}
