/**
 * @typedef { import("../engine.mjs").Game } Game
 * @typedef { import("../types").GameData } GameData
 */

import { renderButton } from '../components/button.mjs';
import { mouseIsInArea } from '../components/checks.mjs';
import { showPopupOnMousePosition } from '../components/popup.mjs';
import { showScrollableHorizontalList } from '../components/scroll-horizontal.mjs';
import { showScrollableTextBox } from '../components/scroll-vertical-text.mjs';
import { generateReadableBackstory } from '../data/backstory-readable.mjs';
import { getFaceImageFromTransmutedCharacter } from '../data/equipment.mjs';
import { getCharacterDescription, getFaceImageFromCharacter } from '../data/player-character.mjs';
import { getAllEquipments, getAllLiveCharacters } from '../data/utils.mjs';
import { advanceCycle } from '../events/advance-cycle.mjs';
import { leaveDungeonForTown } from '../events/leave-dungeon-for-town.mjs';

/**
 * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
 */
export function getGameOverSystem(images) {
  /**
   * @param {Game} game The current game
   * @param {number} elapsedTicks The elapsed ticks since the last update
   */
  return function gameOverSystem(game, elapsedTicks) {
    if (game.game_data.game_state.game_over_cooldown < 0) return;
    handleGameOver(game, game.game_data, elapsedTicks);
  };

  /**
   * @param {Game} game The current game
   * @param {GameData} data
   * @param {number} elapsedTicks The elapsed ticks since the last update
   */
  function handleGameOver(game, data, elapsedTicks) {
    data.game_state.game_over_cooldown -= elapsedTicks;
    if (data.game_state.game_over_cooldown <= 0) data.game_state.game_over_cooldown = 0;

    const town = game.game_data.town;
    const characters_in_town =
      town.available_team.length +
      (town.upgrade_equipment.active && town.upgrade_equipment.handler ? 1 : 0) +
      (town.change_role.active ? 1 : 0);
    if (characters_in_town > 0) {
      showExpeditionFailedScreen(game, data);
      return;
    }

    showGameOverScreen(game, data);
  }

  /**
   * @param {Game} game The current game
   * @param {GameData} data
   */
  function showExpeditionFailedScreen(game, data) {
    const alpha = Math.max(0.0, Math.min(0.9, 1.0 - data.game_state.game_over_cooldown / 100));
    game.context.fillStyle = `rgba(0, 0, 0, ${alpha})`;
    game.context.fillRect(0, 0, 800, 600);
    game.context.fillStyle = `rgba(255, 0, 0, ${alpha})`;
    game.context.font = '40px sans-serif';
    const text = 'Expedition Failed!';
    const text_size = game.context.measureText(text);
    game.context.fillText(text, 400 - text_size.width / 2, 220);

    if (data.game_state.game_over_cooldown <= 0) {
      const return_button = renderButton(game, !game.game_data.pause.is_paused, {
        x: 300,
        y: 450,
        background: { base: '#AAA', hover: '#CCC' },
        content: {
          type: 'text',
          text: 'Back to town',
          font_size: 30,
          padding: { x: 10, y: 10 },
        },
      });
      if (return_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
        leaveDungeonForTown(game.game_data, 1.0, 1.0);
        data.game_state.game_over_cooldown = -1;
      }
    }
  }

  /**
   * @param {Game} game The current game
   * @param {GameData} data
   */
  function showGameOverScreen(game, data) {
    if (data.game_state.game_over_cooldown > 0) {
      const alpha = Math.max(0.0, Math.min(0.9, 1.0 - data.game_state.game_over_cooldown / 100));
      game.context.fillStyle = `rgba(0, 0, 0, ${alpha})`;
      game.context.fillRect(0, 0, 800, 600);
      game.context.fillStyle = `rgba(255, 0, 0, ${alpha})`;
      game.context.font = '40px sans-serif';
      const text = 'Game Over!';
      const text_size = game.context.measureText(text);
      game.context.fillText(text, 400 - text_size.width / 2, 220);
      return;
    }

    game.context.fillStyle = `#000`;
    game.context.fillRect(0, 0, 800, 600);
    game.context.fillStyle = `#F00`;
    game.context.font = '40px sans-serif';
    const text = 'Game Over!';
    const text_size = game.context.measureText(text);
    game.context.fillText(text, 400 - text_size.width / 2, 120);
    if (data.game_state.screen !== 'game_over') {
      leaveDungeonForTown(data, 1.0, 1.0);
      for (let i = 0; i < 30; ++i) {
        advanceCycle(game.game_data, { advance_mode: 'dungeon_fight', fight_mode: 'small_group' });
      }
      data.game_state.screen = 'game_over';
      return;
    }

    /** @type {Array<{ character: import('../types').StoredCharacterData, equipment?: import('../types').StoredEquipmentData }>} */
    const character_info = getAllLiveCharacters(data).map((c) => ({ character: c }));
    for (const equipment of getAllEquipments(data)) {
      if (equipment.character) {
        character_info.push({
          character: equipment.character.data,
          equipment: equipment,
        });
      }
    }

    const selected_info = character_info.find(
      (info) => info.character.id === data.game_state.game_over_backstory.selected_character,
    );
    if (selected_info) {
      const x = 640;
      const y = 5;
      game.context.fillStyle = '#555';
      game.context.fillRect(x, y, 110, 170);
      let base_image = selected_info.equipment
        ? getFaceImageFromTransmutedCharacter(selected_info.equipment)
        : getFaceImageFromCharacter(selected_info.character);
      if (base_image === 'character_face_base_smiling') {
        game.renderer.renderRawImage(images['character_face_magic_binds_angry'], x, y + 20, 200, 150);
      } else {
        game.renderer.renderRawImage(images[base_image], x, y + 40, 110, 110);
      }
      game.context.fillStyle = `#F00`;
      game.context.font = '20px sans-serif';
      game.context.fillText(selected_info.character.name, x, y + 20, 110);

      const back_button = renderButton(game, !game.game_data.pause.is_paused, {
        x: 48,
        y: 138,
        background: { base: '#555', hover: '#777' },
        content: { type: 'text', text: 'Back', font_size: 20, padding: { x: 5, y: 5 } },
      });
      const raw_backstory = generateReadableBackstory(game.game_data, selected_info.character);
      showScrollableTextBox(
        game,
        !game.game_data.pause.is_paused,
        { x: 48, y: 178, total_width: 704, total_height: 404, border_size: 2, elevator_width: 12, padding: 10 },
        { background: '#333', border: '#222', text: '#CCC', elevator: '#555', hovered_elevator: '#777' },
        data.game_state.game_over_backstory,
        [{ text: `${selected_info.character.name}'s story`, font_size: 30 }].concat(
          raw_backstory.map((t) => ({ text: t, font_size: 15 })),
        ),
      );
      if (back_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
        data.game_state.game_over_backstory.selected_character = undefined;
      }
    } else {
      showScrollableHorizontalList(
        game,
        !game.game_data.pause.is_paused,
        {
          x: 50,
          y: 170,
          total_width: 700,
          total_height: 200,
          padding: 10,
          padding_between: 20,
          elevator_height: 12,
          border_size: 2,
        },
        { background: '#000', border: '#222', elevator: '#555', hovered_elevator: '#777' },
        data.game_state.game_over_scroll,
        character_info.map((info) => ({
          width: 110,
          display: (x, y) => {
            let is_hovering = mouseIsInArea(game, { x: x, y: y, width: 110, height: 170 });

            game.context.fillStyle = is_hovering ? '#777' : '#555';
            game.context.fillRect(x, y, 110, 170);

            let base_image = info.equipment
              ? getFaceImageFromTransmutedCharacter(info.equipment)
              : getFaceImageFromCharacter(info.character);
            if (base_image === 'character_face') {
              game.renderer.renderRawImage(images['character_magic_bindings_gag'], x, y + 20, 200, 150);
            } else {
              game.renderer.renderRawImage(images[base_image], x, y + 40, 110, 110);
            }
            game.context.fillStyle = `#F00`;
            game.context.font = '20px sans-serif';
            game.context.fillText(info.character.name, x, y + 20, 110);

            if (is_hovering) showPopupOnMousePosition(game, getCharacterDescription(info.character));
            if (is_hovering && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
              game.sound_queue.push({ sound: 'click' });
              data.game_state.game_over_backstory.scroll_offset = 0;
              data.game_state.game_over_backstory.selected_character = info.character.id;
            }
          },
        })),
      );
    }
    return;
  }
}
