/**
 * @typedef { import("../engine.mjs").Game } Game
 */

export function getPopupSystem() {
  /**
   * @param {Game} game The current game
   * @param {number} elapsedTicks The elapsed ticks since the last update
   */
  return function getPopupSystem(game, elapsedTicks) {
    if (game.game_data.popup.state === 'displayed') {
      handlePopup(game, game.game_data.popup, elapsedTicks);
    }
    game.game_data.popup = { state: 'none' };
  };

  /**
   * @param {Game} game The current game
   * @param {import("../types").DisplayPopupInformation} popup
   * @param {number} elapsedTicks The elapsed ticks since the last update
   */
  function handlePopup(game, popup, elapsedTicks) {
    const content_size = getPopupContentSize(game, popup);
    const content_width = content_size.width;
    const content_height = content_size.height;
    const full_width = content_size.width + 4;
    const full_height = content_size.height + 4;

    const game_width = 800;
    const game_height = 600;

    const content_start_x = (() => {
      let start_x = popup.x;
      if (popup.coordinates_base_position === 'top_right' || popup.coordinates_base_position === 'bottom_right') {
        start_x = popup.x - content_width;
      }

      if (start_x + content_width > game_width) {
        if (popup.coordinates_adjusted_position === 'corner_on_border') {
          return start_x - content_width;
        }
        if (popup.coordinates_adjusted_position === 'sticky_border') {
          return start_x - (start_x + content_width - game_width);
        }
      }

      return start_x;
    })();
    const content_start_y = (() => {
      let start_y = popup.y;
      if (popup.coordinates_base_position === 'bottom_left' || popup.coordinates_base_position === 'bottom_right') {
        start_y = popup.y - content_height;
      }

      if (start_y + content_height > game_height) {
        if (popup.coordinates_adjusted_position === 'corner_on_border') {
          return start_y - content_height;
        }
        if (popup.coordinates_adjusted_position === 'sticky_border') {
          return start_y - (start_y + content_height - game_height);
        }
      }

      return start_y;
    })();
    const border_start_x = content_start_x - 2;
    const border_start_y = content_start_y - 2;

    game.context.fillStyle = popup.color.border;
    game.context.fillRect(border_start_x, border_start_y, full_width, full_height);
    game.context.fillStyle = popup.color.background;
    game.context.fillRect(content_start_x, content_start_y, content_width, content_height);
    if (popup.content.type === 'text') {
      let text_y = content_start_y;
      for (const fragment of popup.content.text.split('\n')) {
        game.context.font = `${popup.content.font_size}px sans-serif`;
        game.context.fillStyle = `black`;
        game.context.fillText(
          fragment,
          content_start_x + popup.content.padding.x,
          text_y + popup.content.padding.y + popup.content.font_size,
        );
        text_y += popup.content.font_size;
      }
    }
  }

  /**
   * @param {Game} game The current game
   * @param {import("../types").DisplayPopupInformation} popup
   */
  function getPopupContentSize(game, popup) {
    if (popup.content.type === 'text') {
      game.context.font = `${popup.content.font_size}px sans-serif`;
      game.context.fillStyle = `black`;
      let max_text_width = 0;
      let line_count = 0;
      for (const fragment of popup.content.text.split('\n')) {
        const text_size = game.context.measureText(fragment);
        max_text_width = Math.max(max_text_width, text_size.width);
        line_count++;
      }
      return {
        width: max_text_width + 2 * popup.content.padding.x,
        height: popup.content.font_size * line_count + 2 * popup.content.padding.y,
      };
    }
    return {
      width: popup.content.width,
      height: popup.content.height,
    };
  }
}
