import { renderButton } from '../../components/button.mjs';
import { showScrollableHorizontalList } from '../../components/scroll-horizontal.mjs';
import { showScrollableTextBox } from '../../components/scroll-vertical-text.mjs';
import { generateReadableBackstory } from '../../data/backstory-readable.mjs';
import { getFaceImageFromTransmutedCharacter } from '../../data/equipment.mjs';
import { getCharacterDescription, getFaceImageFromCharacter } from '../../data/player-character.mjs';
import { getAllCharactersInTown, getAllEquipmentsInTown } from '../../data/utils.mjs';
import { display_image } from './utils.mjs';

/**
 * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
 */
export function getTownBackstoryScreenSystem(images) {
  return {
    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {import('../../types').TownGameData} data The current game
     */
    displayTownBackstoryScreen: function (game, data) {
      const back_to_town = renderButton(game, !game.game_data.pause.is_paused, {
        x: 50,
        y: 30,
        background: { base: '#AAA', hover: '#CCC' },
        content: { type: 'text', text: 'Back', font_size: 20, padding: { x: 5, y: 5 } },
      });
      if (back_to_town.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
        game.sound_queue.push({ sound: 'click' });
        data.backstory.selected_character = undefined;
        data.screen = 'town';
      }

      /** @type {Array<{ status: "free" | "transmuted", character: import('../../types').StoredCharacterData, face: string }>} */
      const characters = getAllCharactersInTown(game.game_data).map((c) => ({
        status: 'free',
        character: c,
        face: getFaceImageFromCharacter(c),
      }));
      for (const equipment of getAllEquipmentsInTown(game.game_data)) {
        if (equipment.character) {
          characters.push({
            status: 'transmuted',
            character: equipment.character.data,
            face: getFaceImageFromTransmutedCharacter(equipment),
          });
        }
      }

      displayUnassignedCharacters(game, data, characters);

      displaySelectedCharacterBackstory(game, data, characters);
    },
  };

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').TownGameData} data The current town data
   * @param {Array<{ status: "free" | "transmuted", character: import('../../types').StoredCharacterData, face: string }>} all_characters The current game
   */
  function displayUnassignedCharacters(game, data, all_characters) {
    game.context.fillStyle = 'black';
    game.context.font = `16px sans-serif`;
    game.context.fillText(`Available characters`, 25, 85);

    showScrollableHorizontalList(
      game,
      !game.game_data.pause.is_paused,
      {
        x: 25,
        y: 90,
        total_width: 750,
        total_height: 68,
        elevator_height: 5,
        padding: 5,
        border_size: 1,
        padding_between: 10,
      },
      { background: 'transparent', border: 'white', elevator: '#AAA', hovered_elevator: '#CCC' },
      data.backstory_characters_controller,
      all_characters.map((info) => ({
        width: 50,
        display: (x, y) => {
          let image = images[info.face];
          display_image.content(
            game,
            image,
            getCharacterDescription(info.character),
            { x: x, y: y },
            () => true,
            () => {
              data.backstory.selected_character = info.character.id;
              data.backstory.scroll_offset = 1.0;
            },
          );
        },
      })),
    );
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').TownGameData} data The current town data
   * @param {Array<{ status: "free" | "transmuted", character: import('../../types').StoredCharacterData, face: string }>} all_characters The current game
   */
  function displaySelectedCharacterBackstory(game, data, all_characters) {
    const view_start = 170;
    const selected_character_info = all_characters.find((c) => c.character.id === data.backstory.selected_character);
    if (!selected_character_info) {
      game.context.fillStyle = '#555';
      game.context.fillRect(48, view_start - 2, 704, 404);
      game.context.fillStyle = '#777';
      game.context.fillRect(50, view_start, 700, 400);
      game.context.fillStyle = 'black';
      game.context.font = `30px sans-serif`;
      game.context.fillText(`Select a character`, 275, view_start + 200);
      return;
    }
    const selected_character = selected_character_info.character;
    const raw_backstory = generateReadableBackstory(game.game_data, selected_character);
    showScrollableTextBox(
      game,
      !game.game_data.pause.is_paused,
      { x: 48, y: 168, total_width: 704, total_height: 404, border_size: 2, elevator_width: 12, padding: 10 },
      { background: '#AAA', border: '#777', elevator: '#CCC', text: 'black', hovered_elevator: '#CCC' },
      data.backstory,
      [{ text: `${selected_character.name}'s story so far`, font_size: 30 }].concat(
        raw_backstory.map((t) => ({ text: t, font_size: 15 })),
      ),
    );
  }
}
