import { renderButton } from '../../components/button.mjs';
import { mouseIsInArea } from '../../components/checks.mjs';
import { showPopupOnMousePosition } from '../../components/popup.mjs';
import { recomputeEquipmentTargetCycle } from '../../data/equipment.mjs';
import {
  displayAvailableEquipments,
  displayLargeAvailableCharacters,
  display_character,
  display_equipment,
} from './utils.mjs';

/**
 * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
 */
export function getEquipmentUpgradeScreenSystem(images) {
  return {
    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {import('../../types').TownGameData} data The current game
     */
    displayEquipmentUpgradeScreen: function (game, data) {
      const back_to_town = renderButton(game, !game.game_data.pause.is_paused, {
        x: 50,
        y: 30,
        background: { base: '#AAA', hover: '#CCC' },
        content: { type: 'text', text: 'Back', font_size: 20, padding: { x: 5, y: 5 } },
      });
      if (back_to_town.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
        game.sound_queue.push({ sound: 'click' });
        data.screen = 'town';
      }

      displayUnassignedCharacters(game, data);

      displayUnusedItems(game, data);

      displayEquipmentUpgrade(game, data);

      display_equipment.dust.simplified(
        game,
        images,
        { x: 260, y: 430 },
        {
          available: data.available_dust > 0,
          used: data.use_dust_for_upgrade,
          action: () => {
            data.use_dust_for_upgrade = !data.use_dust_for_upgrade;
            recomputeEquipmentTargetCycle(game.game_data, data);
          },
        },
      );
    },
  };

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').TownGameData} data The current game
   */
  function displayUnassignedCharacters(game, data) {
    game.context.fillStyle = 'black';
    game.context.font = `16px sans-serif`;
    game.context.fillText(`Unassigned characters`, 25, 95);

    displayLargeAvailableCharacters(
      game,
      images,
      { x: 25, y: 100, width: 750, height: 120 },
      data.upgrade_characters_controller,
      {
        can_action: (_character) => !data.upgrade_equipment.active || !data.upgrade_equipment.handler,
        on_action: (character) => {
          if (!data.upgrade_equipment.active) {
            data.upgrade_equipment = {
              active: true,
              target: null,
              handler: null,
              retarget_end_cycle: 0,
            };
          }
          data.available_team = data.available_team.filter((t) => t !== character);
          data.upgrade_equipment.handler = character;
          recomputeEquipmentTargetCycle(game.game_data, data);
        },
      },
    );
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').TownGameData} data The current game
   */
  function displayUnusedItems(game, data) {
    game.context.fillStyle = 'black';
    game.context.font = `16px sans-serif`;
    game.context.fillText(`Unused equipment`, 25, 265);

    displayAvailableEquipments(
      game,
      images,
      { x: 25, y: 270, width: 750, height: 100 },
      data.upgrade_equipment_controller,
      {
        can_action: (_equipment) => !data.upgrade_equipment.active || !data.upgrade_equipment.target,
        on_action: (equipment) => {
          if (!data.upgrade_equipment.active) {
            data.upgrade_equipment = {
              active: true,
              target: null,
              handler: null,
              retarget_end_cycle: 0,
            };
          }
          data.unused_equipment = data.unused_equipment.filter((t) => t !== equipment);
          data.upgrade_equipment.target = equipment;
          recomputeEquipmentTargetCycle(game.game_data, data);
        },
      },
    );
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').TownGameData} data The current game
   */
  function displayEquipmentUpgrade(game, data) {
    if (!data.upgrade_equipment.active) {
      game.context.font = `16px sans-serif`;
      const title_size = game.context.measureText(`No valid upgrade selected`);
      game.context.fillStyle = '#C77';
      game.context.fillRect(25, 430, title_size.width + 20, 30);
      game.context.fillStyle = 'black';
      game.context.fillText(`No valid upgrade selected`, 35, 450);
      display_character.detailed_slot.empty(game, { x: 50, y: 470 }, 'No character', undefined);
      display_equipment.detailed_slot.empty(game, { x: 175, y: 470 }, 'No equipment', undefined);
      const is_hovering = mouseIsInArea(game, { x: 25, y: 430, width: title_size.width + 20, height: 30 });
      if (is_hovering) showPopupOnMousePosition(game, 'No equipment to upgrade');
      return;
    }

    if (data.upgrade_equipment.retarget_end_cycle > 0) {
      game.context.font = `16px sans-serif`;
      const remaining_cycles = data.upgrade_equipment.retarget_end_cycle - game.game_data.game_state.current_cycle;
      const remaining_cycles_text = `Upgrade Time: ${remaining_cycles} ${remaining_cycles > 1 ? 'cycles' : 'cycle'}`;
      const title_size = game.context.measureText(remaining_cycles_text);
      game.context.fillStyle = '#7C7';
      game.context.fillRect(25, 430, title_size.width + 20, 30);
      game.context.fillStyle = 'black';
      game.context.fillText(remaining_cycles_text, 35, 450);
    } else {
      game.context.font = `16px sans-serif`;
      const title_size = game.context.measureText(`No valid upgrade selected`);
      game.context.fillStyle = '#C77';
      game.context.fillRect(25, 430, title_size.width + 20, 30);
      game.context.fillStyle = 'black';
      game.context.fillText(`No valid upgrade selected`, 35, 450);
      const is_hovering = mouseIsInArea(game, { x: 25, y: 430, width: title_size.width + 20, height: 30 });
      if (is_hovering) {
        if (!data.upgrade_equipment.target) {
          showPopupOnMousePosition(game, 'No equipment to upgrade');
        } else if (!data.upgrade_equipment.handler) {
          showPopupOnMousePosition(game, 'No character to upgrade the equipment');
        } else {
          showPopupOnMousePosition(game, "The character's level is not greater than the equipment level");
        }
      }
    }

    if (data.upgrade_equipment.handler) {
      display_character.detailed_slot.content(
        game,
        images,
        data.upgrade_equipment.handler,
        { x: 50, y: 470 },
        {
          text: 'Unassign',
          can_action: () => true,
          action: () => {
            if (!data.upgrade_equipment.active) return;
            if (!data.upgrade_equipment.handler) return;
            data.available_team.push(data.upgrade_equipment.handler);
            data.upgrade_equipment.handler = null;
            if (!data.upgrade_equipment.target) data.upgrade_equipment = { active: false };
            recomputeEquipmentTargetCycle(game.game_data, data);
          },
        },
      );
    } else {
      display_character.detailed_slot.empty(game, { x: 50, y: 470 }, 'No character', undefined);
    }

    if (data.upgrade_equipment.target) {
      display_equipment.detailed_slot.content(
        game,
        images,
        data.upgrade_equipment.target,
        { x: 175, y: 470 },
        {
          text: 'Unassign',
          can_action: () => true,
          action: () => {
            if (!data.upgrade_equipment.active) return;
            if (!data.upgrade_equipment.target) return;
            data.unused_equipment.push(data.upgrade_equipment.target);
            data.upgrade_equipment.target = null;
            if (!data.upgrade_equipment.handler) data.upgrade_equipment = { active: false };
            recomputeEquipmentTargetCycle(game.game_data, data);
          },
        },
      );
    } else {
      display_equipment.detailed_slot.empty(game, { x: 175, y: 470 }, 'No equipment', undefined);
    }
  }
}
