import { renderButton } from '../../components/button.mjs';
import { leaveTownForDungeon } from '../../events/leave-town-for-dungeon.mjs';
import {
  displayAvailableEquipments,
  displayLargeAvailableCharacters,
  display_character,
  display_equipment,
} from './utils.mjs';

/**
 * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
 */
export function getExpeditionScreenSystem(images) {
  return {
    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {import('../../types').TownGameData} data The current game
     */
    displayExpeditionScreen: function (game, data) {
      const back_to_town = renderButton(game, !game.game_data.pause.is_paused, {
        x: 50,
        y: 30,
        background: { base: '#AAA', hover: '#CCC' },
        content: { type: 'text', text: 'Back', font_size: 20, padding: { x: 5, y: 5 } },
      });
      if (back_to_town.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
        game.sound_queue.push({ sound: 'click' });
        data.screen = 'town';
      }

      // Unassigned team
      displayUnassignedCharacters(game, data);

      // Unused items
      displayUnusedItems(game, data);

      // Expedition
      displayExpeditionTeam(game, data);
    },
  };

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').TownGameData} data The current game
   */
  function displayUnassignedCharacters(game, data) {
    game.context.fillStyle = 'black';
    game.context.font = `16px sans-serif`;
    game.context.fillText(`Unassigned characters`, 25, 95);

    displayLargeAvailableCharacters(
      game,
      images,
      { x: 25, y: 100, width: 750, height: 120 },
      data.expedition_characters_controller,
      {
        can_action: (_character) => data.expedition_slots.some((r) => !r.character),
        on_action: (character) => {
          const empty_character_slot = data.expedition_slots.find((r) => !r.character);
          if (!empty_character_slot) return;
          data.available_team = data.available_team.filter((t) => t !== character);
          empty_character_slot.character = character;
        },
      },
    );
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').TownGameData} data The current game
   */
  function displayUnusedItems(game, data) {
    game.context.fillStyle = 'black';
    game.context.font = `16px sans-serif`;
    game.context.fillText(`Unused equipment`, 25, 295);

    displayAvailableEquipments(
      game,
      images,
      { x: 25, y: 300, width: 750, height: 100 },
      data.expedition_equipment_controller,
      {
        can_action: (equipment) => !!data.expedition_slots.find((r) => !r.equipment[equipment.type]),
        on_action: (equipment) => {
          const empty_character_slot = data.expedition_slots.find((r) => !r.equipment[equipment.type]);
          if (!empty_character_slot) return;
          data.unused_equipment = data.unused_equipment.filter((t) => t !== equipment);
          empty_character_slot.equipment[equipment.type] = equipment;
        },
      },
    );
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').TownGameData} data The current game
   */
  function displayExpeditionTeam(game, data) {
    let start_x = 25;
    const start_y = 470;
    for (let i = 0; i < 4; ++i) {
      const slot = data.expedition_slots[i];
      if (slot && slot.character) {
        display_character.detailed_slot.content(
          game,
          images,
          slot.character,
          { x: start_x, y: start_y },
          {
            text: 'Unassign',
            can_action: () => true,
            action: () => {
              if (!slot.character) return;
              data.available_team.push(slot.character);
              slot.character = undefined;
            },
          },
        );
      } else {
        display_character.detailed_slot.empty(game, { x: start_x, y: start_y }, 'No character', undefined);
      }
      displayExpeditionEquipmentSlot(game, data, slot.equipment, { x: start_x - 5, y: start_y + 90 }, 'body');
      displayExpeditionEquipmentSlot(game, data, slot.equipment, { x: start_x + 35, y: start_y + 90 }, 'accessory');
      displayExpeditionEquipmentSlot(game, data, slot.equipment, { x: start_x + 75, y: start_y + 90 }, 'weapon');
      start_x += 125;
    }

    const can_start_expedition = data.expedition_slots.some((e) => !!e.character);
    const status = renderButton(game, !game.game_data.pause.is_paused, {
      x: 530,
      y: 500,
      background: { base: can_start_expedition ? '#AAA' : '#777', hover: can_start_expedition ? '#CCC' : '#777' },
      content: { type: 'text', text: 'Start Expedition', font_size: 30, padding: { x: 20, y: 20 } },
    });
    if (status.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
      if (can_start_expedition) {
        game.sound_queue.push({ sound: 'click' });
        leaveTownForDungeon(game.game_data);
      } else {
        game.sound_queue.push({ sound: 'error' });
      }
    }
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').TownGameData} data
   * @param {import('../../types').InGameCharacterEquipment} equipment_group
   * @param {{ x: number, y: number }} coordinates
   * @param {import('../../types').StoredEquipmentData["type"]} equipment_name
   */
  function displayExpeditionEquipmentSlot(game, data, equipment_group, coordinates, equipment_name) {
    const actual_equipment = equipment_group[equipment_name];
    if (actual_equipment) {
      display_equipment.simplified_slot.content(
        game,
        images,
        actual_equipment,
        coordinates,
        () => !!actual_equipment,
        () => {
          data.unused_equipment.push(actual_equipment);
          equipment_group[equipment_name] = undefined;
        },
      );
    } else {
      display_equipment.simplified_slot.empty(
        game,
        coordinates,
        () => false,
        () => {},
      );
    }
  }
}
