import { changeStoredCharacterRoleTo } from '../../data/player-character.mjs';
import { assertTypeIsHandled } from '../../utils.mjs';
import { getExpeditionScreenSystem } from './expedition.mjs';
import { getTownScreenSystem } from './town.mjs';
import { getRoleChangeScreenSystem } from './role.mjs';
import { getEquipmentUpgradeScreenSystem } from './equipment-upgrade.mjs';
import { getTownBackstoryScreenSystem } from './backstory.mjs';
import { renderButton } from '../../components/button.mjs';
import { advanceCycle } from '../../events/advance-cycle.mjs';
import { getEquipmentCleanupScreenSystem } from './equipment-cleanup.mjs';

/**
 * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
 */
export function getTownSystem(images) {
  const townScreenSystem = getTownScreenSystem(images);
  const expeditionScreenSystem = getExpeditionScreenSystem(images);
  const roleChangeScreenSystem = getRoleChangeScreenSystem(images);
  const equipmentUpgradeScreenSystem = getEquipmentUpgradeScreenSystem(images);
  const equipmentCleanupScreenSystem = getEquipmentCleanupScreenSystem(images);
  const townBackstoryScreenSystem = getTownBackstoryScreenSystem(images);

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {number} elapsedTicks The elapsed ticks since the last update
   */
  return function townSystem(game, elapsedTicks) {
    if (game.game_data.game_state.screen !== 'town') return;
    const data = game.game_data.town;
    game.renderer.renderRawImage(images[data.background_image], 0, 0, 800, 600);

    displayCorruptionInformation(game, data);
    const wait_one_cycle = renderButton(game, !game.game_data.pause.is_paused, {
      x: 470,
      y: 50,
      background: { base: '#AAA', hover: '#CCC' },
      content: { type: 'fixed_text', text: 'Wait one cycle', font_size: 20, width: 150, height: 40 },
    });
    if (wait_one_cycle.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
      game.sound_queue.push({ sound: 'click' });
      advanceCycle(game.game_data, { advance_mode: 'skip_in_town' });
    }

    game.audio.stop_loop('combat_drum');
    game.audio.stop_loop('combat_bass');
    game.audio.stop_loop('combat_kicks');
    game.audio.stop_loop('combat_lead');
    game.audio.start_loop('town_bass');
    game.audio.start_loop('town_twinkles');
    game.audio.start_loop('town_lead_prep');
    game.audio.start_loop('town_lead_wait');

    game.audio.loop_enable('town_bass');

    if (data.screen === 'town') {
      game.audio.loop_disable('town_lead_wait');
      game.audio.loop_disable('town_lead_prep');
      game.audio.loop_disable('town_twinkles');
      townScreenSystem.displayTownScreen(game, data);
      return;
    }
    if (data.screen === 'role_change') {
      game.audio.loop_disable('town_lead_wait');
      game.audio.loop_enable('town_lead_prep');
      game.audio.loop_disable('town_twinkles');
      roleChangeScreenSystem.displayRoleChangeScreen(game, data);
      return;
    }
    if (data.screen === 'equipment_upgrade') {
      game.audio.loop_disable('town_lead_wait');
      game.audio.loop_enable('town_lead_prep');
      game.audio.loop_disable('town_twinkles');
      equipmentUpgradeScreenSystem.displayEquipmentUpgradeScreen(game, data);
      return;
    }
    if (data.screen === 'equipment_cleanup') {
      game.audio.loop_disable('town_lead_wait');
      game.audio.loop_enable('town_lead_prep');
      game.audio.loop_disable('town_twinkles');
      equipmentCleanupScreenSystem.displayEquipmentCleanupScreen(game, data);
      return;
    }
    if (data.screen === 'expedition') {
      game.audio.loop_disable('town_lead_wait');
      game.audio.loop_disable('town_lead_prep');
      game.audio.loop_enable('town_twinkles');
      expeditionScreenSystem.displayExpeditionScreen(game, data);
      return;
    }
    if (data.screen === 'backstory') {
      game.audio.loop_enable('town_lead_wait');
      game.audio.loop_disable('town_lead_prep');
      game.audio.loop_disable('town_twinkles');
      townBackstoryScreenSystem.displayTownBackstoryScreen(game, data);
      return;
    }
    assertTypeIsHandled(data.screen, 'town screen');
  };

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').TownGameData} data The current game
   */
  function displayCorruptionInformation(game, data) {
    const ratio = game.game_data.game_state.current_corruption_ratio;
    game.context.fillStyle = 'black';
    game.context.font = '15px sans-serif';
    game.context.fillText(`Corruption: ${Math.ceil(ratio * 100)}%`, 650, 55);
    game.context.fillStyle = '#555';
    game.context.fillRect(648, 63, 104, 24);
    game.context.fillStyle = '#585';
    game.context.fillRect(650, 65, 100, 20);
    if (ratio > 0) {
      const pixels = Math.ceil(ratio * 100);
      game.context.fillStyle = '#A25';
      game.context.fillRect(650 + (100 - pixels), 65, pixels, 20);
    }
  }
}
