import { renderButton } from '../../components/button.mjs';
import { showPopupOnMousePosition } from '../../components/popup.mjs';
import { displayLargeAvailableCharacters, display_character } from './utils.mjs';

/**
 * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
 */
export function getRoleChangeScreenSystem(images) {
  return {
    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {import('../../types').TownGameData} data The current game
     */
    displayRoleChangeScreen: function (game, data) {
      const back_to_town = renderButton(game, !game.game_data.pause.is_paused, {
        x: 50,
        y: 30,
        background: { base: '#AAA', hover: '#CCC' },
        content: { type: 'text', text: 'Back', font_size: 20, padding: { x: 5, y: 5 } },
      });
      if (back_to_town.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
        game.sound_queue.push({ sound: 'click' });
        data.screen = 'town';
      }

      displayUnassignedCharacters(game, data);

      displayCharacterRoleChange(game, data);
    },
  };

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').TownGameData} data The current game
   */
  function displayUnassignedCharacters(game, data) {
    game.context.fillStyle = 'black';
    game.context.font = `16px sans-serif`;
    game.context.fillText(`Unassigned characters`, 25, 95);
    displayLargeAvailableCharacters(
      game,
      images,
      { x: 25, y: 100, width: 750, height: 320 },
      data.role_characters_controller,
      {
        can_action: (_character) => !data.change_role.active,
        on_action: (character) => {
          data.change_role = {
            active: true,
            target: character,
            target_role: 'adventurer',
            retarget_end_cycle: game.game_data.game_state.current_cycle + Math.ceil(character.level * 0.5),
          };
          data.available_team = data.available_team.filter((t) => t !== character);
        },
      },
    );
  }

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {import('../../types').TownGameData} data The current game
   */
  function displayCharacterRoleChange(game, data) {
    if (!data.change_role.active) {
      game.context.fillStyle = 'black';
      game.context.font = `16px sans-serif`;
      game.context.fillText(`Target Role: none`, 25, 450);
      display_character.detailed_slot.empty(game, { x: 50, y: 470 }, 'No character', undefined);
      return;
    }

    game.context.fillStyle = 'black';
    game.context.font = `16px sans-serif`;
    game.context.fillText(`Target Role: ${data.change_role.target_role}`, 25, 450);
    const remaining_cycles = data.change_role.retarget_end_cycle - game.game_data.game_state.current_cycle;
    const remaining_cycles_text = `Time: ${remaining_cycles} ${remaining_cycles > 1 ? 'cycles' : 'cycle'}`;
    game.context.fillText(remaining_cycles_text, 325, 450);

    const can_select_advanced_role = data.change_role.target.level >= 3;
    const invalid_bg = { base: '#777', hover: '#777' };
    const selectable_bg = { base: '#AAA', hover: '#CCC' };
    const selected_bg = { base: '#CCC', hover: '#CCC' };
    const advanced_bg = can_select_advanced_role ? selectable_bg : invalid_bg;
    const adventurer_button = renderButton(game, !game.game_data.pause.is_paused, {
      x: 200,
      y: 470,
      background: data.change_role.target_role === 'adventurer' ? selected_bg : selectable_bg,
      content: { type: 'fixed_text', text: 'Adventurer', font_size: 15, width: 100, height: 25 },
    });
    if (adventurer_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
      game.sound_queue.push({ sound: 'click' });
      data.change_role.target_role = 'adventurer';
      data.change_role.retarget_end_cycle =
        game.game_data.game_state.current_cycle + Math.ceil(data.change_role.target.level * 0.5);
    }
    const warrior_button = renderButton(game, !game.game_data.pause.is_paused, {
      x: 200,
      y: 500,
      background: data.change_role.target_role === 'warrior' ? selected_bg : advanced_bg,
      content: { type: 'fixed_text', text: 'Warrior', font_size: 15, width: 100, height: 25 },
    });
    if (warrior_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
      if (can_select_advanced_role) {
        game.sound_queue.push({ sound: 'click' });
        data.change_role.target_role = 'warrior';
        data.change_role.retarget_end_cycle =
          game.game_data.game_state.current_cycle + Math.ceil(data.change_role.target.level * 0.75);
      } else {
        game.sound_queue.push({ sound: 'error' });
      }
    }
    if (warrior_button.is_hovered && !can_select_advanced_role)
      showPopupOnMousePosition(game, 'Requires level 3 or more');

    const rogue_button = renderButton(game, !game.game_data.pause.is_paused, {
      x: 200,
      y: 530,
      background: data.change_role.target_role === 'rogue' ? selected_bg : advanced_bg,
      content: { type: 'fixed_text', text: 'Rogue', font_size: 15, width: 100, height: 25 },
    });
    if (rogue_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
      if (can_select_advanced_role) {
        game.sound_queue.push({ sound: 'click' });
        data.change_role.target_role = 'rogue';
        data.change_role.retarget_end_cycle =
          game.game_data.game_state.current_cycle + Math.ceil(data.change_role.target.level * 0.75);
      } else {
        game.sound_queue.push({ sound: 'error' });
      }
    }
    if (rogue_button.is_hovered && !can_select_advanced_role)
      showPopupOnMousePosition(game, 'Requires level 3 or more');

    const mage_button = renderButton(game, !game.game_data.pause.is_paused, {
      x: 200,
      y: 560,
      background: data.change_role.target_role === 'mage' ? selected_bg : advanced_bg,
      content: { type: 'fixed_text', text: 'Mage', font_size: 15, width: 100, height: 25 },
    });
    if (mage_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
      if (can_select_advanced_role) {
        game.sound_queue.push({ sound: 'click' });
        data.change_role.target_role = 'mage';
        data.change_role.retarget_end_cycle =
          game.game_data.game_state.current_cycle + Math.ceil(data.change_role.target.level * 0.75);
      } else {
        game.sound_queue.push({ sound: 'error' });
      }
    }
    if (mage_button.is_hovered && !can_select_advanced_role) showPopupOnMousePosition(game, 'Requires level 3 or more');

    display_character.detailed_slot.content(
      game,
      images,
      data.change_role.target,
      { x: 50, y: 470 },
      {
        text: 'Unassign',
        can_action: () => true,
        action: () => {
          if (!data.change_role.active) return;
          data.available_team.push(data.change_role.target);
          data.change_role = { active: false };
        },
      },
    );
  }
}
