import { renderButton } from '../../components/button.mjs';
import { mouseIsInArea } from '../../components/checks.mjs';
import { showPopupOnMousePosition } from '../../components/popup.mjs';
import { showScrollableVerticalGrid } from '../../components/scroll-vertical-grid.mjs';
import { getEquipmentDescription } from '../../data/equipment.mjs';
import { getCharacterDescription, getImageFromCharacter } from '../../data/player-character.mjs';

const DUST_DESCRIPTION = `Dust is used to speed up and reach higher equipment level upgrades.\nWhen the character's level is less than 10 above the item's level,\nthe upgrade operation will cost 1 dust per extra level needed.\n`;

/**
 * @param {import("../../engine.mjs").Game} game The current game
 * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
 * @param {{ x: number, y: number, width: number, height: number }} layout
 * @param {import('../../types').ScrollController} controller
 * @param {{
 *  can_action: (equipment: import('../../types').StoredEquipmentData) => boolean,
 *  on_action: (equipment: import('../../types').StoredEquipmentData) => void
 * }} [actions]
 */
export function displayAvailableEquipments(game, images, layout, controller, actions) {
  showScrollableVerticalGrid(
    game,
    !game.game_data.pause.is_paused,
    {
      x: layout.x,
      y: layout.y,
      total_width: layout.width,
      total_height: layout.height,
      border_size: 1,
      elevator_width: 6,
      item_width: 30,
      item_height: 30,
      padding: 5,
      padding_between: 10,
    },
    {
      border: '#CCC',
      background: 'transparent',
      elevator: '#777',
      hovered_elevator: '#AAA',
    },
    controller,
    game.game_data.town.unused_equipment.map((e) => ({
      display: (x, y) => {
        if (actions) {
          display_equipment.simplified_slot.content(
            game,
            images,
            e,
            { x: x, y: y },
            () => actions.can_action(e),
            () => actions.on_action(e),
          );
        } else {
          display_equipment.simplified_slot.content(game, images, e, { x: x, y: y });
        }
      },
    })),
  );
}

/**
 * @param {import("../../engine.mjs").Game} game The current game
 * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
 * @param {{ x: number, y: number, width: number, height: number }} layout
 * @param {import('../../types').ScrollController} controller
 * @param {{
 *  can_action: (equipment: import('../../types').StoredCharacterData) => boolean,
 *  on_action: (equipment: import('../../types').StoredCharacterData) => void
 * }} [actions]
 */
export function displaySmallAvailableCharacters(game, images, layout, controller, actions) {
  showScrollableVerticalGrid(
    game,
    !game.game_data.pause.is_paused,
    {
      x: layout.x,
      y: layout.y,
      total_width: layout.width,
      total_height: layout.height,
      border_size: 1,
      elevator_width: 6,
      item_width: 50,
      item_height: 50,
      padding: 5,
      padding_between: 10,
    },
    {
      border: '#CCC',
      background: 'transparent',
      elevator: '#777',
      hovered_elevator: '#AAA',
    },
    controller,
    game.game_data.town.available_team.map((c) => ({
      display: (x, y) => {
        if (actions) {
          display_character.simplified_slot.content(
            game,
            images,
            c,
            { x: x, y: y },
            () => actions.can_action(c),
            () => actions.on_action(c),
          );
        } else {
          display_character.simplified_slot.content(game, images, c, { x: x, y: y });
        }
      },
    })),
  );
}

/**
 * @param {import("../../engine.mjs").Game} game The current game
 * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
 * @param {{ x: number, y: number, width: number, height: number }} layout
 * @param {import('../../types').ScrollController} controller
 * @param {{
 *  can_action: (equipment: import('../../types').StoredCharacterData) => boolean,
 *  on_action: (equipment: import('../../types').StoredCharacterData) => void
 * }} [actions]
 */
export function displayLargeAvailableCharacters(game, images, layout, controller, actions) {
  showScrollableVerticalGrid(
    game,
    !game.game_data.pause.is_paused,
    {
      x: layout.x,
      y: layout.y,
      total_width: layout.width,
      total_height: layout.height,
      border_size: 1,
      elevator_width: 6,
      item_width: 100,
      item_height: 80,
      padding: 5,
      padding_between: 10,
    },
    {
      border: '#CCC',
      background: 'transparent',
      elevator: '#777',
      hovered_elevator: '#AAA',
    },
    controller,
    game.game_data.town.available_team.map((c) => ({
      display: (x, y) => {
        if (actions) {
          display_character.detailed_slot.content(
            game,
            images,
            c,
            { x: x, y: y },
            {
              text: 'Assign',
              can_action: () => actions.can_action(c),
              action: () => actions.on_action(c),
            },
          );
        } else {
          display_character.detailed_slot.content(game, images, c, { x: x, y: y }, undefined);
        }
      },
    })),
  );
}

export const display_image = {
  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {{ x: number, y: number }} coordinates
   * @param {() => boolean} [can_action]
   * @param {() => void} [on_action]
   * @returns {{ width: number, height: number }}
   */
  empty: function (game, coordinates, can_action, on_action) {
    const action_enabled = can_action && on_action && can_action();
    const equipment_button = renderButton(game, !game.game_data.pause.is_paused, {
      x: coordinates.x,
      y: coordinates.y,
      background: { base: action_enabled ? '#AAA' : '#777', hover: action_enabled ? '#CCC' : '#777' },
      content: { type: 'fixed', width: 50, height: 50 },
    });
    if (equipment_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
      if (action_enabled) {
        game.sound_queue.push({ sound: 'click' });
        on_action();
      } else {
        game.sound_queue.push({ sound: 'error' });
      }
    }
    return { width: 50, height: 50 };
  },

  /**
   * @param {import("../../engine.mjs").Game} game The current game
   * @param {ImageBitmap} image The image to use for the renderer
   * @param {string} popup_description
   * @param {{ x: number, y: number }} coordinates
   * @param {() => boolean} [can_action]
   * @param {() => void} [on_action]
   * @returns {{ width: number, height: number }}
   */
  content: function (game, image, popup_description, coordinates, can_action, on_action) {
    const action_enabled = can_action && on_action && can_action();
    const equipment_button = renderButton(game, !game.game_data.pause.is_paused, {
      x: coordinates.x,
      y: coordinates.y,
      background: { base: action_enabled ? '#AAA' : '#777', hover: action_enabled ? '#CCC' : '#777' },
      content: { type: 'fixed', width: 50, height: 50 },
    });
    if (image) {
      game.renderer.renderRawImage(image, coordinates.x, coordinates.y, 50, 50);
    }
    if (equipment_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
      if (action_enabled) {
        game.sound_queue.push({ sound: 'click' });
        on_action();
      } else {
        game.sound_queue.push({ sound: 'error' });
      }
    }
    if (equipment_button.is_hovered) showPopupOnMousePosition(game, popup_description);
    return { width: 50, height: 50 };
  },
};

export const display_character = {
  simplified_slot: {
    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {{ x: number, y: number }} coordinates
     * @param {() => boolean} [can_action]
     * @param {() => void} [on_action]
     * @returns {{ width: number, height: number }}
     */
    empty: function (game, coordinates, can_action, on_action) {
      const action_enabled = can_action && on_action && can_action();
      const equipment_button = renderButton(game, !game.game_data.pause.is_paused, {
        x: coordinates.x,
        y: coordinates.y,
        background: { base: action_enabled ? '#AAA' : '#777', hover: action_enabled ? '#CCC' : '#777' },
        content: { type: 'fixed', width: 50, height: 50 },
      });
      if (equipment_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
        if (action_enabled) {
          game.sound_queue.push({ sound: 'click' });
          on_action();
        } else {
          game.sound_queue.push({ sound: 'error' });
        }
      }
      return { width: 50, height: 50 };
    },

    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
     * @param {import('../../types').StoredCharacterData} character
     * @param {{ x: number, y: number }} coordinates
     * @param {() => boolean} [can_action]
     * @param {() => void} [on_action]
     * @returns {{ width: number, height: number }}
     */
    content: function (game, images, character, coordinates, can_action, on_action) {
      const action_enabled = can_action && on_action && can_action();
      const equipment_button = renderButton(game, !game.game_data.pause.is_paused, {
        x: coordinates.x,
        y: coordinates.y,
        background: { base: action_enabled ? '#AAA' : '#777', hover: action_enabled ? '#CCC' : '#777' },
        content: { type: 'fixed', width: 50, height: 50 },
      });
      const image = images[getImageFromCharacter(character)];
      if (image) {
        game.renderer.renderRawImage(image, coordinates.x, coordinates.y, (8 * 50) / 6, 50);
      }
      if (equipment_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
        if (action_enabled) {
          game.sound_queue.push({ sound: 'click' });
          on_action();
        } else {
          game.sound_queue.push({ sound: 'error' });
        }
      }
      if (equipment_button.is_hovered) showPopupOnMousePosition(game, getCharacterDescription(character));
      return { width: 50, height: 50 };
    },
  },
  detailed_slot: {
    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {{ x: number, y: number }} coordinates
     * @param {string} empty_text
     * @param {{ text: string, can_action: () => boolean, action: () => void } | undefined} action
     * @returns {{ width: number, height: number }}
     */
    empty: function (game, coordinates, empty_text, action) {
      game.context.fillStyle = '#888';
      game.context.fillRect(coordinates.x, coordinates.y, 100, 80);
      game.context.font = `15px sans-serif`;
      game.context.fillStyle = 'black';
      game.context.fillText(empty_text, coordinates.x + 8, coordinates.y + 45);
      if (action) {
        const enabled = action.can_action();
        const action_button = renderButton(game, !game.game_data.pause.is_paused, {
          x: coordinates.x + 15,
          y: coordinates.y + 45,
          background: { base: enabled ? '#AAA' : '#777', hover: enabled ? '#CCC' : '#777' },
          content: { type: 'text', text: action.text, font_size: 15, padding: { x: 5, y: 5 } },
        });
        if (action_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
          if (enabled) {
            game.sound_queue.push({ sound: 'click' });
            action.action();
          } else {
            game.sound_queue.push({ sound: 'error' });
          }
        }
      }
      return { width: 50, height: 50 };
    },

    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
     * @param {import('../../types').StoredCharacterData} character
     * @param {{ x: number, y: number }} coordinates
     * @param {{ text: string, can_action: () => boolean, action: () => void } | undefined} action
     * @returns {{ width: number, height: number }}
     */
    content: function (game, images, character, coordinates, action) {
      game.context.fillStyle = '#888';
      game.context.fillRect(coordinates.x, coordinates.y, 100, 80);
      game.context.fillStyle = 'black';
      game.context.font = `16px sans-serif`;
      game.context.fillText(`lv.${character.level} ${character.role}`, coordinates.x + 5, coordinates.y + 15, 90);
      game.context.font = `20px sans-serif`;
      game.context.fillText(`${character.name}`, coordinates.x + 5, coordinates.y + 35, 90);
      const is_box_hovered = mouseIsInArea(game, { x: coordinates.x, y: coordinates.y, width: 100, height: 80 });
      if (action) {
        const enabled = action.can_action();
        const action_button = renderButton(game, !game.game_data.pause.is_paused, {
          x: coordinates.x + 15,
          y: coordinates.y + 45,
          background: { base: enabled ? '#AAA' : '#777', hover: enabled ? '#CCC' : '#777' },
          content: { type: 'text', text: action.text, font_size: 15, padding: { x: 5, y: 5 } },
        });
        if (action_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
          if (enabled) {
            game.sound_queue.push({ sound: 'click' });
            action.action();
          } else {
            game.sound_queue.push({ sound: 'error' });
          }
        }
        if (is_box_hovered && !action_button.is_hovered) {
          showPopupOnMousePosition(game, getCharacterDescription(character));
        }
      } else {
        if (is_box_hovered) {
          showPopupOnMousePosition(game, getCharacterDescription(character));
        }
      }
      return { width: 100, height: 80 };
    },
  },
};

export const display_equipment = {
  simplified_slot: {
    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {{ x: number, y: number }} coordinates
     * @param {() => boolean} [can_action]
     * @param {() => void} [on_action]
     * @returns {{ width: number, height: number }}
     */
    empty: function (game, coordinates, can_action, on_action) {
      const action_enabled = can_action && on_action && can_action();
      const equipment_button = renderButton(game, !game.game_data.pause.is_paused, {
        x: coordinates.x,
        y: coordinates.y,
        background: { base: action_enabled ? '#AAA' : '#777', hover: action_enabled ? '#CCC' : '#777' },
        content: { type: 'fixed', width: 30, height: 30 },
      });
      if (equipment_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
        if (action_enabled) {
          game.sound_queue.push({ sound: 'click' });
          on_action();
        } else {
          game.sound_queue.push({ sound: 'error' });
        }
      }
      return { width: 30, height: 30 };
    },

    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
     * @param {import('../../types').StoredEquipmentData} equipment
     * @param {{ x: number, y: number }} coordinates
     * @param {() => boolean} [can_action]
     * @param {() => void} [on_action]
     * @returns {{ width: number, height: number }}
     */
    content: function (game, images, equipment, coordinates, can_action, on_action) {
      const action_enabled = can_action && on_action && can_action();
      const equipment_button = renderButton(game, !game.game_data.pause.is_paused, {
        x: coordinates.x,
        y: coordinates.y,
        background: { base: action_enabled ? '#AAA' : '#777', hover: action_enabled ? '#CCC' : '#777' },
        content: { type: 'fixed', width: 30, height: 30 },
      });
      const image = images[equipment.image];
      if (image) {
        game.renderer.renderRawImage(image, coordinates.x, coordinates.y, 30, 30);
      }
      if (equipment_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
        if (action_enabled) {
          game.sound_queue.push({ sound: 'click' });
          on_action();
        } else {
          game.sound_queue.push({ sound: 'error' });
        }
      }
      if (equipment_button.is_hovered) showPopupOnMousePosition(game, getEquipmentDescription(equipment));
      return { width: 30, height: 30 };
    },
  },
  detailed_slot: {
    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {{ x: number, y: number }} coordinates
     * @param {string} empty_text
     * @param {{ text: string, can_action: () => boolean, action: () => void } | undefined} action
     * @returns {{ width: number, height: number }}
     */
    empty: function (game, coordinates, empty_text, action) {
      game.context.fillStyle = '#888';
      game.context.fillRect(coordinates.x, coordinates.y, 100, 80);
      game.context.font = `15px sans-serif`;
      game.context.fillStyle = 'black';
      game.context.fillText(empty_text, coordinates.x + 8, coordinates.y + 45);
      if (action) {
        const enabled = action.can_action();
        const action_button = renderButton(game, !game.game_data.pause.is_paused, {
          x: coordinates.x + 15,
          y: coordinates.y + 45,
          background: { base: enabled ? '#AAA' : '#777', hover: enabled ? '#CCC' : '#777' },
          content: { type: 'text', text: action.text, font_size: 15, padding: { x: 5, y: 5 } },
        });
        if (action_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
          if (enabled) {
            game.sound_queue.push({ sound: 'click' });
            action.action();
          } else {
            game.sound_queue.push({ sound: 'error' });
          }
        }
      }
      return { width: 50, height: 50 };
    },

    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
     * @param {import('../../types').StoredEquipmentData} equipment
     * @param {{ x: number, y: number }} coordinates
     * @param {{ text: string, can_action: () => boolean, action: () => void } | undefined} action
     * @returns {{ width: number, height: number }}
     */
    content: function (game, images, equipment, coordinates, action) {
      game.context.fillStyle = '#888';
      game.context.fillRect(coordinates.x, coordinates.y, 100, 80);
      game.context.fillStyle = 'black';
      game.context.font = `12px sans-serif`;
      game.context.fillText(`lv.${equipment.level} ${equipment.type}`, coordinates.x + 5, coordinates.y + 10, 90);
      game.context.font = `12px sans-serif`;
      game.context.fillText(`${equipment.name}`, coordinates.x + 5, coordinates.y + 25, 90);
      const image = images[equipment.image];
      if (image) {
        game.renderer.renderRawImage(image, coordinates.x + 5, coordinates.y + 50, 25, 25);
      }
      const is_box_hovered = mouseIsInArea(game, { x: coordinates.x, y: coordinates.y, width: 100, height: 80 });
      if (action) {
        const enabled = action.can_action();
        const action_button = renderButton(game, !game.game_data.pause.is_paused, {
          x: coordinates.x + 50,
          y: coordinates.y + 45,
          background: { base: enabled ? '#AAA' : '#777', hover: enabled ? '#CCC' : '#777' },
          content: { type: 'text', text: action.text, font_size: 15, padding: { x: 5, y: 5 } },
        });
        if (action_button.is_hovered && !game.game_data.pause.is_paused && game.mouse.primary_button_clicked_frame) {
          if (enabled) {
            game.sound_queue.push({ sound: 'click' });
            action.action();
          } else {
            game.sound_queue.push({ sound: 'error' });
          }
        }
        if (is_box_hovered && !action_button.is_hovered) {
          showPopupOnMousePosition(game, getEquipmentDescription(equipment));
        }
      } else {
        if (is_box_hovered) {
          showPopupOnMousePosition(game, getEquipmentDescription(equipment));
        }
      }
      return { width: 100, height: 80 };
    },
  },
  dust: {
    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
     * @param {{ x: number, y: number }} coordinates
     * @param {{ available: boolean, used: boolean, action: () => void } | undefined} action
     * @returns {{ width: number, height: number }}
     */
    simplified: function (game, images, coordinates, action) {
      const base = action ? (action.available ? (action.used ? '#AAA' : '#888') : '#777') : '#AAA';
      const hover = action ? (action.available ? (action.used ? '#CCC' : '#AAA') : '#777') : '#AAA';
      const equipment_button = renderButton(game, !game.game_data.pause.is_paused, {
        x: coordinates.x,
        y: coordinates.y,
        background: { base, hover },
        content: { type: 'fixed', width: 30, height: 30 },
      });
      const image = images['item_dust'];
      if (image) {
        game.renderer.renderRawImage(image, coordinates.x, coordinates.y, 30, 30);
      }
      game.context.font = `14px sans-serif`;
      game.context.fillStyle = 'black';
      game.context.fillText(`x${game.game_data.town.available_dust}`, coordinates.x + 15, coordinates.y + 25, 20);
      if (equipment_button.is_hovered && !game.game_data.pause.is_paused) {
        if (action && game.mouse.primary_button_clicked_frame) {
          game.sound_queue.push({ sound: 'click' });
          action.action();
        }
        let extra_current = '';
        if (
          game.game_data.town.upgrade_equipment.active &&
          game.game_data.town.upgrade_equipment.handler &&
          game.game_data.town.upgrade_equipment.target
        ) {
          if (game.game_data.town.use_dust_for_upgrade) {
            let level_diff =
              game.game_data.town.upgrade_equipment.handler.level -
              game.game_data.town.upgrade_equipment.target.level -
              1;
            const used_dust = Math.max(level_diff, game.game_data.town.available_dust);
            if (game.game_data.town.available_dust === 0) {
              extra_current = `(Click to stop using dust when upgrading)\n(Not dust available to use during the operation)\n`;
            } else if (
              used_dust === game.game_data.town.available_dust &&
              level_diff !== game.game_data.town.available_dust
            ) {
              extra_current = `(Click to stop using dust when upgrading)\n(Not enough dust. Current operation use all ${used_dust} remaining dust)\n`;
            } else if (used_dust > 0) {
              extra_current = `(Click to stop using dust when upgrading)\n(Current operation will use ${used_dust} dust)\n`;
            } else {
              extra_current = `(Click to stop using dust when upgrading)\n(Level difference is high enough, dust will not be used)\n`;
            }
          } else {
            extra_current = `(Click to use dust when upgrading)\n(Dust is currently not used)\n`;
          }
        } else {
          if (game.game_data.town.use_dust_for_upgrade) {
            extra_current = `(Click to stop using dust when upgrading)\n(No upgrade operation ongoing)\n`;
          } else {
            extra_current = `(Click to use dust when upgrading)\n(Dust is currently not used)\n`;
          }
        }
        showPopupOnMousePosition(
          game,
          `${extra_current}${DUST_DESCRIPTION}Available Dust: ${game.game_data.town.available_dust}`,
        );
      }
      return { width: 30, height: 30 };
    },

    /**
     * @param {import("../../engine.mjs").Game} game The current game
     * @param {{ [key: string]: ImageBitmap }} images The loaded images to use by the renderer
     * @param {{ x: number, y: number }} coordinates
     * @returns {{ width: number, height: number }}
     */
    detailed: function (game, images, coordinates) {
      game.context.fillStyle = '#888';
      game.context.fillRect(coordinates.x, coordinates.y, 100, 80);
      game.context.fillStyle = 'black';
      game.context.font = `16px sans-serif`;
      game.context.fillText(`Dust x${game.game_data.town.available_dust}`, coordinates.x + 1, coordinates.y + 15, 90);
      const image = images['item_dust'];
      if (image) {
        game.renderer.renderRawImage(image, coordinates.x + 5, coordinates.y + 10, 80, 80);
      }
      const is_box_hovered = mouseIsInArea(game, { x: coordinates.x, y: coordinates.y, width: 100, height: 80 });
      if (is_box_hovered) {
        showPopupOnMousePosition(game, `${DUST_DESCRIPTION}Available Dust: ${game.game_data.town.available_dust}`);
      }
      return { width: 100, height: 80 };
    },
  },
};
