import { Game } from './engine.mjs';

export type ControllerSystemMethod = (game: Game, elapsedTicks: number) => void;
export type SystemMethod = (game: Game) => void;

export type ImageAssetPartialInformation = {
  origin: { x: number; y: number; w: number; h: number };
  target: { w: number; h: number };
};

export type AudioAssetKind = 'music' | 'effect';

export type AudioAssetPartialInformation = {
  name: string;
  kind: AudioAssetKind;
};

export type AudioAssets = {
  snippets: {
    [id: string]: {
      kind: AudioAssetKind;
      media: HTMLAudioElement;
    };
  };
};

export type CharacterRelationshipStatus = {
  is_friend?: boolean;
  is_team_member?: boolean;
  is_hardship_companion?: boolean;
  is_partner?: boolean;
  is_rival_bully?: boolean;
  is_rival_target?: boolean;
};

export type CharacterBackstoryRelation = {
  id: CharacterId;
  status: CharacterRelationshipStatus;
  interactions: {
    friendship_interactions: number;
    romance_interactions: number;
    rivalry_bullies_interactions: number;
    rivalry_is_bullied_interactions: number;
    did_protect_interactions: number;
    protected_by_interactions: number;
    guilt_interactions: number;
    hardship_interactions: number;
  };
};

export type EventCharacterInfo = {
  character: CharacterId;
  relationship: CharacterRelationshipStatus;
  does_not_recognize_character?: boolean;
  does_not_recognize_target?: boolean;
};

export type EventEnemyInfo = {
  kind: EnemyKind;
};

export type EventItemInfo = {
  name: string;
  level: number;
};

export type CharacterBackstoryOriginsEvent =
  | { type: 'vagrant'; cycle: number }
  | { type: 'vagrant_magic'; cycle: number }
  | { type: 'commoner'; cycle: number }
  | { type: 'commoner_money'; cycle: number }
  | { type: 'commoner_magic'; cycle: number }
  | { type: 'noble_money'; cycle: number }
  | { type: 'noble_magic'; cycle: number }
  | { type: 'rival_encounter_childhood_bullies'; cycle: number; rival: EventCharacterInfo }
  | { type: 'rival_encounter_childhood_is_bullied'; cycle: number; rival: EventCharacterInfo }
  | { type: 'rival_encounter_adulthood_bullies'; cycle: number; rival: EventCharacterInfo }
  | { type: 'rival_encounter_adulthood_is_bullied'; cycle: number; rival: EventCharacterInfo }
  | { type: 'friend_encounter_childhood'; cycle: number; friend: EventCharacterInfo }
  | { type: 'friend_encounter_adulthood'; cycle: number; friend: EventCharacterInfo }
  | { type: 'girlfriend_encounter_childhood'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'girlfriend_encounter_adulthood'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'alone_town_event'; cycle: number }
  | { type: 'group_town_event'; cycle: number; group: EventCharacterInfo[] }
  | { type: 'alone_capture_event'; cycle: number; enemies: EventEnemyInfo[] }
  | { type: 'group_capture_event'; cycle: number; enemies: EventEnemyInfo[]; group: EventCharacterInfo[] };

export type CharacterBackstoryTownSharedEvent =
  | 'upgrading_item'
  | 'watched_upgrade_item'
  | 'training'
  | 'watched_training'
  | 'upgrading_item_watching_training'
  | 'training_watching_upgrading_item'
  | 'patrol_round'
  | 'evening_guard_round'
  | 'gathering_food'
  | 'town_cleaning';

export type CharacterBackstoryFriendshipEvent =
  | CharacterBackstoryTownSharedEvent
  | 'eating_lively_conversation'
  | 'friendly_spar';
export type CharacterBackstoryRomanceEvent =
  | CharacterBackstoryTownSharedEvent
  | 'eating_lively_conversation'
  | 'friendly_spar'
  | 'friendly_spar_with_fun_tops'
  | 'friendly_spar_with_fun_bots'
  | 'late_afternoon_stroll';
export type CharacterBackstoryRivalryEvent =
  | CharacterBackstoryTownSharedEvent
  | 'fight_fair_win'
  | 'fight_taunts'
  | 'prank_hides_bag'
  | 'fight_then_prank_hides_bag'
  | 'fight_cheats'
  | 'evening_guard_round_prank_ties_up'
  | 'town_cleaning_prank_ties_up'
  | 'fight_then_prank_ties_up'
  | 'prank_hides_clothes';

export type CharacterBackstoryTownEvent =
  | { type: 'arrival_base_alone'; cycle: number }
  | { type: 'arrival_base_group'; cycle: number; group: EventCharacterInfo[] }
  | { type: 'arrival_alone_dungeon'; cycle: number }
  | { type: 'arrival_team_dungeon'; cycle: number; group: EventCharacterInfo[] }
  | { type: 'arrival_alone_rescued_people'; cycle: number; rescuees: EventCharacterInfo[] }
  | { type: 'arrival_team_rescued_people'; cycle: number; group: EventCharacterInfo[]; rescuees: EventCharacterInfo[] }
  | { type: 'arrival_team_was_rescued_alone'; cycle: number; group: EventCharacterInfo[] }
  | {
      type: 'arrival_team_was_rescued_group';
      cycle: number;
      group: EventCharacterInfo[];
      rescuees: EventCharacterInfo[];
    }
  | { type: 'alone_event'; cycle: number }
  | { type: 'friendship_start'; cycle: number; friend: EventCharacterInfo }
  | { type: 'friendship_event'; cycle: number; friend: EventCharacterInfo; kind: CharacterBackstoryFriendshipEvent }
  | { type: 'friendship_end'; cycle: number; other: EventCharacterInfo }
  | { type: 'romance_start'; mode: 'standard' | 'rivalry_ends'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'romance_event'; cycle: number; girlfriend: EventCharacterInfo; kind: CharacterBackstoryRomanceEvent }
  | { type: 'rivalry_start_bullies'; cycle: number; rival: EventCharacterInfo }
  | { type: 'rivalry_start_is_bullied'; cycle: number; rival: EventCharacterInfo }
  | { type: 'rivalry_event_bullies'; cycle: number; rival: EventCharacterInfo; kind: CharacterBackstoryRivalryEvent }
  | { type: 'rivalry_event_is_bullied'; cycle: number; rival: EventCharacterInfo; kind: CharacterBackstoryRivalryEvent }
  | { type: 'rivalry_end'; cycle: number; rival: EventCharacterInfo }
  | { type: 'class_changed_event'; cycle: number; new_role: BaseCharacterRole }
  | { type: 'upgrade_event'; cycle: number; item: EventItemInfo }
  | { type: 'alone_capture_town_corruption'; cycle: number }
  | {
      type: 'group_capture_town_corruption';
      cycle: number;
      capturedGroup: EventCharacterInfo[];
      kind: 'fought_back' | 'overpowered' | 'ambushed';
    };

export type EventCharacterCriticalInfo = {
  name: string;
};

export type CharacterBackstoryExplorationEvent =
  | { type: 'alone_start_exploration'; cycle: number }
  | { type: 'team_start_exploration'; cycle: number; group: EventCharacterInfo[] }
  | { type: 'alone_class_changed'; cycle: number; enemies: EventEnemyInfo[]; new_role: CharacterRole }
  | {
      type: 'team_class_changed';
      cycle: number;
      enemies: EventEnemyInfo[];
      group: EventCharacterInfo[];
      new_role: CharacterRole;
    }
  | { type: 'alone_loot'; cycle: number; loot: EventItemInfo }
  | { type: 'team_loot'; cycle: number; group: EventCharacterInfo[]; loot: EventItemInfo }
  | { type: 'alone_rescue'; cycle: number; rescuee: EventCharacterInfo }
  | { type: 'team_rescue'; cycle: number; group: EventCharacterInfo[]; rescuee: EventCharacterInfo }
  | { type: 'alone_becomes_rescued'; cycle: number; rescuer: EventCharacterInfo }
  | { type: 'team_becomes_rescued'; cycle: number; group: EventCharacterInfo[] }
  | { type: 'alone_becomes_critical'; cycle: number; enemies: EventEnemyInfo[]; critical: EventCharacterCriticalInfo }
  | {
      type: 'team_becomes_critical';
      cycle: number;
      enemies: EventEnemyInfo[];
      group: EventCharacterInfo[];
      critical: EventCharacterCriticalInfo;
    }
  | {
      type: 'alone_has_critical_companion_event';
      cycle: number;
      enemies: EventEnemyInfo[];
      teammate: EventCharacterInfo;
      critical: EventCharacterCriticalInfo;
    }
  | {
      type: 'team_has_critical_companion_event';
      cycle: number;
      enemies: EventEnemyInfo[];
      group: EventCharacterInfo[];
      teammate: EventCharacterInfo;
      critical: EventCharacterCriticalInfo;
    }
  | { type: 'alone_has_saved_companion'; cycle: number; rescuee: EventCharacterInfo }
  | { type: 'alone_has_saved_companions'; cycle: number; rescuees: EventCharacterInfo[] }
  | { type: 'team_has_saved_companion'; cycle: number; group: EventCharacterInfo[]; rescuee: EventCharacterInfo }
  | { type: 'team_has_saved_companions'; cycle: number; group: EventCharacterInfo[]; rescuees: EventCharacterInfo[] }
  | {
      type: 'alone_has_lost_companion_during_exploration';
      cycle: number;
      enemies: EventEnemyInfo[];
      captured: EventCharacterInfo;
    }
  | {
      type: 'alone_has_lost_companions_during_exploration';
      cycle: number;
      enemies: EventEnemyInfo[];
      capturedGroup: EventCharacterInfo[];
    }
  | {
      type: 'team_has_lost_companion_during_exploration';
      cycle: number;
      enemies: EventEnemyInfo[];
      group: EventCharacterInfo[];
      captured: EventCharacterInfo;
    }
  | {
      type: 'team_has_lost_companions_during_exploration';
      cycle: number;
      enemies: EventEnemyInfo[];
      group: EventCharacterInfo[];
      capturedGroup: EventCharacterInfo[];
    }
  | {
      type: 'alone_has_lost_companion_during_exit';
      cycle: number;
      enemies: EventEnemyInfo[];
      captured: EventCharacterInfo;
    }
  | {
      type: 'alone_has_lost_companions_during_exit';
      cycle: number;
      enemies: EventEnemyInfo[];
      capturedGroup: EventCharacterInfo[];
    }
  | {
      type: 'team_has_lost_companion_during_exit';
      cycle: number;
      enemies: EventEnemyInfo[];
      group: EventCharacterInfo[];
      captured: EventCharacterInfo;
    }
  | {
      type: 'team_has_lost_companions_during_exit';
      cycle: number;
      enemies: EventEnemyInfo[];
      group: EventCharacterInfo[];
      capturedGroup: EventCharacterInfo[];
    }
  | { type: 'alone_becomes_captured'; cycle: number; enemies: EventEnemyInfo[] }
  | {
      type: 'alone_becomes_captured_with_companions';
      cycle: number;
      enemies: EventEnemyInfo[];
      capturedGroup: EventCharacterInfo[];
    }
  | { type: 'team_becomes_captured_alone'; cycle: number; enemies: EventEnemyInfo[]; group: EventCharacterInfo[] }
  | {
      type: 'team_becomes_captured_with_companions';
      cycle: number;
      enemies: EventEnemyInfo[];
      group: EventCharacterInfo[];
      capturedGroup: EventCharacterInfo[];
    };

export type CharacterBackstoryCaptureEvent =
  | { type: 'alone_capture_start'; cycle: number }
  | { type: 'alone_capture_start_brought_to_girlfriend'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'group_capture_start'; cycle: number; capturedGroup: EventCharacterInfo[] }
  | {
      type: 'group_capture_start_brought_to_girlfriend';
      cycle: number;
      girlfriend: EventCharacterInfo;
      capturedGroup: EventCharacterInfo[];
    }
  | {
      type: 'group_capture_start_with_girlfriend';
      cycle: number;
      girlfriend: EventCharacterInfo;
      capturedGroup: EventCharacterInfo[];
    }
  | { type: 'paired_capture_start_with_girlfriend'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'girlfriend_reunited_after_capture'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'girlfriend_reunited_after_talked'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'girlfriend_reunited_after_other_talked'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'alone_wait_event'; cycle: number }
  | { type: 'alone_play_with_event'; cycle: number; enemies: EventEnemyInfo[] }
  | {
      type: 'alone_drain_into_girlfriend_event';
      cycle: number;
      girlfriend: EventCharacterInfo;
      equipment: { name: string; level: number };
    }
  | { type: 'alone_drain_event'; cycle: number; equipment: { name: string; level: number } }
  | { type: 'alone_magic_event'; cycle: number }
  | { type: 'companion_play_with_event'; cycle: number; captured: EventCharacterInfo; enemies: EventEnemyInfo[] }
  | { type: 'companion_magic_event'; cycle: number; captured: EventCharacterInfo }
  | { type: 'girlfriend_wait_event'; cycle: number; girlfriend: EventCharacterInfo }
  | {
      type: 'girlfriend_play_with_event';
      cycle: number;
      girlfriend: EventCharacterInfo;
      enemies: EventEnemyInfo[];
      state: 'secret' | 'known';
    }
  | { type: 'girlfriend_magic_event'; cycle: number; girlfriend: EventCharacterInfo; state: 'secret' | 'known' }
  | { type: 'alone_interrogate_event_stayed_silent'; cycle: number }
  | { type: 'alone_interrogate_event_talked'; cycle: number }
  | { type: 'alone_interrogate_event_denounced_girlfriend'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'companion_interrogate_event_stayed_silent'; cycle: number; captured: EventCharacterInfo }
  | {
      type: 'companion_interrogate_event_self_denounced_own_girlfriend';
      cycle: number;
      captured: EventCharacterInfo;
      girlfriend: EventCharacterInfo;
    }
  | {
      type: 'companion_interrogate_event_self_denounced_other_girlfriend';
      cycle: number;
      captured: EventCharacterInfo;
      targetGirlfriend: EventCharacterInfo;
    }
  | {
      type: 'companion_interrogate_event_other_denounced_self_girlfriend';
      cycle: number;
      captured: EventCharacterInfo;
      girlfriend: EventCharacterInfo;
    }
  | {
      type: 'companion_interrogate_event_other_denounced_other_girlfriend';
      cycle: number;
      captured: EventCharacterInfo;
      targetGirlfriend: EventCharacterInfo;
    }
  | { type: 'girlfriend_interrogate_event_stayed_silent'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'girlfriend_interrogate_event_broke'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'transmutation_prepare_1_alone'; cycle: number }
  | { type: 'transmutation_prepare_2_alone'; cycle: number }
  | { type: 'transmutation_prepare_3_alone'; cycle: number }
  | { type: 'transmutation_prepare_1_in_front_of_girlfriend'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'transmutation_girlfriend_is_prepared_1'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'transmutation_prepare_2_in_front_of_girlfriend'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'transmutation_girlfriend_is_prepared_2'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'transmutation_prepare_3_in_front_of_girlfriend'; cycle: number; girlfriend: EventCharacterInfo }
  | { type: 'transmutation_girlfriend_is_prepared_3'; cycle: number; girlfriend: EventCharacterInfo }
  | {
      type: 'transmutation_girlfriend_is_executed';
      cycle: number;
      girlfriend: EventCharacterInfo;
      equipment: { name: string; level: number };
    };

export type CharacterBackstoryTransmutedEvent =
  | { type: 'transmutation_execute_alone'; cycle: number; equipment: { name: string; level: number } }
  | {
      type: 'transmutation_execute_in_front_of_girlfriend';
      cycle: number;
      equipment: { name: string; level: number };
      girlfriend: EventCharacterInfo;
    }
  | { type: 'alone_event'; cycle: number }
  | {
      type: 'drain_upgrade_event';
      cycle: number;
      teammate: EventCharacterInfo;
      equipment: { name: string; level: number };
    }
  | { type: 'enemy_use_event'; cycle: number; enemies: EventEnemyInfo[] }
  | { type: 'enemy_group_use_event'; cycle: number; enemies: EventEnemyInfo[]; capturedGroup: EventCharacterInfo[] }
  | { type: 'team_use_event'; cycle: number; teammate: EventCharacterInfo }
  | { type: 'team_group_use_event'; cycle: number; teammate: EventCharacterInfo; capturedGroup: EventCharacterInfo[] }
  | {
      type: 'team_upgrade_event';
      cycle: number;
      teammate: EventCharacterInfo;
      equipment: { name: string; level: number };
    }
  | { type: 'turned_into_dust_event'; cycle: number };

export type CharacterBackstoryEvent =
  | { location: 'origins'; content: CharacterBackstoryOriginsEvent[] }
  | { location: 'town'; cycle_start: number; content: CharacterBackstoryTownEvent[] }
  | { location: 'exploration'; cycle_start: number; content: CharacterBackstoryExplorationEvent[] }
  | { location: 'captured'; cycle_start: number; content: CharacterBackstoryCaptureEvent[] }
  | { location: 'transmuted'; cycle_start: number; content: CharacterBackstoryTransmutedEvent[] };

export type CharacterBackstory = {
  relations: CharacterBackstoryRelation[];
  events: CharacterBackstoryEvent[];
};

export type SkillId = string;
export type StatusEffectId = string;
export type EnemyAttackId = string;

export type CharacterId = string;

export type BaseCharacterRole = 'adventurer' | 'warrior' | 'rogue' | 'mage';
export type ExtraCharacterRole = 'mirror' | 'statue';
export type CharacterRole = BaseCharacterRole | ExtraCharacterRole;

export type StatsModifierCharacteristics =
  | { type: 'hp_modifier'; additional_hp: number }
  | { type: 'mp_modifier'; additional_mp: number }
  | { type: 'mp_regen_modifier'; additional_mp_regen: number }
  | { type: 'damage_modifier'; additional_damage: number }
  | { type: 'heal_modifier'; additional_heal: number }
  | { type: 'mp_cost_modifier'; mp_cost_divider: number }
  | { type: 'cooldown_modifier'; cooldown_divider: number }
  | { type: 'redirect_attack_to_character_modifier'; increase_percent: number }
  | { type: 'redirected_attack_damage_modifier'; damage_multiplier: number }
  | { type: 'evasion_modifier'; increase_percent: number }
  | { type: 'counterspell_self_modifier'; increase_percent: number }
  | { type: 'counterspell_team_modifier'; increase_percent: number };

export type InGameStatistics = {
  /** Maximum number of hit points for this character. */
  max_hp: number;
  /** Maximum number of magic points for this character */
  max_mp: number;
  /** Passive HP regen, in HP per tick. There is 20 ticks in one second. */
  hp_regen: number;
  /** Passive MP regen, in MP per tick. There is 20 ticks in one second. */
  mp_regen: number;
  /** Modifier for attack damage. +10 means that attacks do +10 HP of damage. Damage is capped to -0HP and can't "heal". */
  attack_additional_hp: number;
  /** Modifier for heal regeneration. +10 means that heals regen +10 HP, -10 means that heals regen -10 HP. Heals are capped to +0HP and can't "damage". */
  heal_additional_hp: number;
  /** Modifier for the cooldown. 0.8 means that everything is 25% slower, 2.0 means that everything is 50% faster. */
  skill_cooldown_divider: number;
  /** Modifier for MP cost of spells. 0.8 means that everything is 25% more expensive, 2.0 means that everything is 50% cheaper */
  skill_mp_cost_divider: number;
  /** Chance for how likely the character is to evade an attack. 0 is a 0% chance, 25 is a 25% chance, and so on. Capped between 0.0 and 1.0 */
  evasion_chance: number;
  /** Chance for how likely the character is to block an incoming spell against themself. 0 is a 0% chance, 25 is a 25% chance, and so on. Capped between 0.0 and 1.0 */
  self_counterspell_chance: number;
  /** Chance for how likely the character is to block an incoming spell against themself or a team member. 0 is a 0% chance, 25 is a 25% chance, and so on. Capped between 0.0 and 1.0 */
  team_counterspell_chance: number;
  /** Chance for how likely the character is to take an incoming attack against a team member for themself. 0 is a 0% chance, 25 is a 25% chance, and so on. Capped between 0.0 and 1.0 */
  team_attack_redirect_chance: number;
  /** Modifier for the damage of incoming redirected attacks. 0.8 means that attacks inflict 25% more damage, 2.0 means that attacks inflict 50% less damage. */
  redirected_attack_damage_modifier: number;
};

export type PartialInGameStatistics = {
  [property in keyof InGameStatistics]?: InGameStatistics[property];
};

export type InGameCurrentStats = {
  hp: number;
  mp: number;
};

export type SkillActionTargets =
  | { type: 'random_enemy' }
  | { type: 'lowest_hp_enemy' }
  | { type: 'highest_hp_enemy' }
  | { type: 'first_enemy' }
  | { type: 'last_enemy' }
  | { type: 'all_enemies' }
  | { type: 'self' }
  | { type: 'all_allies' };

export type SkillActionCharacteristics =
  | { type: 'damage'; damage_hp: number }
  | { type: 'damage_scaling_magic_cost'; ratio: number }
  | { type: 'heal'; heal_hp: number }
  | { type: 'drain_hp'; damage_hp: number; drain_ratio: number }
  | { type: 'drain_mana'; damage_mp: number; drain_ratio: number }
  | { type: 'drain_experience'; damage_experience: number; drain_ratio: number }
  | { type: 'dispel_all_effects' }
  | { type: 'apply_effect'; effect: StatusEffectId }
  | { type: 'dispel_effect'; effect: StatusEffectId }
  | { type: 'resist_effect'; effect: StatusEffectId };

export type SkillCharacteristics = {
  id: SkillId;
  name: string;
  cost_data: {
    mana: number;
    mana_percent: number;
    cooldown: number;
  };
  target: SkillActionTargets;
  action: SkillActionCharacteristics;
};

export type SkillBlockInformation =
  | { type: 'none' }
  | { type: 'all' }
  | { type: 'all_except_exact'; skills: Array<SkillId> }
  | { type: 'exact'; skills: Array<SkillId> };

export type StoredCharacterData = {
  id: CharacterId;
  name: string;
  role: CharacterRole;
  story: CharacterBackstory;
  level: number;
  partial_exp: number;
  skills: Array<SkillId>;
};

export type StoredEquipmentData = {
  type: 'body' | 'accessory' | 'weapon';
  name: string;
  image: string;
  level: number;
  stats: StatsModifierCharacteristics;
  character?: {
    is_discovered_by_player: boolean;
    data: StoredCharacterData;
  };
};

export type StatusEffectsCharacteristics = {
  id: StatusEffectId;
  name: string;
  modify_image:
    | { type: 'none' }
    | { type: 'static'; image: string }
    | { type: 'per_stack'; stacks: { [stack: number]: string } };
  block_skills:
    | { type: 'static'; blocks: SkillBlockInformation }
    | { type: 'per_stack'; stacks: { [stack: number]: SkillBlockInformation } };
  behavior:
    | { type: 'cooldown'; tick_count: number }
    | { type: 'game_over'; new_role?: CharacterRole }
    | {
        type: 'staged_progression';
        resist_skill?: SkillId;
        stages: Array<{
          stack: number;
          tick_count: number;
          next_messages: Array<string>;
          next:
            | { type: 'stack_upgrade'; next_stack: number }
            | { type: 'change_status'; next_status_effect_id: StatusEffectId };
        }>;
      };
};

export type InGameStatusEffect = {
  id: StatusEffectId;
  characteristics: StatusEffectsCharacteristics;
  current_stack: number;
  lifetime:
    | { type: 'to_remove' }
    | { type: 'fixed' }
    | { type: 'cooldown'; remaining_ticks: number; total_ticks: number }
    | { type: 'coolup'; ongoing_ticks: number; max_ticks: number };
};

export type InGameCharacterSkill = {
  id: SkillId;
  characteristics: SkillCharacteristics;
  cooldown: { remaining_ticks: number };
};

export type InGameCharacterEquipment = {
  body?: StoredEquipmentData;
  accessory?: StoredEquipmentData;
  weapon?: StoredEquipmentData;
};

export type InGameCharacterData = {
  type: 'character';
  base: StoredCharacterData;
  image: string;
  animation: { current: 'none' } | { current: 'attack'; remaining_frames: number };
  base_stats: InGameStatistics;
  current_stats: InGameCurrentStats;
  status_effects: Array<InGameStatusEffect>;
  skills: Array<InGameCharacterSkill>;
  equipment: InGameCharacterEquipment;
};

export type EnemyKind = 'ghost' | 'book' | 'mage' | 'high_sorceress' | 'rogue' | 'slime' | 'mirror';

export type EnemyAttackConditionCharacteristics =
  | { type: 'no_condition' }
  | { type: 'target_not_critical' }
  | { type: 'target_vulnerable'; vulnerability_ratio: number }
  | { type: 'target_critical' }
  | { type: 'target_game_over' };

export type InGameEnemySkill = {
  id: SkillId;
  condition: EnemyAttackConditionCharacteristics;
  characteristics: SkillCharacteristics;
  cooldown: { remaining_ticks: number };
};

export type InGameEnemyData = {
  type: 'enemy';
  name: string;
  kind: EnemyKind;
  image: string;
  animation: { current: 'none' } | { current: 'attack'; remaining_frames: number };
  level: number;
  exp_value: number;
  base_stats: InGameStatistics;
  current_stats: InGameCurrentStats;
  skills: Array<InGameEnemySkill>;
};

export type ScrollController = {
  scroll_offset: number;
  is_elevator_selected: boolean;
};

export type TownGameData = {
  screen: 'town' | 'role_change' | 'equipment_upgrade' | 'equipment_cleanup' | 'expedition' | 'backstory';
  background_image: string;
  unused_equipment: Array<StoredEquipmentData>;
  available_team: Array<StoredCharacterData>;
  available_dust: number;
  change_role:
    | { active: false }
    | {
        active: true;
        target: StoredCharacterData;
        target_role: BaseCharacterRole;
        retarget_end_cycle: number;
      };
  use_dust_for_upgrade: boolean;
  upgrade_equipment:
    | { active: false }
    | {
        active: true;
        handler: StoredCharacterData | null;
        target: StoredEquipmentData | null;
        retarget_end_cycle: number;
      };
  cleanup_equipment: { active: false } | { active: true; target: StoredEquipmentData };
  expedition_slots: [
    { character?: StoredCharacterData; equipment: InGameCharacterEquipment },
    { character?: StoredCharacterData; equipment: InGameCharacterEquipment },
    { character?: StoredCharacterData; equipment: InGameCharacterEquipment },
    { character?: StoredCharacterData; equipment: InGameCharacterEquipment },
  ];
  backstory: { selected_character?: CharacterId } & ScrollController;
  backstory_characters_controller: ScrollController;
  town_characters_controller: ScrollController;
  role_characters_controller: ScrollController;
  upgrade_characters_controller: ScrollController;
  expedition_characters_controller: ScrollController;
  town_equipment_controller: ScrollController;
  upgrade_equipment_controller: ScrollController;
  cleanup_equipment_controller: ScrollController;
  expedition_equipment_controller: ScrollController;
};

export type LevelEnemySpawnData = {
  min_level: number;
  max_level: number;
  type: 'standard' | 'small' | 'large' | 'ambush';
  kind?: {
    kind: EnemyKind;
    weight: number;
  };
  quantity?: {
    quantity: number;
    weight: number;
  };
};

export type DungeonGameData = {
  level: {
    background_image: string;
    exploration_cycles: number;
    is_round_active: boolean;
  };
  spawn: {
    ambush_cooldown: number;
    ambush_time: number;
  };
  rewards: {
    current_reward_chance: number;
    last_reward_message: string;
    saved_characters: Array<StoredCharacterData>;
    gathered_equipments: Array<StoredEquipmentData>;
  };
  team: Array<InGameCharacterData>;
  enemy: Array<InGameEnemyData>;
};

export type DisplayButtonContent =
  | { type: 'fixed'; width: number; height: number }
  | { type: 'text'; text: string; font_size: number; padding: { x: number; y: number } }
  | { type: 'fixed_text'; text: string; font_size: number; width: number; height: number };

export type DisplayButtonInformation = {
  x: number;
  y: number;
  background: { base: string; hover: string };
  content: DisplayButtonContent;
};

export type DisplayPopupInformation = {
  coordinates_base_position: 'top_left' | 'bottom_left' | 'top_right' | 'bottom_right';
  coordinates_adjusted_position: 'overflow' | 'sticky_border' | 'corner_on_border';
  x: number;
  y: number;
  color: { border: string; background: string };
  content: DisplayButtonContent;
};

export type DisplayParticleInformation = {
  text: string;
  color: { r: number; g: number; b: number };
  x: number;
  y: number;
  dx: number;
  dy: number;
  lifetime: number;
  alpha_lifetime: number;
};

export type PauseGameStateData = {
  is_paused: boolean;
  pause_button_held: boolean;
  effects_volume: ScrollController;
  music_volume: ScrollController;
};

export type GlobalGameStateData = {
  screen: 'town' | 'dungeon' | 'game_over';
  current_cycle: number;
  current_corruption_ratio: number;
  items_outside_town: Array<StoredEquipmentData>;
  characters_outside_town: Array<StoredCharacterData>;
  dusted_characters: Array<StoredCharacterData>;
  dungeon_knowledge: {
    interrogations_completed: Array<CharacterId>;
    girlfriends_known: Array<CharacterId>;
    transmutation_states: Array<{ id: CharacterId; step: 1 | 2 | 3 }>;
    forced_drain_into: Array<{ character: CharacterId; item_character: CharacterId }>;
  };
  game_over_cooldown: number;
  game_over_scroll: ScrollController;
  game_over_backstory: { selected_character?: string } & ScrollController;
};

export type GameData = {
  version: string;
  game_state: GlobalGameStateData;
  pause: PauseGameStateData;
  town: TownGameData;
  dungeon: DungeonGameData;
  particles: Array<DisplayParticleInformation>;
  popup: { state: 'none' } | ({ state: 'displayed' } & DisplayPopupInformation);
};
