const fs = require('fs');
const path = require('path');

/**
 * @param {string} file
 * @returns {Promise<void>}
 */
function createDirectory(directory) {
  return new Promise((resolve, reject) => {
    fs.mkdir(directory, (err) => {
      if (err) {
        reject(err);
        return;
      }
      resolve();
    });
  });
}

/**
 * @param {string} file
 * @returns {Promise<void>}
 */
function deleteFileOrDirectory(file) {
  return new Promise((resolve, reject) => {
    fs.rm(file, { recursive: true, force: true }, (err) => {
      if (err) {
        reject(err);
        return;
      }
      resolve();
    });
  });
}

/**
 * @param {string} path
 * @returns {Promise<fs.Dirent[]>}
 */
function getDirectoryContents(path) {
  return new Promise((resolve, reject) => {
    fs.readdir(path, { withFileTypes: true }, (err, files) => {
      if (err) {
        reject(err);
        return;
      }
      resolve(files);
    });
  });
}

/**
 * @param {string} directoryPath
 * @param {{ excluding?: string[] }} [options]
 * @returns {Promise<Array<{ action: 'createDir' | 'copy', path: string }>>}
 */
async function getActionsFromDirectory(directoryPath, options) {
  options = options || {};
  const excluding = options.excluding || [];
  if (excluding.some((f) => directoryPath === f)) return [];
  const files = await getDirectoryContents(directoryPath);
  const result = [{ action: 'createDir', path: directoryPath }];
  for (const file of files) {
    const filePath = path.join(directoryPath, file.name);
    if (excluding.some((f) => filePath === f)) continue;
    if (file.isDirectory()) {
      result.push(...(await getActionsFromDirectory(filePath, options)));
    } else if (file.isFile()) {
      result.push({ action: 'copy', path: filePath });
    }
  }
  return result;
}

(async () => {
  console.info("Deploying version to 'dist/'...");
  try {
    await deleteFileOrDirectory('dist');
  } catch (e) {
    // NO OP
  }

  await createDirectory('dist');

  /** @type {Awaited<ReturnType<getActionsFromDirectory>>} */
  const actionList = [
    { action: 'copy', path: 'index.html' },
    ...(await getActionsFromDirectory('code', { excluding: [path.join('code', 'types.d.ts')] })),
    ...(await getActionsFromDirectory('assets', {
      excluding: [path.join('assets', 'sources'), path.join('assets', 'COPYRIGHT'), path.join('assets', 'LICENSE')],
    })),
  ];

  for (const action of actionList) {
    if (action.action === 'copy') {
      fs.copyFileSync(action.path, path.join('dist', action.path));
    }
    if (action.action === 'createDir') {
      await createDirectory(path.join('dist', action.path));
    }
  }
  console.info("Deployment successful!");
})();
